/*
 * SmoothB2.cpp
 * Dennis Lew (av504@torfree.net)
 *
 * ROBOT       : SmoothB2
 * RACES       : Can be used for the remaining races.
 * CONFIDENTIAL: No
 * DATA FILE   : "SmoothB.dat" (or whatever DATA_FILE is set to)
 *               Contains data for: -cstlcomb.trk
 *                                  -hock.trk
 *                                  -hungary.trk
 *                                  -indygp.trk
 *                                  -monza-76.trk
 *
 * Uses the OOP structure, call 'getSmoothBInstance()' to get a
 * new instance.  All instances of SmoothB in a race will share
 * the path data.
 *    
 * SmoothB will try and create its own path around the track its
 * on.  The line is o.k, but there are problems with certain
 * combinations of sections.  It will make it around any track
 * with a decent pace (with no data file), but for maximum
 * performance I need to create a better path than its own.  The
 * path (in data file) is created manually using another program
 * I have made.
 *
 * If you want to use any part of this code in your own robot, or
 * are curious about anything I've done here, just let me know.
 *
 * <CHANGES>
 * v1.4: -fixed NULL pointer bug when trying to load data
 *       -improved collision detection
 *       -robot will not actively try to pass other cars (UNDER DEVELOPMENT)
 *       -robot should be able to reliably finish races now
 *
 * v1.5: -bugs introduced by earlier improvements have been fixed
 *       -robot runs better on tracks with no data
 *
 * v1.6: -complete rewrite of the anti-collision code (MUCH BETTER)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "os.h"
#include "car.h"
#include "track.h"

//#define SB_TESTING 1

// CONSTANTS
static const char   *SB1_AUTHOR     = "Dennis Lew (av504@torfree.net)";
static const char   *SB1_DESCRIPTION= "v1.6: Sep 21, 2001";
static const char   *SB1_NAME1      = "SBv1r4";
static const char   *SB1_NAME2      = "SBv1r4-%d";
static const int     SB1_NAME_LEN   = 10;
static const int     SB1_NOSE       = oYELLOW;
static const int     SB1_TAIL       = oBLACK;
static const char   *SB1_BITMAP2D   = "car_yellow_black";
static const char   *SB1_MODEL3D    = NULL;

static const char   *DATA_FILE      = "SBv1r4.dat";
static const double  MY_PI          =  3.14159265358979323846;
static const double  MY_PI21        =  2.0 * MY_PI;
static const double  MY_PI12        =  0.5 * MY_PI;
static const double  MY_PI14        =  0.25 * MY_PI;
static const double  SCALE_CORNSPD  =  0.88;
static const double  SB1_FC         =  6.0 / (5.0 * 5280.0);
static const double  SB1_DT         =  1.95 * delta_time;
static const double  SB1_SAFE_TIME  =  7.5;
static const int     SB1_PASS_COUNT = 60;
static const double  SB1_IN_WID     =  2.0;
static const double  SB1_OUT_WID    =  5.5;
static const int     SB1_NO_PASSING =  1;   // make non-zero to prevent passing

// GLOBALS
static double MAX_ACCEL;


// STATIC FUNCTIONS
//-----START spd_func.h-----//
static double crn_spd(double radius);
static double crn_RAD2FT(double len, double rad, double wid);
static double crn_FT2RAD(double len, double rad, double wid);

static double d_brk(double v);
static double brkdist(double vi, double vf);
//-----END spd_func.h-----//


// CLASSES
//-----START Vector.h-----//
class SB1Vector {
public:
   // MEMBERS
   double x, y;

   // CONSTRUCTORS
   SB1Vector();
   SB1Vector(double vx, double vy);
   SB1Vector(double angle);
   SB1Vector(const SB1Vector &v);
   
   // METHODS
   void   add(const SB1Vector &v);
   void   sub(const SB1Vector &v);
   double mag();
   double mag2();
   double ang();
   void   scale(double factor);
   void   normalize();
   double dot(const SB1Vector &v);
   void   rotate(double angle);
};
//-----END Vector.h-----//

//-----START Line.h-----//
class SB1Line {
public:
   // MEMBERS
   SB1Vector pt;
   SB1Vector dir;

   // CONSTRUCTORS
   SB1Line();
   SB1Line(double px, double py, double dx, double dy);
   SB1Line(double px, double py, double d_ang);

   // METHODS
   SB1Vector *intersect(SB1Line &l);
   void      intersect(SB1Vector &np, SB1Line &l);
   SB1Vector *getPoint(double dist);
   void      getPoint(SB1Vector &np, double dist);
   void      movePoint(double dist);
};
//-----END Line.h-----//

//-----START Circle.h-----//
class SB1Circle {
public:
   // MEMBERS
   SB1Vector c;
   double   R;

   // CONSTRUCTORS
   SB1Circle();
   SB1Circle(double rad);
   SB1Circle(double cx, double cy, double rad);

   // METHODS
   SB1Vector *intersect(SB1Line &l);
   SB1Vector *intersect_F(SB1Line &l);
   void      intersect_F(SB1Vector &np, SB1Line &l);
   void      tangent2(SB1Line &s, SB1Line &f, SB1Circle &a);
   SB1Vector *getPoint(double angle);
   void      getPoint(SB1Vector &np, double angle);

private:
   // PRIVATE METHODS
   int intersect(SB1Line &l, double *d1, double *d2);
};
//-----END Circle.h-----//

//-----START PathComponent.h-----//
class SB1PathComponent {
public:
   // MEMBERS
   int    type;           // type of seg this component is on (1,0,-1)
   int    num;
   double len, wid, rad;  // length, width and radius of this seg
   double Sl, Sw;         // start position
   double Fl, Fw;         // finish position

   // CONSTRUCTORS
   SB1PathComponent();

   // METHODS
           void   set_seg_data(double l, double w, double r);
   virtual int    valid();
   virtual void   construct();
   virtual double to_side(double pos);
   virtual double get_nex_pos(SB1Line &cp, SB1Vector &np, 
                              SB1PathComponent &nex);
   virtual double dist(SB1Vector &pt);
};
//-----END PathComponent.h-----//

//-----START Straight.h-----//
class SB1Straight : public SB1PathComponent {
public:
   // MEMBERS
   SB1Line l;
   
   // CONSTRUCTORS
   SB1Straight();
   
   // METHODS
   int    valid();
   void   construct();
   double to_side(double pos);
   double get_nex_pos(SB1Line &cp, SB1Vector &np, SB1PathComponent &nex);
   double dist(SB1Vector &pt);

private:
   // PRIVATE METHODS
   void   construct_S();
   void   construct_T();
   double to_side_S(double pos);
   double to_side_T(double pos);
};
//-----END Straight.h-----//

//-----START Turn.h-----//
class SB1Turn : public SB1PathComponent {
public:
   // MEMBERS
   SB1Circle curv;
   double   Aw;    // Apex position
   
   // CONSTRUCTORS
   SB1Turn();
   
   // METHODS
   int    valid();
   void   construct();
   double to_side(double pos);
   double get_nex_pos(SB1Line &cp, SB1Vector &np, SB1PathComponent &nex);
   double dist(SB1Vector &pt);

private:
   // PRIVATE METHODS
   void   construct_S();
   void   construct_T();
   double to_side_S(double pos);
   double to_side_T(double pos);
};
//-----END Turn.h-----//

//-----START DrvPath.h-----//
class SB1DrvPath {
public:
   // MEMBERS
   // Driving path parameters
   int               seg_type;  // -1 right, 0 straight, 1 left
   int               num;       // number of components in this segment
   SB1PathComponent **pc;
   double            spd, spd_cur;   // spd parameters
   double            apex;

   // CONSTRUCTORS
   SB1DrvPath(int type, int n);

   // DESTRUCTOR
   ~SB1DrvPath();
   
   // METHODS
          void             construct();
          SB1PathComponent *get_PC(double pos);
   static void             best_path(SB1Turn &cs, SB1Turn &ns, SB1Turn &as);
   static void             adjust_path(SB1Turn &ps, SB1Turn &cs, SB1Turn &ns, SB1Turn &as);
   
private:
   // PRIVATE METHODS
          void construct_S();
          void construct_T();
   static void Update_cs(SB1Turn &cs, SB1Turn &ns);
   static void Update_as(SB1Turn &as, SB1Turn &ns);
};
//-----END DrvPath.h-----//

//-----START TrkData.h-----//
class SB1TrkData {
public:
   // MEMBERS
   char       *name;
   SB1DrvPath **path;
   int         nseg;

   // CONSTRUCTORS
   SB1TrkData();
   
   // DESTUCTOR
   ~SB1TrkData();
   
   // METHODS
   void set_path(Track *trk);
   void saveData(const char *fname, const int ver);
   int  loadData(const char *fname);
   
private:
   // PRIVATE METHODS
   void saveTXTData(const char *fname);
   int  loadTXTData(const char *fname);
   void saveBINData(const char *fname);
   int  loadBINData(const char *fname);
   void saveTurn(FILE *fp, SB1Turn &t);
   void saveStraight(FILE *fp, SB1Straight &t);
};
//-----END TrkData.h-----//

//-----START SBDriver.h-----//
// GLOBALS
static int        trk_num = -1;
static SB1TrkData trk_data;

class SBv1r4 : public Driver
{
public:
   // CONSTRUCTORS
   SBv1r4();
   SBv1r4(int i);
   
   // DESTRUCTOR
   ~SBv1r4();

   // METHODS
   con_vec drive(situation &s);
   
private:
   // PRIVATE MEMBERS
   char    *name;
   int      init;
   con_vec  result;
   con_vec *cp;
   int      lastpit;
   double   fuelstart;
   double   err;
   int      passTime;
   double   nearby[NEARBY_CARS];
#ifdef SB_TESTING
   int      SBnum;
#endif

   // PRIVATE METHODS
   void   get_RARS_ctrl_params(situation &s);
   double fuel_needed(Stage s, int lapsleft, double fpl);
   double car_mass(double fuel);
   double u(double slip);
   double iu(double cur_u);
   void   update_con_vec(SB1Vector &cV, SB1Vector &nV);
   void   get_pos(SB1PathComponent &pc, double Pl, double Pw, 
                  SB1Vector &pos);
   void   get_vel(SB1PathComponent &pc, double v, double vn, double Pl, 
                  SB1Vector &vel);
   double get_tgt_v(situation &s);
   double get_dist(int seg_ID, double ahead, SB1Vector &pos);
   double locate_cars(SB1Line &cV, situation &s);
   int    anti_collision(situation &s, SB1Vector &nV, double my_dist, 
                         int dir);
   void   alt_update_con_vec(situation &s, double lane, double bias, 
                             int trouble);
   double alt_get_tgt_v(situation &s, double lane);
};
//-----END SBDriver.h-----//


//-----START Vector.cpp-----//
// CONSTRUCTORS
SB1Vector::SB1Vector() 
{
   x = y = 0.0;
}

SB1Vector::SB1Vector(double vx, double vy) 
{
   x = vx;
   y = vy;
}

SB1Vector::SB1Vector(double angle)
{
   x = cos(angle);
   y = sin(angle);
}

SB1Vector::SB1Vector(const SB1Vector &v)
{
   x = v.x;
   y = v.y;
}

// METHODS
void SB1Vector::add(const SB1Vector &v) 
{
   x += v.x;
   y += v.y;
}
   
void SB1Vector::sub(const SB1Vector &v) 
{
   x -= v.x;
   y -= v.y;
}

double SB1Vector::mag() 
{
   return sqrt((x*x) + (y*y));
}

double SB1Vector::mag2() 
{
   return ((x*x) + (y*y));
}

double SB1Vector::ang()
{
   double retval;

   if (x != 0.0)
      retval = atan(y / x);
   else {
      if (y >= 0.0)
         retval =  MY_PI12;
      else
         retval = -MY_PI12;
   }

   if (x < 0.0)                       // Quad 2, Quad 3
      retval += MY_PI;
   else if ( (x >= 0.0)&&(y < 0.0) )  // Quad 4
      retval += MY_PI21;

   return retval;
}
   
void SB1Vector::scale(double factor) 
{
   x *= factor;
   y *= factor;
}
   
void SB1Vector::normalize()
{
   double m;

   m = mag();
   if (m != 0.0)
      scale(1.0 / m);
}

double SB1Vector::dot(const SB1Vector &v)
{
   return (x*v.x) + (y*v.y);
}

void SB1Vector::rotate(double angle)
{
   double m, a;

   m = mag();
   a = ang();
   x = m * cos(a + angle);
   y = m * sin(a + angle);
}
//-----END Vector.cpp-----//

//-----START Line.cpp-----//
// CONSTRUCTORS
SB1Line::SB1Line()
   : pt(), dir()
{
}

SB1Line::SB1Line(double px, double py, double dx, double dy)
   : pt(px, py), dir(dx, dy)
{
   dir.normalize();
}

SB1Line::SB1Line(double px, double py, double d_ang)
   : pt(px, py), dir(d_ang)
{
}

// METHODS
SB1Vector *SB1Line::intersect(SB1Line &l)
{
   SB1Vector *ret;

   ret = new SB1Vector(pt);
   intersect(*ret, l);
   return ret;
}

void SB1Line::intersect(SB1Vector &np, SB1Line &l)
{
   double N, D;

   N = (dir.x*(l.pt.y - pt.y)) - (dir.y*(l.pt.x - pt.x));
   D = (l.dir.x*dir.y) - (l.dir.y*dir.x);

   if (D == 0.0) {
      np.x = pt.x;
      np.y = pt.y;
      return;
   }

   l.getPoint(np, (N/D));
}

SB1Vector *SB1Line::getPoint(double dist)
{
   SB1Vector *ret;
		
   ret = new SB1Vector();
   getPoint(*ret, dist);
   return ret;
}

void SB1Line::getPoint(SB1Vector &np, double dist)
{
   np.x = pt.x + (dist*dir.x);
   np.y = pt.y + (dist*dir.y);
}

void SB1Line::movePoint(double dist)
{
   pt.x += dist * dir.x;
   pt.y += dist * dir.y;
}
//-----END Line.cpp-----//

//-----START Circle.cpp-----//
// CONSTRUCTORS
SB1Circle::SB1Circle()
   : c()
{
   R = 0.0;
}

SB1Circle::SB1Circle(double rad)
   : c()
{
   R = rad;
}

SB1Circle::SB1Circle(double cx, double cy, double rad)
   : c(cx, cy)
{
   R = rad;
}

// METHODS
SB1Vector *SB1Circle::intersect(SB1Line &l)
{
   SB1Vector *ret;
   double    E, F;
   double    d1, d2;
   int       rc;

   d1 = d2 = 0.0;
   rc = intersect(l, &d1, &d2);

   if (rc == 0) {
      ret = new SB1Vector(c);
      return ret;
   }
   else if (rc == 1) {
      // Only 1 solution
      ret = l.getPoint(d1);
      return ret;
   }

   E = l.pt.y + (d1 * l.dir.y);
   F = l.pt.y + (d2 * l.dir.y);
   if (  ( (E >= 0.0)&&(F >= 0.0) )||( (E < 0.0)&&(F < 0.0) )  ) {
      // both points are on same side of x axis
      // choose closest point
      if (fabs(d1) < fabs(d2))
         ret = l.getPoint(d1);
      else
         ret = l.getPoint(d2);
   }
   else if (E >= 0.0)
      ret = l.getPoint(d1);
   else
      ret = l.getPoint(d2);

   return ret;
}
	
SB1Vector *SB1Circle::intersect_F(SB1Line &l)
{
   SB1Vector *ret;

   ret = new SB1Vector();
   intersect_F(*ret, l);
   return ret;
}

void SB1Circle::intersect_F(SB1Vector &np, SB1Line &l)
{
   double d1, d2;
   int    rc;

   d1 = d2 = 0.0;
   rc = intersect(l, &d1, &d2);

   if (rc == 0) {
      np.x = c.x;
      np.y = c.y;
      return;
   }
   else if (rc == 1) {
      // Only 1 solution
      l.getPoint(np, d1);
      return;
   }

   // choose closest point
   if (fabs(d1) < fabs(d2))
      l.getPoint(np, d1);
   else
      l.getPoint(np, d2);
}

void SB1Circle::tangent2(SB1Line &s, SB1Line &f, SB1Circle &a)
{
   SB1Vector I_p, O_p, *A_p;
   double   d;
   SB1Line   I;

   f.dir.rotate(MY_PI12);
   if (f.dir.x != -1.0)
      d = (s.pt.x - f.pt.x) / (1.0 + f.dir.x);
   else
      d = 0.0;

   if (d <= (f.pt.mag() - 2.0)) {
      // this situation causes problems, so must adjust (s) and (f)
      d = 0.5 * (s.pt.mag() + f.pt.mag());
      s.pt.normalize(); s.pt.scale(d);
      f.pt.normalize(); f.pt.scale(d);
   }
   f.getPoint(O_p, d);

   f.dir.rotate(-MY_PI12);
   if (fabs(s.dir.ang()-f.dir.ang()) > 0.0436332)
      s.intersect(I_p, f);
   else  { // s and f are parallel
      I_p.x = O_p.x - (10.0*s.dir.x);
      I_p.y = O_p.y - (10.0*s.dir.y);
   }

   I.pt.add(I_p);
   I.dir.add(O_p);
   I.dir.sub(I_p);
   I.dir.normalize();

   if (fabs(s.pt.mag()-a.R) <= 5.0)
      A_p = new SB1Vector(f.pt);
   else if (fabs(f.pt.mag()-a.R) <= 5.0)
      A_p = new SB1Vector(f.pt);
   else {
      A_p = a.intersect(I);
      if (A_p->ang() > f.pt.ang()) {
         A_p->x = f.pt.x; A_p->y = f.pt.y;
      }
   }

   I.dir.x = O_p.x; I.dir.y = O_p.y;
   I.dir.sub(*A_p);
   I.dir.normalize();

   if (I.dir.x != -1.0)
      d = (s.pt.x - A_p->x) / (1.0 + I.dir.x);
   else
      d = 0.0;

   R   = d;
   c.x = A_p->x + (d * I.dir.x);
   c.y = A_p->y + (d * I.dir.y);

   I_p.x = R - c.mag();
   if (I_p.x < a.R) {
      // Adjust turn
      I_p.y = a.R - I_p.x;
      I_p.x = c.mag();
      d     = I_p.x / I_p.y;
      if (d > 2.0)
         d = 2.0;
      I_p.x -= d * I_p.y;
      d     -= 1.0;
      R     -= d * I_p.y;
      c.normalize();
      c.scale(I_p.x);
   }

   delete A_p;
}

SB1Vector *SB1Circle::getPoint(double angle)
{
   SB1Vector *ret;
		
   ret = new SB1Vector();
   getPoint(*ret, angle);
   return ret;
}

void SB1Circle::getPoint(SB1Vector &np, double angle)
{
   np.x = c.x + (R * cos(angle));
   np.y = c.y + (R * sin(angle));
}


// PRIVATE METHODS
int SB1Circle::intersect(SB1Line &l, double *d1, double *d2)
{
   SB1Vector dif(l.pt);
   double   E, F, G, det;

   dif.sub(c);
		
   E = l.dir.mag() * l.dir.mag();
   F = 2.0 * dif.dot(l.dir);
   G = (dif.mag() * dif.mag()) - (R * R);

   det = (F*F) - (4.0*E*G);
   if (  (det < 0.0)||( (E == 0.0)&&(F == 0.0) )  ) {
      return 0;
   }

   if (E == 0.0) {
      // Only 1 solution
      *d1 = -G / F;
      return 1;
   }

   *d1 = (-F + sqrt(det)) / (2.0*E);
   *d2 = (-F - sqrt(det)) / (2.0*E);
   return 2;
}
//-----END Circle.cpp-----//

//-----START PathComponent.cpp-----//
// CONSTRUCTORS
SB1PathComponent::SB1PathComponent()
{
   type = num = 0;
   len  = wid = rad = 0.0;
   Sl   = Sw  = Fl  = Fw  = 0.0;
}

// METHODS
void SB1PathComponent::set_seg_data(double l, double w, double r)
{
   len = l;
   wid = w;
   rad = fabs(r);  // all turns treated as left turns

   if (r == 0.0)
      type =  0;  // straight
   else if (r > 0.0)
      type =  1;  // left turn
   else
      type = -1;  // right turn
}

int SB1PathComponent::valid()
{
   return 1;
}

void SB1PathComponent::construct()
{
}
	
double SB1PathComponent::to_side(double pos)
{
   return 0.0;
}	

double SB1PathComponent::get_nex_pos(SB1Line &cp, SB1Vector &np, 
                                    SB1PathComponent &nex)
{
   return 0.0;
}

double SB1PathComponent::dist(SB1Vector &pt)
{
   return 0.0;
}	
//-----END PathComponent.cpp-----//

//-----START Straight.cpp-----//
// CONSTRUCTORS
SB1Straight::SB1Straight()
   : SB1PathComponent(), l()
{
}

// METHODS
int SB1Straight::valid()
{
   return ((Fl - Sl) > 1.0);
}

void SB1Straight::construct()
{
   if (type == 0)
      construct_S();
   else
      construct_T();

   // make sure S/F positions are in the segment
   if (Sl < 0.0) {
      Sl = 0.0;
      Sw = to_side(Sl);
   }
   else if (Sl > len) {
      Sl = len;
      Sw = to_side(Sl);
   }
		
   if (Fl > len) {
      Fl = len;
      Fw = to_side(Fl);
   }
   else if (Fl < 0.0) {
      Fl = 0.0;
      Fw = to_side(Fl);
   }

   if (type == 0)
      construct_S();
   else
      construct_T();
}

double SB1Straight::to_side(double pos)
{
   if (type == 0)
      return to_side_S(pos);

   return to_side_T(pos);
}	

double SB1Straight::get_nex_pos(SB1Line &cp, SB1Vector &np, SB1PathComponent &nex)
{
   double MAX_dA = acos(1.0 - (0.048/cp.dir.mag2()));
   double dp     = np.x * delta_time;
   double pos, ideal, err_pct, cur_dir;

   MAX_dA = (MAX_dA > 0.0035) ? 0.0035 : MAX_dA;

   // find ideal direction and error pct
   cur_dir = cp.dir.ang();
   ideal   = l.dir.ang() - cur_dir;
   if (ideal > MY_PI)
      ideal -= MY_PI21;
   else if (ideal < -MY_PI)
      ideal += MY_PI21;
   pos     = to_side(cp.pt.y);
   err_pct = fabs(cp.pt.x - pos) / wid;
   err_pct = (err_pct > 1.0) ? 1.0 : err_pct;

   // find target position
   np.y = cp.pt.y + (l.dir.y * dp);
   if (np.y <= Fl)   // target is in this component
      np.x = pos + (l.dir.x * dp);
   else {            // target is in next component
      if (nex.type == 0) {
         pos  = nex.Sl + (np.y - Fl);
         np.x = nex.to_side(pos);
      }
      else {
         pos  = nex.Sl + ((np.y - Fl) / nex.rad);
         np.x = nex.to_side(pos) + nex.rad;
         np.y = 0.0;
         np.rotate(pos);
         if (nex.type == 1) {
            // left turn
            np.x -= nex.rad;
            np.y += Fl;
         }
         else {
            // right turn
            np.x  = wid + nex.rad - np.x;
            np.y += Fl;
         }
      }
   }
   
   // limit the change in direction
   np.sub(cp.pt);
   pos = np.ang() - cur_dir;
   if (pos > MY_PI)
      pos -= MY_PI21;
   else if (pos < -MY_PI)
      pos += MY_PI21;
   pos *= err_pct;
   pos += (1.0 - err_pct) * ideal;

   if (pos > MAX_dA)
      pos =  MAX_dA;
   else if (pos < -MAX_dA)
      pos = -MAX_dA;

   np.x = dp;
   np.y = 0.0;
   np.rotate(cur_dir + pos);
   return -err_pct;
}

double SB1Straight::dist(SB1Vector &pt)
{
   SB1Line   tl;
   SB1Vector tv;

   tl.pt.x  = pt.x;    tl.pt.y  = pt.y;
   tl.dir.x = l.dir.x; tl.dir.y = l.dir.y;
   tl.dir.rotate(-MY_PI12);

   tl.intersect(tv, l);
   if (tl.dir.x != 0.0)
      return ((tl.pt.x - tv.x) / tl.dir.x);
   else if (tl.dir.y != 0.0)
      return ((tl.pt.y - tv.y) / tl.dir.y);
   else
      return 100000.0;
}

// PRIVATE METHODS
void SB1Straight::construct_S()
{
   double Ml, Mw, a;
		
   l.pt.x  = Sw;
   l.pt.y  = Sl;
   l.dir.x = Fw - l.pt.x;
   l.dir.y = Fl - l.pt.y;

   a = l.dir.ang() - MY_PI12;
   if ( (fabs(a) > 0.4363)&&(Sl != Fl) ) {
      // Sw and Fw are too far apart
      Ml = 0.5*(Sl + Fl);
      Mw = to_side_S(Ml);
			
      l.pt.x  = Mw;
      l.pt.y  = Ml;
      l.dir.x = 0.0;
      l.dir.y = 1.0;
      if (a > 0.0)
         l.dir.rotate( 0.4363);
      else
         l.dir.rotate(-0.4363);

      Sw = to_side_S(Sl);
      Fw = to_side_S(Fl);
   }
   else
      l.dir.normalize();
}

void SB1Straight::construct_T()
{
   l.pt.x  = (Sw + rad) * cos(Sl);
   l.pt.y  = (Sw + rad) * sin(Sl);
   l.dir.x = ((Fw + rad) * cos(Fl)) - l.pt.x;
   l.dir.y = ((Fw + rad) * sin(Fl)) - l.pt.y;
   l.dir.normalize();
}

double SB1Straight::to_side_S(double pos)
{
   double    ret;
   SB1Vector *i;
   SB1Line    ln(0.0, pos, 1.0, 0.0);

   i = ln.intersect(l);
   i->sub(ln.pt);
   ret = i->mag();
   
   delete i;
   return ((ret < SB1_OUT_WID)?SB1_OUT_WID:((ret > (wid-SB1_OUT_WID))?(wid-SB1_OUT_WID):ret));
}	

double SB1Straight::to_side_T(double pos)
{
   double    ret;
   SB1Vector *i;
   SB1Line    ln(rad*cos(pos), rad*sin(pos), pos);

   i = ln.intersect(l);
   i->sub(ln.pt);
   ret = i->mag();

   delete i;
   return ((ret < SB1_OUT_WID)?SB1_OUT_WID:((ret > (wid-SB1_OUT_WID))?(wid-SB1_OUT_WID):ret));
}
//-----END Straight.cpp-----//

//-----START Turn.cpp-----//
// CONSTRUCTORS
SB1Turn::SB1Turn()
   : SB1PathComponent(), curv()
{
   Aw = 0.0;
}

// METHODS
int SB1Turn::valid()
{
   return ( ((Fl - Sl) > 1.0)||(type != 0) );
}

void SB1Turn::construct()
{
   if (type == 0)
      construct_S();
   else
      construct_T();
}

double SB1Turn::to_side(double pos)
{
   if (type == 0)
      return to_side_S(pos);

   return to_side_T(pos);
}	

double SB1Turn::get_nex_pos(SB1Line &cp, SB1Vector &np, SB1PathComponent &nex)
{
   double MAX_dA  = 2.0 * ((cp.dir.mag() * delta_time) / curv.R);
   double dp      = np.x * delta_time;
   int    left    = 1;
   int    trouble = 0;
   double pos, ideal, err_pct, cur_dir;

   MAX_dA = (MAX_dA < 0.0075) ? 0.0075 : MAX_dA;

   // find error pct
   cur_dir = cp.dir.ang();
   if (type) {
      pos     = to_side(cp.pt.ang());
      err_pct = fabs(cp.pt.mag() - rad - pos) / wid;
   }
   else {
      pos     = to_side(cp.pt.y);
      err_pct = fabs(cp.pt.x - pos) / wid;
   }
   err_pct = (err_pct > 1.0) ? 1.0 : err_pct;

   // find target position
   if (type) {
      // turn
      if ((pos+CARWID+rad) < cp.pt.mag())
         trouble = 1;
      np.x = cp.pt.x - curv.c.x;
      np.y = cp.pt.y - curv.c.y;
      if (Aw < 0.0) {
         left = 0;
         np.rotate(MY_PI);
      }
      np.normalize();
      np.scale(curv.R);
      ideal = np.ang() + MY_PI12;
      np.rotate(dp / curv.R);
      np.add(curv.c);
   }
   else {
      // straight
      np.x = pos - curv.c.x;
      np.y = cp.pt.y - curv.c.y;
      if (curv.c.x <= 0.0) {  // left turn
         if ((pos+CARWID) < cp.pt.x)
            trouble = 1;
         ideal = np.ang() + MY_PI12;
         np.rotate( dp / curv.R);
      }
      else {                  // right turn
         if ((pos-CARWID) > cp.pt.x)
            trouble = 1;
         left  = 0;
         ideal = np.ang() - MY_PI12;
         np.rotate(-dp / curv.R);
      }
      np.add(curv.c);

      if (np.y > Fl) {  // target is in next component
         if (nex.type == 0) {
            pos   = nex.Sl + (np.y - Fl);
            np.x  = nex.to_side(pos);
            ideal = MY_PI12;
         }
      }
   }

   ideal -= cur_dir;
   if (ideal > MY_PI)
      ideal -= MY_PI21;
   else if (ideal < -MY_PI)
      ideal += MY_PI21;

   // limit the change in direction
   np.sub(cp.pt);
   pos = np.ang() - cur_dir;
   if (pos > MY_PI)
      pos -= MY_PI21;
   else if (pos < -MY_PI)
      pos += MY_PI21;
   pos *= err_pct;
   pos += (1.0 - err_pct) * ideal;

   if (left) {
      if (pos > MAX_dA)
         pos =  MAX_dA;
      else if ( (pos < -0.05*MAX_dA)&&(ideal > -MAX_dA) )
         pos = -0.05*MAX_dA;
      else if (pos < -0.275*MAX_dA)
         pos = -0.275*MAX_dA;
   }
   else {
      if ( (pos > 0.05*MAX_dA)&&(ideal < MAX_dA) )
         pos =  0.05*MAX_dA;
      else if (pos > 0.275*MAX_dA)
         pos =  0.275*MAX_dA;
      else if (pos < -MAX_dA)
         pos = -MAX_dA;
   }

   np.x = dp;
   np.y = 0.0;
   np.rotate(cur_dir + pos);
   return ((trouble) ? err_pct : -err_pct);
}

double SB1Turn::dist(SB1Vector &pt)
{
   SB1Vector tv;

   tv.x = pt.x; tv.y = pt.y;
   tv.sub(curv.c);
   return (tv.mag() - curv.R);
}

// PRIVATE METHODS
void SB1Turn::construct_S()
{
   // make sure S/F positions are in the segment
   if (Sl < 0.0) {
      Sl = 0.0;
      Sw = to_side(Sl);
   }
   else if (Sl > len) {
      Sl = len;
      Sw = to_side(Sl);
   }
		
   if (Fl > len) {
      Fl = len;
      Fw = to_side(Fl);
   }
   else if (Fl < 0.0) {
      Fl = 0.0;
      Fw = to_side(Fl);
   }
}

void SB1Turn::construct_T()
{
   // make sure S/F positions are in the segment
   if ((Sl+0.087266) >= MY_PI21) {
      Sl = 0.0;
      Sw = to_side(Sl);
   }
}

double SB1Turn::to_side_S(double pos)
{
   double    ret;
   SB1Vector *i;
   SB1Line    ln(0.0, pos, 1.0, 0.0);

   i = curv.intersect_F(ln);
   i->sub(ln.pt);
   ret = i->mag();

   delete i;
   return ((ret < SB1_IN_WID)?SB1_IN_WID:((ret > (wid-SB1_IN_WID))?(wid-SB1_IN_WID):ret));
}

double SB1Turn::to_side_T(double pos)
{
   double    ret;
   SB1Vector *i;
   SB1Line    ln(rad*cos(pos), rad*sin(pos), pos);

   i = curv.intersect_F(ln);
   i->sub(ln.pt);
   ret = i->mag();

   delete i;
   return ((ret < SB1_IN_WID)?SB1_IN_WID:((ret > (wid-SB1_OUT_WID))?(wid-SB1_OUT_WID):ret));
}
//-----END Turn.cpp-----//

//-----START DrvPath.cpp-----//
// CONSTRUCTORS
SB1DrvPath::SB1DrvPath(int type, int n)
{
   int i;
   
   seg_type = type;
   pc       = new SB1PathComponent*[n];
   if (pc != 0)
      num = n;
   else
      num = 0;
      
   for (i = 0; i < num; i++) {
      pc[i] = 0;
   }

   spd_cur = 0.0;
   spd     = (type == 0) ? 1.0 : SCALE_CORNSPD;
   apex    = 0.0;
}

// DESTRUCTOR
SB1DrvPath::~SB1DrvPath()
{
   int i;

   for (i = 0; i < num; i++) {
      delete pc[i];
   }
   delete [] pc;
}

// METHODS
void SB1DrvPath::construct()
{
   if (seg_type == 0)
      construct_S();
   else
      construct_T();
}

SB1PathComponent *SB1DrvPath::get_PC(double pos)
{
   int i, s, f;

   s = f = -1;
   for (i = 0; i < num; i++) {
      if (pc[i]->valid()) {
         f = i;
         if (s < 0)
            s = i;
      }
   }

   if (s < 0)
      return NULL;
   else if (pos < pc[s]->Sl)
      return pc[s];

   for (i = s; i <= f; i++) {
      if ( (pos >= pc[i]->Sl)&&(pos < pc[i]->Fl) )
         return pc[i];
   }

   return pc[f];
}

// Finds best path through corner.  Assumes (cs) is on straight before
// corner, (ns) is in the corner, and (as) is on a straight after corner
void SB1DrvPath::best_path(SB1Turn &cs, SB1Turn &ns, SB1Turn &as)
{
   SB1Line   *s, *f;
   SB1Circle *a;

   // find the best circle for Turn (ns)
   s = new SB1Line((ns.rad + ns.Sw), 0.0, 0.0, -1.0);
   a = new SB1Circle(ns.rad + ns.Aw);
   f = new SB1Line((ns.rad + ns.Fw) * cos(ns.len),
                  (ns.rad + ns.Fw) * sin(ns.len),
                  (ns.len + MY_PI12));
   ns.curv.tangent2(*s, *f, *a);

   // Update segements cs, as
   Update_cs(cs, ns);
   Update_as(as, ns);

   delete s; delete f;
   delete a;
}

// Adjusts path through corners (cs,ns). Assumes (ps) is on straight
// before 1st corner, (cs) is in the 1st corner, (ns) is in the 2nd
// corner, and (as) is on a straight after 2nd corner.
void SB1DrvPath::adjust_path(SB1Turn &ps, SB1Turn &cs, SB1Turn &ns, SB1Turn &as)
{
   double   c_wgt, dist;
   SB1Line   move;
   SB1Vector newC;

   if ( (cs.type == 0)||(ns.type == 0) )
      return;

   c_wgt = cs.rad / (cs.rad + ns.rad);
   dist  = cs.to_side(cs.len);
   dist -= ns.wid - ns.to_side(0.0);

   // move (cs) Turn
   move.pt.x  = cs.curv.c.x; move.pt.y  = cs.curv.c.y;
   move.dir.x = 1.0;         move.dir.y = 0.0;
   move.dir.rotate(cs.len + MY_PI);
   move.getPoint(newC, (c_wgt * dist));
   cs.curv.c.x = newC.x; cs.curv.c.y = newC.y;
   cs.Fl = cs.len;
   cs.Fw = cs.to_side(cs.Fl);

   // move (ns) Turn
   move.pt.x  = ns.curv.c.x; move.pt.y  = ns.curv.c.y;
   move.dir.x = -1.0;        move.dir.y = 0.0;
   move.getPoint(newC, ((1.0 - c_wgt) * dist));
   ns.curv.c.x = newC.x; ns.curv.c.y = newC.y;
   ns.Sl = 0.0;
   ns.Sw = ns.to_side(ns.Sl);

   // Update segements ps, as
   Update_cs(ps, cs);
   Update_as(as, ns);
}

// PRIVATE METHODS
void SB1DrvPath::construct_S()
{
   double      n_pct, n_len, p_len;
   SB1Turn     *n, *p;
   SB1Straight *s;

   pc[0]->num = 0;
   pc[1]->num = 1;
   pc[2]->num = 2;

   // Initialize speed parameters
   spd_cur = spd * crn_spd(0.0);

   if (pc[0]->valid())
      pc[0]->construct();
   if (pc[2]->valid())
      pc[2]->construct();

   pc[1]->Sl = pc[0]->Fl;
   pc[1]->Sw = pc[0]->Fw;
   pc[1]->Fl = pc[2]->Sl;
   pc[1]->Fw = pc[2]->Sw;

   if ( (pc[0]->valid())||(pc[2]->valid()) )
      pc[1]->construct();
   else {
      s = (SB1Straight *)pc[1];
      s->l.pt.x  = s->Sw;
      s->l.pt.y  = s->Sl;
      s->l.dir.x = s->Fw - s->l.pt.x;
      s->l.dir.y = s->Fl - s->l.pt.y;
      s->l.dir.normalize();
      return;
   }

   if ((pc[1]->Fl - pc[1]->Sl) > CARLEN)
      return;

   // fix overlap
   p = (SB1Turn *)pc[0];
   n = (SB1Turn *)pc[2];

   n_pct = p->curv.R + n->curv.R;
   if (n_pct != 0.0)
      n_pct = p->curv.R / n_pct;
   else
      n_pct = 0.5;

   // cut segement into two sections
   n_len = n_pct * pc[1]->len;
   p_len = pc[1]->len - n_len;
   if (n_len > (pc[2]->Fl - pc[2]->Sl)) {  // n_len too big
      n_len = pc[2]->Fl - pc[2]->Sl;
      p_len = pc[1]->len - n_len;
      if (p_len > (pc[0]->Fl - pc[0]->Sl))
         p_len = pc[0]->Fl - pc[0]->Sl;
   }
   if (p_len > (pc[0]->Fl - pc[0]->Sl)) {  // p_len too big
      p_len = pc[0]->Fl - pc[0]->Sl;
      n_len = pc[1]->len - p_len;
      if (n_len > (pc[2]->Fl - pc[2]->Sl))
         n_len = pc[2]->Fl - pc[2]->Sl;
   }

   pc[1]->Sl  = pc[1]->Fl = p_len;
   pc[1]->Fl += pc[1]->len - n_len - p_len;
   pc[1]->Sw  = n_pct       * pc[2]->to_side(pc[1]->Fl);
   pc[1]->Sw += (1.0-n_pct) * pc[0]->to_side(pc[1]->Sl);
   n_len = 1.5 * CARWID;
   if (pc[1]->Sw < n_len)
      pc[1]->Sw = n_len;
   else if (pc[1]->Sw > (pc[1]->wid-n_len))
      pc[1]->Sw = pc[1]->wid - n_len;
   pc[1]->Fw = pc[1]->Sw;
   pc[1]->construct();
		
   pc[0]->Fl = pc[1]->Sl;
   pc[0]->Fw = pc[1]->Sw;
   pc[2]->Sl = pc[1]->Fl;
   pc[2]->Sw = pc[1]->Fw;
}

void SB1DrvPath::construct_T()
{
   int     i;
   double  rad;
   SB1Turn *t;

   for (i = 0; i < num; i++) {
      pc[i]->num = i;
      pc[i]->construct();
      if ((i+1) < num) {
         pc[i+1]->Sl = pc[i]->Fl;
         pc[i+1]->Sw = pc[i]->Fw;
      }
      else {
         pc[i]->Fl = pc[i]->len;
         pc[i]->Fw = pc[i]->to_side(pc[i]->len);
      }
   }

   // Initialize speed parameters
   if (num == 1)
      t = (SB1Turn *)pc[0];
   else if (num == 3)
      t = (SB1Turn *)pc[1];
   else {
      t   = (SB1Turn *)pc[0];
      rad = t->curv.R;
      t   = (SB1Turn *)pc[1];
      if (rad < t->curv.R)
         t = (SB1Turn *)pc[0];
   }
   spd_cur  = crn_spd(t->curv.R);
   if (spd == SCALE_CORNSPD) {
      spd  += (0.1/100.0) * (150.0 - spd_cur);
      spd   = (spd < SCALE_CORNSPD) ? SCALE_CORNSPD : spd;
      spd   = (spd > (SCALE_CORNSPD+0.1)) ? (SCALE_CORNSPD+0.1) : spd;
   }
   spd_cur *= spd;
   apex     = t->curv.c.ang() - MY_PI;
}
	
void SB1DrvPath::Update_cs(SB1Turn &cs, SB1Turn &ns)
{
   double   dist;
   SB1Vector i;

   // place circle on segments cs
   cs.curv.R = ns.curv.R;

   if (ns.type == 1)  // left turn
      cs.curv.c.x = ns.curv.c.x - ns.rad;
   else               // right turn
      cs.curv.c.x = cs.wid + ns.rad - ns.curv.c.x;
   cs.curv.c.y = ns.curv.c.y + cs.len;


   // Turn in (cs.Sl, cs.Sw)
   ns.curv.getPoint(i, 0.0);
   cs.Sl = cs.len + i.y;
   if (ns.type == 1)  // left turn
      cs.Sw = i.x - ns.rad;
   else               // right turn
      cs.Sw = cs.wid - i.x + ns.rad;

   // Exit straight / Enter turn (cs.Fl, cs.Fw, ns.Sl, ns.Sw)
   if (ns.type == 1)  // left turn
      dist = cs.to_side(cs.len);
   else               // right turn
      dist = cs.wid - cs.to_side(cs.len);
   i.x = ns.rad + dist; i.y = 0.0;

   cs.Fl = cs.len + i.y;
   if (ns.type == 1)  // left turn
      cs.Fw = i.x - ns.rad;
   else               // right turn
      cs.Fw = cs.wid - i.x + ns.rad;

   ns.Sl = i.ang();
   ns.Sw = i.x - ns.rad;
}

void SB1DrvPath::Update_as(SB1Turn &as, SB1Turn &ns)
{
   double   dist;
   SB1Vector i;

   // place circle on segment as
   as.curv.R = ns.curv.R;

   as.curv.c.x = ns.curv.c.x;
   as.curv.c.y = ns.curv.c.y;
   as.curv.c.rotate(-ns.len);
   if (ns.type == 1)
      as.curv.c.x = as.curv.c.x - ns.rad;
   else
      as.curv.c.x = as.wid + ns.rad - as.curv.c.x;


   // Exit turn / Enter straight (ns.Fl, ns.Fw, as.Sl, as.Sw)
   if (ns.type == 1)  // left turn
      dist = as.to_side(0.0);
   else               // right turn
      dist = as.wid - as.to_side(0.0);
   i.x = cos(ns.len); i.y = sin(ns.len);
   i.scale(ns.rad + dist);

   ns.Fl = i.ang();
   i.rotate(-ns.len);
   ns.Fw = i.x - ns.rad;

   as.Sl = i.y;
   if (ns.type == 1)  // left turn
      as.Sw = i.x - ns.rad;
   else               // right turn
      as.Sw = as.wid - i.x + ns.rad;

   // Turn out (as.Fl, as.Fw)
   ns.curv.getPoint(i, ns.len);
   i.rotate(-ns.len);
   as.Fl = i.y;
   if (ns.type == 1)  // left turn
      as.Fw = i.x - ns.rad;
   else               // right turn
      as.Fw = as.wid - i.x + ns.rad;
}
//-----END DrvPath.cpp-----//

//-----START TrkData.cpp-----//
// CONSTRUCTORS
SB1TrkData::SB1TrkData()
{
   name = 0;
   path = 0;
   nseg = 0;
}

// DESTRUCTOR
SB1TrkData::~SB1TrkData()
{
   int i;

   if (path) {
      for (i = 0; i < nseg; i++) {
         delete path[i];
      }
      delete [] path;
      path = 0;
   }

   if (name) {
      free(name);
      name = 0;
   }
}

// METHODS
void SB1TrkData::set_path(Track *trk)
{
   int         i, j;
   int         ps, cs, ns, as;
   double      dw;
   SB1Turn      seg_cs, seg_as;
   SB1Turn     *tp, *ts, *tc, *tf;
   SB1Straight *ss;
   
   this->~SB1TrkData();  // clear out old data
   path = new SB1DrvPath*[trk->NSEG];
   if (path != 0)
      nseg = trk->NSEG;
   else {
      nseg = 0;
      return;
   }

   // store track name
   j    = strlen(trk->m_sFileName);
   name = (char *)malloc(j + 1);
   for (i = 0; i < j; i++) {
      name[i] = (char)tolower((int)trk->m_sFileName[i]);
   }
   name[j] = 0;

   // try to load data if available
   if (loadData(DATA_FILE) == 0)
      return;
   
   // Initialize
   seg_cs.set_seg_data(5000.0, trk->width, 0.0);
   seg_as.set_seg_data(5000.0, trk->width, 0.0);

   dw = trk->width - 50.0;
   dw = (dw < 10.0) ? 10.0 : dw;

   for (i = 0; i < nseg; i++) {
      if (trk->lftwall[i].radius > 0.0) {       // left turn
         path[i] = new SB1DrvPath( 1, 1);
         ts      = new SB1Turn();
         ts->set_seg_data(trk->lftwall[i].length, trk->width,
                          trk->lftwall[i].radius);
      }
      else if (trk->lftwall[i].radius < 0.0) {  // right turn
         path[i] = new SB1DrvPath(-1, 1);
         ts      = new SB1Turn();
         ts->set_seg_data(trk->rgtwall[i].length, trk->width,
                          trk->rgtwall[i].radius);
      }
      else {                                    // straight
         path[i] = new SB1DrvPath( 0, 3);
         ts      = new SB1Turn();
         ss      = new SB1Straight();
         tf      = new SB1Turn();
         ts->set_seg_data(trk->lftwall[i].length, trk->width, 0.0);
         ss->set_seg_data(trk->lftwall[i].length, trk->width, 0.0);
         tf->set_seg_data(trk->lftwall[i].length, trk->width, 0.0);
      }

      if (path[i]->seg_type != 0) {
         ts->Sw = trk->width - dw;
         ts->Sl = 0.0;
         ts->Aw = 5.0;
         ts->Fw = trk->width - dw;
         ts->Fl = ts->len;
         path[i]->pc[0] = ts;
      }
      else {
         ts->Sw = 0.0;
         ts->Sl = 0.0;
         ts->Aw = 0.0;
         ts->Fw = 0.0;
         ts->Fl = 0.0;
         ss->Sw = 0.0;
         ss->Sl = 0.0;
         ss->Fw = 0.0;
         ss->Fl = ss->len;
         tf->Sw = 0.0;
         tf->Sl = 0.0;
         tf->Aw = 0.0;
         tf->Fw = 0.0;
         tf->Fl = 0.0;
         path[i]->pc[0] = ts;
         path[i]->pc[1] = ss;
         path[i]->pc[2] = tf;
      }
   }

   for (j = 0; j < 2; j++) {
      // Initial run through track, to find preliminary values
      for (i = 0; i < nseg; i++) {
         cs = i;
         ns = (cs + 1) % trk->NSEG;
         as = (cs + 2) % trk->NSEG;
         ps = (cs + trk->NSEG - 1) % trk->NSEG;

         if (path[cs]->seg_type == 0) {  // straight
            if ( (path[ns]->seg_type != 0)&&(path[as]->seg_type == 0) ) {
               // straight, turn, straight
               ts = (SB1Turn *)path[cs]->pc[2];
               tc = (SB1Turn *)path[ns]->pc[0];
               tf = (SB1Turn *)path[as]->pc[0];
               SB1DrvPath::best_path(*ts, *tc, *tf);
            }
            else if ( (path[ns]->seg_type != 0)&&(path[as]->seg_type != 0) ) {
               // straight, turn, turn
               ts = (SB1Turn *)path[cs]->pc[2];
               tc = (SB1Turn *)path[ns]->pc[0];
               tf = (SB1Turn *)path[as]->pc[0];

               if ((tc->type - tf->type) == 0) {
                  // turns in same direction
                  tc->Aw = tc->Fw = 2.0 * CARWID;
                  tf->Aw = tf->Sw = 2.0 * CARWID;
               }
               else {
                  // turns in different directions
                  dw = tc->wid - (2.5 * CARWID);
                  if (tc->Sw > dw)
                     tc->Sw = dw;
                  tc->Fw = tc->Sw;

                  if (tf->Fw > dw)
                     tf->Fw = dw;
                  tf->Sw = tf->Fw;
               }
               
               SB1DrvPath::best_path(*ts, *tc, seg_as);
            }
         }
         else {                         // turn
            if ( (path[ns]->seg_type != 0)&&(path[as]->seg_type == 0) ) {
               // turn, turn, straight
               ts = (SB1Turn *)path[cs]->pc[0];
               tc = (SB1Turn *)path[ns]->pc[0];
               tf = (SB1Turn *)path[as]->pc[0];
               SB1DrvPath::best_path(seg_cs, *tc, *tf);

               if ((ts->type - tc->type) != 0) {
                  // turns in different directions
                  if (path[ps]->seg_type == 0) 
                     tp = (SB1Turn *)path[ps]->pc[2];
                  else
                     tp = &seg_cs;
                  SB1DrvPath::adjust_path(*tp, *ts, *tc, *tf);
               }
            }
            else if ( (path[ns]->seg_type != 0)&&(path[as]->seg_type != 0) ) {
               // turn, turn, turn
               ts = (SB1Turn *)path[cs]->pc[0];
               tc = (SB1Turn *)path[ns]->pc[0];
               tf = (SB1Turn *)path[as]->pc[0];

               if ((tc->type - tf->type) == 0) {
                  // turns in same direction
                  tc->Aw = tc->Fw = 2.0 * CARWID;
                  tf->Aw = tf->Sw = 2.0 * CARWID;
               }
               else {
                  // turns in different directions
                  dw = tc->wid - (2.5 * CARWID);
                  if (tc->Sw > dw)
                     tc->Sw = dw;
                  tc->Fw = tc->Sw;

                  if (tf->Fw > dw)
                     tf->Fw = dw;
                  tf->Sw = tf->Fw;
               }
               
               SB1DrvPath::best_path(seg_cs, *tc, seg_as);

               if ((ts->type - tc->type) != 0) {
                  // turns in different directions
                  if (path[ps]->seg_type == 0) 
                     tp = (SB1Turn *)path[ps]->pc[2];
                  else
                     tp = &seg_cs;
                  SB1DrvPath::adjust_path(*tp, *ts, *tc, seg_as);
               }
            }
         }
      }

      // Second run to fix straights
      for (i = 0; i < nseg; i++) {
         cs = i;
         ns = (cs + 1) % trk->NSEG;
         as = (cs + 2) % trk->NSEG;
         if ( (path[cs]->seg_type == 0)&&(path[ns]->seg_type == 0) ) {
            // straight, straight
            path[cs]->pc[2]->Sl = path[cs]->pc[0]->len;
            path[cs]->pc[2]->Sw = path[cs]->pc[0]->Fw;
            path[cs]->pc[2]->Fl = path[cs]->pc[0]->len;
            path[cs]->pc[2]->Fw = path[cs]->pc[0]->Fw;

            path[ns]->pc[0]->Sl = 0.0;
            path[ns]->pc[0]->Sw = path[cs]->pc[2]->Fw;
            path[ns]->pc[0]->Fl = 0.0;
            path[ns]->pc[0]->Fw = path[cs]->pc[2]->Fw;
         }
      }

      // Construct path
      for (i = 0; i < nseg; i++) {
         cs = i;
         ns = (cs + 1) % trk->NSEG;
         as = (cs + 2) % trk->NSEG;

         path[ns]->construct();

         if (j == 0) {
            dw = trk->width - 50.0;
            dw = (dw < 10.0) ? 10.0 : dw;

            if (path[cs]->seg_type != 0) {
               path[cs]->pc[0]->Fw = trk->width - dw;
               path[cs]->pc[0]->Fl = path[cs]->pc[0]->len;
            }
            if (path[as]->seg_type != 0) {
               path[as]->pc[0]->Sw = trk->width - dw;
               path[as]->pc[0]->Sl = 0.0;
            }

            if (path[ns]->seg_type == 0) {
               dw = path[cs]->pc[0]->Fw;
               if (path[cs]->seg_type == 1)
                  dw -= path[ns]->pc[0]->Fw;
               else if (path[cs]->seg_type == -1) 
                  dw -= trk->width - path[ns]->pc[0]->Fw;
               else
                  dw  = 0.0;
						
               if (fabs(dw) > (0.2*CARWID))
                  path[cs]->pc[0]->Fw -= dw;


               dw = path[as]->pc[0]->Sw;
               if (path[as]->seg_type == 1)
                  dw -= path[ns]->pc[2]->Sw;
               else if (path[as]->seg_type == -1)
                  dw -= trk->width - path[ns]->pc[2]->Sw;
               else
                  dw  = 0.0;

               if (fabs(dw) > (0.2*CARWID))
                  path[as]->pc[0]->Sw -= dw;
            }
         }
      }
   }
}

void SB1TrkData::saveData(const char *fname, const int ver)
{
   if (ver % 2)
      saveTXTData(fname);
   else
      saveBINData(fname);
}

int SB1TrkData::loadData(const char *fname)
{
   FILE *fp;
   char  buf[4];

   fp = fopen(fname, "r");
   if (fp == NULL)
      return -1;

   if (!fgets(buf, 4, fp)) {
      fclose(fp);
      return -1;
   }
   fclose(fp);
   buf[2] = 0;
   
   if (!strcmp(buf, "v1"))
      return loadTXTData(fname);
   else if (!strcmp(buf, "v2"))
      return loadBINData(fname);
   else
      return -1;
}


// PRIVATE METHODS
void SB1TrkData::saveTXTData(const char *fname)
{
   FILE       *fp;
   int         i, j, datalen;
   SB1Turn     *ts;
   SB1Straight *ss;

   fp = fopen(fname, "w");
   if (fp == NULL)
      return;

   datalen = 0;
   for (i = 0; i < nseg; i++) {
      if (path[i]->seg_type == 0)
         datalen += 13;
      else
         datalen +=  2 + (4*path[i]->num);
   }

   fprintf(fp, "v1\n");
   fprintf(fp, "TRACK: %s  LEN: %d [CARWID=%Lf][CARLEN=%Lf]\n",name,datalen,CARWID,CARLEN);
   for (i = 0; i < nseg; i++) {
      fprintf(fp, "<Seg: %3d, ", i);
      fprintf(fp, "type: %2d, ", path[i]->seg_type);
      fprintf(fp, "num: %1d>  ", path[i]->num);

      fprintf(fp, "<spd=%5.3Lf, spd_cur=%7.3Lf>\n", path[i]->spd, path[i]->spd_cur);
      
      for (j = 0; j < path[i]->num; j++) {
         fprintf(fp, "%1d", j);
         if ( ((j%2) == 0)||(path[i]->seg_type != 0) ) {
            ts = (SB1Turn *)path[i]->pc[j];
            saveTurn(fp, *ts);
         }
         else {
            ss = (SB1Straight *)path[i]->pc[j];
            saveStraight(fp, *ss);
         }
      }
      if ((i + 1) < nseg)
         fprintf(fp, "\n");
   }
	fclose(fp);
}

void SB1TrkData::saveTurn(FILE *fp, SB1Turn &t)
{
   int p = (t.type == 0) ? 3 : 6;
   
   fprintf(fp, "     TURN:");
   fprintf(fp, " type = %2d", t.type);
   fprintf(fp, " len = %8.*Lf", p, t.len);
   fprintf(fp, " wid = %5.1Lf", t.wid);
   fprintf(fp, " rad = %8.3Lf", t.rad);
   fprintf(fp, "\n");
		
   fprintf(fp, "  ");
   fprintf(fp, " Sl = %8.*Lf", p, t.Sl);
   fprintf(fp, " Sw = %5.1Lf", t.Sw);
   fprintf(fp, " Fl = %8.*Lf", p, t.Fl);
   fprintf(fp, " Fw = %5.1Lf", t.Fw);
   fprintf(fp, "\n");
		
   fprintf(fp, "  ");
   fprintf(fp, " Aw = %5.2Lf", t.Aw);
   fprintf(fp, "\n");

   fprintf(fp, "  ");
   fprintf(fp, " Circle{");
   fprintf(fp, "c.x=%9.3Lf",   t.curv.c.x);
   fprintf(fp, ", c.y=%9.3Lf", t.curv.c.y);
   fprintf(fp, ", R=%8.3Lf",   t.curv.R);		
   fprintf(fp, "}\n");
}

void SB1TrkData::saveStraight(FILE *fp, SB1Straight &t)
{
   fprintf(fp, " STRAIGHT:");
   fprintf(fp, " type = %2d", t.type);
   fprintf(fp, " len = %8.3Lf", t.len);
   fprintf(fp, " wid = %5.1Lf", t.wid);
   fprintf(fp, " rad = %3.1Lf", t.rad);
   fprintf(fp, "\n");
   
   fprintf(fp, "  ");
   fprintf(fp, " Sl = %8.3Lf", t.Sl);
   fprintf(fp, " Sw = %5.1Lf", t.Sw);
   fprintf(fp, " Fl = %8.3Lf", t.Fl);
   fprintf(fp, " Fw = %5.1Lf", t.Fw);
   fprintf(fp, "\n");
		
   fprintf(fp, "  ");
   fprintf(fp, " Line{");
   fprintf(fp, "pt.x=%9.3Lf",    t.l.pt.x);
   fprintf(fp, ", pt.y=%9.3Lf",  t.l.pt.y);
   fprintf(fp, ", dir.x=%6.3Lf", t.l.dir.x);
   fprintf(fp, ", dir.y=%6.3Lf", t.l.dir.y);
   fprintf(fp, "}\n");
}

int SB1TrkData::loadTXTData(const char *fname)
{
   char        buf[128], *tok;
   int         i, j, type, num_comp, datalen, straight;
   double      len, wid, rad, Sw, Sl, Fw, Fl;
   SB1Turn     *ts;
   SB1Straight *ss;
   char       *sep = " \n\t,=<>{}:";
   FILE       *fp = fopen(fname, "r");

   // skip first line
   if (!fgets(buf, 128, fp)) {
      fclose(fp);
      return -1;
   }

   // look for track
   while (1) {
      if (!fgets(buf, 128, fp)) {
         fclose(fp);
         return -1;
      }
      tok = strtok(buf, sep);    // TRACK:
      tok = strtok(NULL, sep);   // Track name
      if (!tok) {
         fclose(fp);
         return -1;
      }
      if (!strcmp(name, tok))
         break;   // track found

      // skip this track
      tok     = strtok(NULL, sep);        // LEN:
      datalen = atoi(strtok(NULL, sep));  // data length
      for (i = 0; i < datalen; i++) {
         if (!fgets(buf, 128, fp)) {
            fclose(fp);
            return -1;
         }
      }
   }

   for (i = 0; i < nseg; i++) {
      if (!fgets(buf, 128, fp)) {
         fclose(fp);
         return -1;
      }
      tok          = strtok(buf, sep);          // Seg:
      tok          = strtok(NULL, sep);         // Seg value
      tok          = strtok(NULL, sep);         // type:
      type         = atoi(strtok(NULL, sep));   // type value
      tok          = strtok(NULL, sep);         // num:
      num_comp     = atoi(strtok(NULL, sep));   // number of components
      path[i]      = new SB1DrvPath(type, num_comp);
      tok          = strtok(NULL, sep);         // spd=
      path[i]->spd = atof(strtok(NULL, sep));   // spd value

      for (j = 0; j < num_comp; j++) {
         if (!fgets(buf, 128, fp)) {
            fclose(fp);
            return -1;
         }
         tok = strtok(buf, sep);          // Comp_num
         tok = strtok(NULL, sep);         // Comp_name
         if (!strcmp(tok, "STRAIGHT"))
            straight = 1;
         else if (!strcmp(tok, "TURN"))
            straight = 0;
         else {
            fclose(fp);
            return -1;
         }
         tok = strtok(NULL, sep);         // Type
         tok = strtok(NULL, sep);         // Type_val
         tok = strtok(NULL, sep);         // Len
         len = atof(strtok(NULL, sep));   // Len_val
         tok = strtok(NULL, sep);         // Wid
         wid = atof(strtok(NULL, sep));   // Wid_val
         tok = strtok(NULL, sep);         // Rad
         rad = atof(strtok(NULL, sep));   // Rad_val

         if (!fgets(buf, 128, fp)) {
            fclose(fp);
            return -1;
         }
         tok = strtok(buf, sep);          // Sl
         Sl  = atof(strtok(NULL, sep));   // Sl_val
         tok = strtok(NULL, sep);         // Sw
         Sw  = atof(strtok(NULL, sep));   // Sw_val
         tok = strtok(NULL, sep);         // Fl
         Fl  = atof(strtok(NULL, sep));   // Fl_val
         tok = strtok(NULL, sep);         // Fw
         Fw  = atof(strtok(NULL, sep));   // Fw_val
					
         if (straight) {
            ss = new SB1Straight();
            ss->set_seg_data(len, wid, rad);
            ss->Sl = Sl; ss->Sw = Sw;
            ss->Fl = Fl; ss->Fw = Fw;

            if (!fgets(buf, 128, fp)) {
               fclose(fp);
               return -1;
            }
            tok         = strtok(buf, sep);        // Line
            tok         = strtok(NULL, sep);       // pt.x
            ss->l.pt.x  = atof(strtok(NULL, sep)); // pt.x_val
            tok         = strtok(NULL, sep);       // pt.y
            ss->l.pt.y  = atof(strtok(NULL, sep)); // pt.y_val
            tok         = strtok(NULL, sep);       // dir.x
            ss->l.dir.x = atof(strtok(NULL, sep)); // dir.x_val
            tok         = strtok(NULL, sep);       // dir.y
            ss->l.dir.y = atof(strtok(NULL, sep)); // dir.y_val
            path[i]->pc[j] = ss;
         }
         else {
            ts = new SB1Turn();
            ts->set_seg_data(len, wid, rad);
            ts->Sl = Sl; ts->Sw = Sw;
            ts->Fl = Fl; ts->Fw = Fw;
				ts->type = type;

            if (!fgets(buf, 128, fp)) {
               fclose(fp);
               return -1;
            }
            tok    = strtok(buf, sep);          // Aw
            ts->Aw = atof(strtok(NULL, sep));   // Aw_val

            if (!fgets(buf, 128, fp)) {
               fclose(fp);
               return -1;
            }
            tok          = strtok(buf, sep);          // Circle
            tok          = strtok(NULL, sep);         // c.x
            ts->curv.c.x = atof(strtok(NULL, sep));   // c.x_val
            tok          = strtok(NULL, sep);         // c.y
            ts->curv.c.y = atof(strtok(NULL, sep));   // c.y_val
            tok          = strtok(NULL, sep);         // R
            ts->curv.R   = atof(strtok(NULL, sep));   // R_val
            path[i]->pc[j] = ts;
         }
      }
      if ( ((i + 1) < nseg)&&(!fgets(buf, 128, fp)) ) {
         fclose(fp);
         return -1;
      }
      path[i]->construct();
   }
   fclose(fp);
   return 0;
}

void SB1TrkData::saveBINData(const char *fname)
{
}

int SB1TrkData::loadBINData(const char *fname)
{
   return -1;
}
//-----END TrkData.cpp-----//

//-----START SBDriver.cpp-----//
// CONSTRUCTORS
SBv1r4::SBv1r4()
{
   m_sName         = SB1_NAME1;
   m_sAuthor       = SB1_AUTHOR;
   m_sDescription  = SB1_DESCRIPTION;
   m_iNoseColor    = SB1_NOSE;
   m_iTailColor    = SB1_TAIL;
   m_sBitmapName2D = SB1_BITMAP2D;
   m_sModel3D      = SB1_MODEL3D;

   name      = NULL;
   init      = 2;
   cp        = &result;
   lastpit   = 0;
   fuelstart = MAX_FUEL;
   err       = 0.0;
   passTime  = 0;
#ifdef SB_TESTING
   SBnum     = 0;
#endif
}

SBv1r4::SBv1r4(int i)
{
   name = new char[SB1_NAME_LEN];
   sprintf(name, SB1_NAME2, (i%100));
   
   m_sName         = name;
   m_sAuthor       = SB1_AUTHOR;
   m_sDescription  = SB1_DESCRIPTION;
   m_iNoseColor    = SB1_NOSE;
   m_iTailColor    = SB1_TAIL;
   m_sBitmapName2D = SB1_BITMAP2D;
   m_sModel3D      = SB1_MODEL3D;

   init      = 2;
   cp        = &result;
   lastpit   = 0;
   fuelstart = MAX_FUEL;
   err       = 0.0;
   passTime  = 0;
#ifdef SB_TESTING
   SBnum     = i;
#endif
}


// DESTRUCTOR
SBv1r4::~SBv1r4()
{
   if (name)
      delete [] name;
}


// METHODS
con_vec SBv1r4::drive(situation &s)
{
   if ( (args.m_iCurrentTrack != trk_num)&&(currentTrack) ) {
      // set surface type
      MAX_ACCEL = g;
      switch (args.m_iSurface) {
      case 0:
         MAX_ACCEL *= MYU_MAX0;
         break;
      case 1:
         MAX_ACCEL *= MYU_MAX1;
         break;
      case 2:
         MAX_ACCEL *= MYU_MAX2;
      }
      trk_data.set_path(currentTrack);
      trk_num = args.m_iCurrentTrack;
   }

   if (s.starting) {
      init               = (init < 1) ? 1 : init;
      result.request_pit = 0;
      result.fuel_amount = fuel_needed(s.stage, (s.laps_to_go+5), 
                                       (SB1_FC*currentTrack->length));
      lastpit            = 0;
      fuelstart          = result.fuel_amount;
   }

   if (init == 2) {
      init--;
      result.vc    = 400.0;
      result.alpha =   0.0;
      return result;
   }
   else if (init == 1) {
      if ( ((!s.seg_ID)&&((s.cur_len - s.to_end) >= trk_data.path[0]->pc[1]->Fl))||
           (s.v > 60.0) )
         init--;
   }

   get_RARS_ctrl_params(s);

   if (init == 1) {
      result.alpha *= 0.25;
      err           = 0.0;
   }
   return result;
}


// PRIVATE METHODS
void SBv1r4::get_RARS_ctrl_params(situation &s)
{
   SB1Line           cV;
   SB1Vector         nV;
   SB1PathComponent *pc, *nex_pc;
   double           acc, tgt_v, tv, lane, bias, my_dist;
   double           pos = s.cur_len - s.to_end;
   int              sID2, dir;
   int              brk_hard = 0, passing = 0;


   // Handle fuel / damage
   cp->repair_amount = (s.damage > 18000) ? 18000 : s.damage;
   if ( (s.out_pits)&&(cp->request_pit) ) {
      cp->request_pit = 0;
      lastpit         = s.laps_done;
      fuelstart       = s.fuel;
      passTime        = 5 * SB1_PASS_COUNT;
   }
   if ( (s.lap_flag)&&(s.laps_to_go > 1)&&(s.laps_done > lastpit) ) {
      cp->fuel_amount = (fuelstart - s.fuel)/(double)(s.laps_done - lastpit);
      if ( (s.fuel < (cp->fuel_amount*2.25))||
           ((s.laps_to_go > 4)&&(s.damage > 25000)) ) {
         cp->fuel_amount = fuel_needed(s.stage, (s.laps_to_go+2), cp->fuel_amount);
         cp->request_pit = 1;
      }
   }


   // Get new control parameters
   // get current and next path components
   sID2 = s.seg_ID;
   pc   = trk_data.path[sID2]->get_PC(pos);
   if (pc->type) {
      acc = 0.5 * s.v * delta_time / (pc->rad + pc->wid);
      acc = (acc < 0.0349) ? 0.0349 : acc;
   }
   else {
      acc = 0.5 * s.v * delta_time;
      acc = (acc < 2.0) ? 2.0 : acc;
   }
   acc += pc->Fl;
   if (acc < s.cur_len)
      nex_pc = trk_data.path[sID2]->get_PC(acc);
   else {
      acc -= s.cur_len;
      if (pc->type)
         acc *= pc->rad;
      if (s.nex_rad != 0.0)
         acc /= fabs(s.nex_rad);
      nex_pc = trk_data.path[(sID2+1)%trk_data.nseg]->get_PC(acc);
   }

   // find current position and velocity in current segment
   if (pc->type >= 0)
      get_pos(*pc, pos, s.to_lft, cV.pt);
   else
      get_pos(*pc, pos, s.to_rgt, cV.pt);
   get_vel(*pc, s.v, s.vn, pos, cV.dir);

   // target velocity
   tgt_v = tv = get_tgt_v(s);

   // find bias and lane for alt_update_con_vec()
   if (!pc->type) {
      if (pc == trk_data.path[sID2]->pc[0])
         sID2 = (sID2+trk_data.nseg-1) % trk_data.nseg;
      else if (pc == trk_data.path[sID2]->pc[2])
         sID2 = (sID2+1) % trk_data.nseg;
   }
   acc  = (double)trk_data.path[sID2]->pc[0]->type;
   acc *= trk_data.path[sID2]->pc[0]->len;
   if (acc == 0.0)
      bias = 0.0;
   else {
      if (fabs(tgt_v) > 10.273)
         bias = atan(16.0 / fabs(tgt_v));
      else
         bias = 1.0;

      acc /= MY_PI;
      if (acc < 1.0)
         bias *= (acc < -1.0) ? -1.0 : acc;

      bias *= (s.v*s.v) / (tgt_v*tgt_v);

      if ( (s.to_lft < SB1_OUT_WID)&&(bias > 0.0) )
         bias *= s.to_lft / SB1_OUT_WID;
      else if ( (s.to_rgt < SB1_OUT_WID)&&(bias < 0.0) )
         bias *= s.to_rgt / SB1_OUT_WID;

      if (bias > 1.0)
         bias =  1.0;
      else if (bias < -1.0)
         bias = -1.0;
   }
   lane = pc->to_side(pos);
   lane = (pc->type < 0) ? (pc->wid - lane) : lane;

   // safety measures
   if ( (s.to_lft < 0.0)||(s.to_rgt < 0.0) ) {
      // car is off track
      alt_update_con_vec(s, lane, bias, 1);
      cp->vc = 400.0;
      return;
   }
   else if (passTime > SB1_PASS_COUNT) {
      // car is leaving pits
      alt_update_con_vec(s, lane, bias, 0);
      cp->vc = fabs(tgt_v * cos(cp->alpha));
      passTime--;
      passTime = (passTime == SB1_PASS_COUNT) ? 0 : passTime;
      return;
   }

   // adjust target velocity
   tgt_v = fabs(tgt_v);
   acc   = (tgt_v - s.v) / delta_time;
   if (acc > MAX_ACCEL)
      tgt_v = s.v + (1.1 * MAX_ACCEL * delta_time);
   else if (acc < -MAX_ACCEL) {
      tgt_v = s.v - (1.1 * MAX_ACCEL * delta_time);
      brk_hard = 1;
   }
   if (tgt_v < 5.0)
      tgt_v = 5.0;
   nV.x = tgt_v; nV.y = 0.0;

   // find target position
   err = pc->get_nex_pos(cV, nV, *nex_pc);
   if (pc->type == -1) {
      cV.pt.x  = -cV.pt.x;
      cV.dir.x = -cV.dir.x;
      nV.x     = -nV.x;
   }
   nV.scale(1.0 / delta_time);

   // find out how close the other cars are
   my_dist = locate_cars(cV, s);

   // avoid other cars
   if ( (!pc->type)&&(pc->num == 2)&&(nex_pc->type) )
      dir = (int)ceil(pc->Sw - pc->Fw);
   else
      dir = 0;
   passing = anti_collision(s, nV, my_dist, dir);

   // get new control parameters
   update_con_vec(cV.dir, nV);
   nV.x = cp->alpha;
   nV.y = cp->vc;
   if (err <= 0.0) {
      alt_update_con_vec(s, lane, bias, 0);
      cp->alpha *= -0.25*err;
      cp->alpha +=  (1.0 + (0.25*err)) * nV.x;
      cp->vc     = nV.y;
      if ( (s.to_lft >= SB1_IN_WID)&&(s.to_rgt >= SB1_IN_WID)&&
           (tv >= 200.0) ) {
         if ( (!brk_hard)&&(!passing) ) {
            if (pc->type == 0)
               cp->vc *= 2.5;
            else if ( (s.cur_len < MY_PI14)&&(s.nex_rad == 0.0) )
               cp->vc *= 2.5;
         }

         if ( (brk_hard)||(passTime) ) {
            cp->alpha *= 0.8;
            cp->alpha += 0.2 * s.alpha;
         }
         else if ( ((s.to_lft > 25.0)||(s.vn <  5.0))&&
                   ((s.to_rgt > 25.0)||(s.vn > -5.0)) ) {
            cp->alpha *= 0.35;
            cp->alpha += 0.65 * s.alpha;
         }
      }
      else {
         if ( ((s.to_lft > 10.0)||(s.vn <  5.0))&&
              ((s.to_rgt > 10.0)||(s.vn > -5.0)) ) {
            cp->alpha *= 0.9;
            cp->alpha += 0.1 * s.alpha;
         }
      }
   }
   else {
      alt_update_con_vec(s, lane, bias, 1);
      cp->alpha *= 0.33*err;
      cp->alpha += (1.0 - (0.33*err)) * nV.x;
      cp->vc    *= 0.07*err;
      cp->vc    += (1.0 - (0.07*err)) * nV.y;
   }
   
   err *= (err < 0.0) ? -0.6 : 1.175;
   err  = (err > 1.0) ?  1.0 : err;
}

double SBv1r4::fuel_needed(Stage s, int lapsleft, double fpl)
{
   double retval;

   if (fpl <= 0.0)
      return MAX_FUEL;
   else if (s == PRACTICE)
      retval = fpl * 10.0;
   else if (s == QUALIFYING)
      retval = fpl * 5.0;
   else
      retval = fpl * (double)lapsleft;

   if (retval > MAX_FUEL)
      return MAX_FUEL;
   return retval;
}

// Returns current mass of the car
double SBv1r4::car_mass(double fuel)
{
   return M + (fuel / 32.2);
}

// friction function
double SBv1r4::u(double slip)
{
   double retval;

   switch (args.m_iSurface) {
   case 2:
      retval = MYU_MAX2 * (1.0 - exp(-slip/SLIPPING));
      if (slip > 20.0)
         retval -= 0.2;
      break;
   case 1:
      retval = MYU_MAX1 * (1.0 - exp(-slip/SLIPPING));
      break;
   default:
      retval = MYU_MAX0 * (slip / (SLIPPING + slip));
   }
   return retval;
}

// inverse friction function
double SBv1r4::iu(double cur_u)
{
   double retval;

   switch (args.m_iSurface) {
   case 2:
      if (cur_u >= MYU_MAX2)
         retval = 20.0;
      else
         retval = -SLIPPING * log(1.0 - (cur_u/MYU_MAX2));
      break;
   case 1:
      if (cur_u >= MYU_MAX1)
         retval = 20.0;
      else
         retval = -SLIPPING * log(1.0 - (cur_u/MYU_MAX1));
      break;
   default:
      if (cur_u >= MYU_MAX0)
         retval = 20.0;
      else
         retval = ((SLIPPING*cur_u)/MYU_MAX0) / (1.0 - (cur_u/MYU_MAX0));
   }
   return retval;
}

void SBv1r4::update_con_vec(SB1Vector &cV, SB1Vector &nV)
{
   SB1Vector dV, L, W;
   double   acc;

   acc = -cV.ang();
   nV.rotate(acc);
   cV.rotate(acc);
   cV.y = 0.0;

   dV.add(nV);
   dV.sub(cV);

   L.add(dV);
   L.normalize();
   acc = (dV.mag() / delta_time) / 32.2;
   L.scale(-iu(acc));

   W.add(L);
   W.sub(cV);

   cp->vc    = W.mag();
   cp->alpha = W.ang() - MY_PI;
}

void SBv1r4::get_pos(SB1PathComponent &pc, double Pl, double Pw, 
                      SB1Vector &pos)
{
   if (pc.type == 0) {
      pos.x = Pw;
      pos.y = Pl;
   }
   else {
      pos.x  = (pc.rad + Pw)*cos(Pl);
      pos.y  = (pc.rad + Pw)*sin(Pl);
   }
}

void SBv1r4::get_vel(SB1PathComponent &pc, double v, double vn, double Pl, 
                      SB1Vector &vel)
{
   if (v < (fabs(vn)+1.0))
      v = fabs(vn) + 1.0;

   if (pc.type >= 0)
      vel.x = -vn;
   else
      vel.x =  vn;
   vel.y = sqrt((v*v) - (vn*vn));
   if (pc.type != 0)
      vel.rotate(Pl);
}

double SBv1r4::get_tgt_v(situation &s)
{
   double cur_spd, nex_spd, tst_spd, dist2seg, brk_dist, end_seg, wid;
   int    i, sID[5];

   // segement IDs
   sID[0] = s.seg_ID;
   for (i = 1; i < 5; i++) {
      sID[i] = (sID[0] + i) % trk_data.nseg;
   }

   wid      = s.to_lft + s.to_rgt;
   dist2seg = ((s.cur_rad == 0.0) ? -1.5 : -2.0) * s.v * delta_time;
   cur_spd  = trk_data.path[sID[0]]->spd_cur;
   end_seg  = -(s.cur_len - s.to_end);
   if (trk_data.path[sID[0]]->seg_type != 0) {
      cur_spd *= 1.0 - err*(1.0 - 0.6/trk_data.path[sID[0]]->spd);
      if ( (err < 0.225)&&(-end_seg > trk_data.path[sID[0]]->apex) )
         cur_spd *= 1.07 - 0.07*(s.to_end/(s.cur_len - trk_data.path[sID[0]]->apex));
      end_seg = crn_RAD2FT(end_seg, s.cur_rad, wid);
   }
   nex_spd = cur_spd;

   // check the next 4 segments
   for (i = 1; i < 5; i++) {
      if (trk_data.nseg <= i)
         break;

      dist2seg += end_seg;
      if ( (trk_data.path[sID[i-1]]->seg_type != 0)&&
           (trk_data.path[sID[i-1]]->num == 1) ) {
         dist2seg += crn_RAD2FT(trk_data.path[sID[i-1]]->pc[0]->len, 
                                trk_data.path[sID[i-1]]->pc[0]->rad, wid);
         brk_dist  = 1.1;
         end_seg   = 0.0;
      }
      else if ( (trk_data.path[sID[i-1]]->seg_type != 0)&&
                (trk_data.path[sID[i-1]]->num == 2) ) {
         dist2seg += crn_RAD2FT(trk_data.path[sID[i-1]]->pc[0]->Fl, 
                                trk_data.path[sID[i-1]]->pc[0]->rad, wid);
         brk_dist  = 1.1;
         end_seg   = trk_data.path[sID[i-1]]->pc[1]->len - trk_data.path[sID[i-1]]->pc[1]->Sl;
         end_seg   = crn_RAD2FT(end_seg, trk_data.path[sID[i-1]]->pc[1]->rad, wid);
      }
      else if (trk_data.path[sID[i-1]]->seg_type != 0) {
         dist2seg += crn_RAD2FT(trk_data.path[sID[i-1]]->pc[1]->Fl, 
                                trk_data.path[sID[i-1]]->pc[1]->rad, wid);
         brk_dist  = 1.1;
         end_seg   = trk_data.path[sID[i-1]]->pc[2]->len - trk_data.path[sID[i-1]]->pc[2]->Sl;
         end_seg   = crn_RAD2FT(end_seg, trk_data.path[sID[i-1]]->pc[2]->rad, wid);
      }
      else {
         dist2seg += trk_data.path[sID[i-1]]->pc[1]->Fl;
         brk_dist  = 1.0;
         end_seg   = trk_data.path[sID[i-1]]->pc[2]->Fl - trk_data.path[sID[i-1]]->pc[2]->Sl;
         if (end_seg < 0.0)
            end_seg = 0.0;
      }

      tst_spd = trk_data.path[sID[i]]->spd_cur;
      if (trk_data.path[sID[i]]->seg_type)
         tst_spd *= 1.0 - err*(1.0 - 0.6/trk_data.path[sID[i]]->spd);
      brk_dist *= brkdist(s.v, tst_spd);
      if ( (brk_dist >= dist2seg)&&(nex_spd > tst_spd) )
         nex_spd = tst_spd;
   }

   if (nex_spd < cur_spd) {
      // some braking might be needed
      if (trk_data.path[sID[0]]->seg_type != 0)
         nex_spd = -nex_spd;
   }
   return nex_spd;
}

double SBv1r4::get_dist(int seg_ID, double ahead, SB1Vector &pos)
{
   SB1Vector rel_pos(pos);
   int      i, sID, next;
   double   seg_len;

   for (i = 0; i < trk_data.nseg; i++) {
      // find correct segment
      sID  = (i + seg_ID) % trk_data.nseg;
      next = (sID + 1) % trk_data.nseg;
      if (trk_data.path[sID]->seg_type == 0) {
         // straight
         seg_len = trk_data.path[sID]->pc[0]->len;

         if (ahead <= seg_len) {
            // found segment, now find correct path component
            return (trk_data.path[sID]->get_PC(ahead))->dist(rel_pos);
         }
         else {
            // make rel_pos relative to next segment
            ahead     -= seg_len;
            rel_pos.y -= seg_len;
            if (trk_data.path[next]->seg_type == 1)
               rel_pos.x += trk_data.path[next]->pc[0]->rad;
            else if (trk_data.path[next]->seg_type == -1)
               rel_pos.x -= trk_data.path[next]->pc[0]->rad + trk_data.path[sID]->pc[0]->wid;
         }
      }
      else {
         // turn
         seg_len = crn_RAD2FT(trk_data.path[sID]->pc[0]->len,
                              trk_data.path[sID]->pc[0]->rad,
                              trk_data.path[sID]->pc[0]->wid);

         if (trk_data.path[sID]->seg_type == -1)
            rel_pos.x = -rel_pos.x;

         if (ahead <= seg_len) {
            // found segment, now find correct path component
            ahead = crn_FT2RAD(ahead,
                               trk_data.path[sID]->pc[0]->rad,
                               trk_data.path[sID]->pc[0]->wid);
            if (trk_data.path[sID]->seg_type == 1)
               return  (trk_data.path[sID]->get_PC(ahead))->dist(rel_pos);
            else
               return -(trk_data.path[sID]->get_PC(ahead))->dist(rel_pos);
         }
         else {
            // make rel_pos relative to next segment
            ahead     -= seg_len;
            rel_pos.x -= cos(trk_data.path[sID]->pc[0]->len) * trk_data.path[sID]->pc[0]->rad;
            rel_pos.y -= sin(trk_data.path[sID]->pc[0]->len) * trk_data.path[sID]->pc[0]->rad;
            rel_pos.rotate(-trk_data.path[sID]->pc[0]->len);
            rel_pos.x += trk_data.path[next]->pc[0]->rad;
         }

         if (trk_data.path[sID]->seg_type == -1)
            rel_pos.x = -rel_pos.x;
      }
   }
   return 100000.0;
}

double SBv1r4::locate_cars(SB1Line &cV, situation &s)
{
   int    i;
   SB1Line other_car;
   double dist, my_dist, my_pos;

   my_pos = s.cur_len - s.to_end;
   if (s.cur_rad != 0.0)
      my_pos = crn_RAD2FT(my_pos, fabs(s.cur_rad), (s.to_lft+s.to_rgt));
   my_dist = get_dist(s.seg_ID, my_pos, cV.pt);

   for (i = 0; i < NEARBY_CARS; i++) {
      if (s.nearby[i].who >= MAX_CARS) {
         nearby[i] = 100000.0;
         continue;
      }

      // start at the current pos
      other_car.pt.x  = cV.pt.x;  other_car.pt.y  = cV.pt.y;
      other_car.dir.x = cV.dir.x; other_car.dir.y = cV.dir.y;
      other_car.dir.normalize();

      // now find pos of other car
      other_car.movePoint(s.nearby[i].rel_y);   // in front
      other_car.dir.rotate(-MY_PI12);
      other_car.movePoint(s.nearby[i].rel_x);   // to right
      other_car.dir.rotate( MY_PI12);

      // other_car's current dist from path
      nearby[i]  = get_dist(s.seg_ID, (my_pos+s.nearby[i].dist), 
                            other_car.pt);
      nearby[i] -= my_dist;

      if (fabs(nearby[i]) <= CARWID) {
         // advance position
         other_car.movePoint(s.nearby[i].rel_ydot*SB1_DT);   // in front
         other_car.dir.rotate(-MY_PI12);
         other_car.movePoint(s.nearby[i].rel_xdot*SB1_DT);   // to right
         other_car.dir.rotate( MY_PI12);

         // see if other_car will get closer to the path in the next SB1_DT sec
         dist  = s.nearby[i].dist + ((s.nearby[i].v-s.v)*SB1_DT);
         dist  = get_dist(s.seg_ID, (my_pos+dist), other_car.pt);
         dist -= my_dist;

         if (nearby[i] > 10000.0)
            nearby[i] = dist;
         else if (nearby[i] > 0.0) {
            if (dist <= 0.0)
               nearby[i] = 0.0;
            else if (dist < nearby[i])
               nearby[i] = dist;
         }
         else if (dist < 10000.0) {
            if (dist >= 0.0)
               nearby[i] = 0.0;
            else if (dist > nearby[i])
               nearby[i] = dist;
         }
      }
   }
   return my_dist;
}

int SBv1r4::anti_collision(situation &s, SB1Vector &nV, double my_dist,
                            int dir)
{
   double dv, min_v, tgt_v, look_ahead, time_ahead;
   double min_d, max_d;
   double rel_x, rel_xdot, sn, cs;
   int    j, braking, in_the_way;

   sn    = sin(s.alpha);
   cs    = cos(s.alpha);
   tgt_v = nV.mag();
   min_v = tgt_v;
   for (j = 0; j < NEARBY_CARS; j++) {
      if ( (s.nearby[j].who >= MAX_CARS)||(nearby[j] > 10000.0)||
           (s.nearby[j].dist < (0.3*CARLEN)) )
         continue;

      min_d  = s.nearby[j].rel_x * s.nearby[j].rel_x;
      min_d += s.nearby[j].rel_y * s.nearby[j].rel_y;
      if (sqrt(min_d) < (CARLEN-0.5))
         continue;

      dv = s.nearby[j].v - tgt_v;
      if (dv < 0.0)
         time_ahead = -s.nearby[j].dist / dv;
      else if (dv < 5.0)
         time_ahead = 0.0;
      else 
         time_ahead = 100000.0;
      if (time_ahead > SB1_SAFE_TIME)
         continue;

      // check if car is too close to the path
      in_the_way = 0;
      if ( (dv < 1.0)||(s.nearby[j].dist < 80.0) ) {
         if (my_dist > 1.0) {
            min_d = -4.0 - my_dist;
            max_d =  5.0;
         }
         else if (my_dist < -1.0) {
            min_d = -5.0;
            max_d =  4.0 - my_dist;
         }
         else {
            min_d = -5.0;
            max_d =  5.0;
         }
         min_d -= CARWID;
         max_d += CARWID;

         if (s.nearby[j].dist <= (s.to_end+CARLEN)) {
            if (dir > 0)
               min_d -= (double)dir;
            else if (dir < 0)
               max_d -= (double)dir;
         }

         if (min_d < (-CARLEN-s.to_lft))
            min_d = -CARLEN - s.to_lft;
         if (max_d > ( CARLEN+s.to_rgt))
            max_d =  CARLEN + s.to_rgt;

         in_the_way = (nearby[j] >= min_d)&&(nearby[j] <= max_d);
      }

      if (!in_the_way) {
         // check if car is in front
         look_ahead  = (dv > -5.0) ? 80.0 : (-149.5*dv*SB1_DT);
         look_ahead -= s.nearby[j].dist + (SB1_DT * dv);
         dv          = s.nearby[j].rel_x;
         in_the_way  = (look_ahead >= 0.0)&&(dv >= min_d)&&(dv <= max_d);
         if ( (!in_the_way)&&(look_ahead >= 0.0)&&
              (s.nearby[j].rel_xdot != 0.0) ) {
            dv        -= (dv > 0.0) ? max_d : min_d;
            dv        /= -s.nearby[j].rel_xdot;
            in_the_way = (dv >= 0.0)&&(dv <= 3.5);
         }
      }
      if ( (!in_the_way)&&(look_ahead >= 0.0) ) {
         // check if car is in front
         rel_x      = (s.nearby[j].rel_x*cs)-(s.nearby[j].rel_y*sn);
         rel_xdot   = (s.nearby[j].rel_xdot*cs)-(s.nearby[j].rel_ydot*sn);
         in_the_way = (rel_x >= min_d)&&(rel_x <= max_d);
         if ( (!in_the_way)&&(rel_xdot != 0.0) ) {
            rel_x     -= (rel_x > 0.0) ? max_d : min_d;
            rel_x     /= -rel_xdot;
            in_the_way = (rel_x >= 0.0)&&(rel_x <= 3.5);
         }
      }

      if (in_the_way) {
         if (s.nearby[j].v < min_v) {
            if ( (s.nearby[j].to_lft >= 0.0)&&
                 (s.nearby[j].to_rgt >= 0.0) )
               min_v  = s.nearby[j].v;
            else if (s.nearby[j].to_lft < 0.0) {
               min_v *= s.nearby[j].to_lft / -30.0;
               min_v += ((-30.0+s.nearby[j].to_lft)/-30.0)*s.nearby[j].v;
            }
            else {
               min_v *= s.nearby[j].to_rgt / -30.0;
               min_v += ((-30.0+s.nearby[j].to_rgt)/-30.0)*s.nearby[j].v;
            }
            braking  = s.nearby[j].braking;
            braking |= s.nearby[j].to_lft < 0.0;
            braking |= s.nearby[j].to_rgt < 0.0;
         }
      }
   }

   if (min_v < tgt_v) {
      // must slow down
      if ( (braking)||((tgt_v - s.v) < 1.5) )
         min_v *= 0.95;
      else
         min_v *= ((s.v - min_v) > 5.0) ? 0.95 : 1.0;

      dv = (min_v - s.v) / delta_time;
      if ( (s.cur_rad == 0.0)||(s.v >= 200.0) )
         max_d = -1.1 * MAX_ACCEL;
      else if (s.v <= 60.0)
         max_d = -0.9 * MAX_ACCEL;
      else
         max_d = ((-0.2*(s.v-60.0)/140.0) - 0.9) * MAX_ACCEL;
      if (dv < max_d)
         min_v = s.v + (max_d * delta_time);
      
      nV.scale(min_v / tgt_v);
      return 2;
   }
   return 0;
}

void SBv1r4::alt_update_con_vec(situation &s, double lane, double bias,
                                 int trouble)
{
   static const double SGs = 1.0;
   static const double SDs = SGs*2.5/1.5;
   double dv, trk_wid, tgt_v;

   trk_wid   = s.to_lft + s.to_rgt;
   cp->alpha = (SGs*(s.to_lft - lane)/trk_wid) - (SDs*(s.vn/s.v)) + bias;
   if (trouble) {
      if (cp->alpha > 0.5)
         cp->alpha =  0.5;
      else if (cp->alpha < -0.5)
         cp->alpha = -0.5;
   }
   else {
      bias *= 0.2;
      if (bias > 0.0) {
         if (cp->alpha > 0.5)
            cp->alpha =  0.5;
         else if (cp->alpha < -bias)
            cp->alpha = -bias;
      }
      else if (bias < 0.0) {
         if (cp->alpha > -bias)
            cp->alpha =  -bias;
         else if (cp->alpha < -0.5)
            cp->alpha = -0.5;
      }
   }

   tgt_v = alt_get_tgt_v(s, lane);
   dv    = 32.2 * delta_time * ((s.cur_rad == 0.0)?1.25:0.95);
   if ((tgt_v - s.v) > dv)
      tgt_v = s.v + dv;
   else if ((tgt_v - s.v) < -dv)
      tgt_v = s.v - dv;
   cp->vc = tgt_v * cos(cp->alpha);
}

double SBv1r4::alt_get_tgt_v(situation &s, double lane)
{
   static const double SD = 1.25;
   static const double TD = 1.75;
   static const double m  = -0.085 / (s.to_lft + s.to_rgt - 25.0);
   static const double b  =  1.0 - (5.0 * m);
   double SPDc;
   double tgt_v, radius, trk_wid;
   double tgt_dist, tst_rad, tst_v, tst_dist;

   trk_wid = s.to_lft + s.to_rgt;
   radius  = s.cur_rad;
   if (radius > 0.0) {
      radius += lane;
      SPDc    = (m * lane) + b;
   }
   else if (radius < 0.0) {
      radius -= trk_wid - lane;
      SPDc    = (m * (trk_wid - lane)) + b;
   }
   else
      SPDc    = 1.0;
   tgt_v    =  SPDc*crn_spd(fabs(radius));
   tgt_dist = -brkdist(s.v, tgt_v) * ((s.cur_rad == 0.0)?SD:TD);
   if ( (fabs(radius) > fabs(s.nex_rad))||(radius == 0.0) ) {
      if (s.nex_rad != 0.0) {
         tst_rad   = s.nex_rad;
         tst_dist  = -2.5 * s.v * delta_time;
         tst_dist += s.to_end*((s.cur_rad == 0.0)?1.0 : fabs(s.cur_rad));
            
         if (tst_rad > 0.0) {
            tst_rad += lane;
            SPDc     = (m * lane) + b;
         }
         else if (tst_rad < 0.0) {
            tst_rad -= trk_wid - lane;
            SPDc     = (m * (trk_wid - lane)) + b;
         }
         else
            SPDc     = 1.0;
         tst_v     = SPDc*crn_spd(fabs(tst_rad));
         tst_dist -= brkdist(s.v, tst_v) * ((s.cur_rad == 0.0)?SD:TD);

         if (tst_dist < tgt_dist) {
            tgt_dist = tst_dist;
            tgt_v    = tst_v;
            radius   = tst_rad;
         }
      }
   }
   if ( (fabs(radius) > fabs(s.after_rad))||(radius == 0.0) ) {
      if (s.after_rad != 0.0) {
         tst_rad   = s.after_rad;
         tst_dist  = -5.0 * s.v * delta_time;
         tst_dist += s.to_end *((s.cur_rad == 0.0)?1.0 : fabs(s.cur_rad));
         tst_dist += s.nex_len*((s.nex_rad == 0.0)?1.0 : fabs(s.nex_rad));

         if (tst_rad > 0.0) {
            tst_rad += lane;
            SPDc     = (m * lane) + b;
         }
         else if (tst_rad < 0.0) {
            tst_rad -= trk_wid - lane;
            SPDc     = (m * (trk_wid - lane)) + b;
         }
         else
            SPDc     = 1.0;
         tst_v     = SPDc*crn_spd(fabs(tst_rad));
         tst_dist -= brkdist(s.v, tst_v) * ((s.cur_rad == 0.0)?SD:TD);

         if (tst_dist < tgt_dist) {
            tgt_dist = tst_dist;
            tgt_v    = tst_v;
            radius   = tst_rad;
         }
      }
   }
   if ( (fabs(radius) > fabs(s.aftaft_rad))||(radius == 0.0) ) {
      if (s.aftaft_rad != 0.0) {
         tst_rad   = s.aftaft_rad;
         tst_dist  = -5.0 * s.v * delta_time;
         tst_dist += s.to_end   *((s.cur_rad == 0.0)  ?1.0 : fabs(s.cur_rad));
         tst_dist += s.nex_len  *((s.nex_rad == 0.0)  ?1.0 : fabs(s.nex_rad));
         tst_dist += s.after_len*((s.after_rad == 0.0)?1.0 : fabs(s.after_rad));

         if (tst_rad > 0.0) {
            tst_rad += lane;
            SPDc     = (m * lane) + b;
         }
         else if (tst_rad < 0.0) {
            tst_rad -= trk_wid - lane;
            SPDc     = (m * (trk_wid - lane)) + b;
         }
         else
            SPDc     = 1.0;
         tst_v     = SPDc*crn_spd(fabs(tst_rad));
         tst_dist -= brkdist(s.v, tst_v) * ((s.cur_rad == 0.0)?SD:TD);

         if (tst_dist < tgt_dist) {
            tgt_dist = tst_dist;
            tgt_v    = tst_v;
            radius   = tst_rad;
         }
      }
   }
   return tgt_v;
}
//-----END SBDriver.cpp-----//


// STATIC FUNCTIONS
//-----START spd_func.cpp-----//
static double crn_spd(double radius)
{
   double retval;

   if (radius == 0.0)
      return 400.0;

   retval = sqrt(fabs(radius) * MAX_ACCEL);

   return (retval < 400.0) ? retval : 400.0;
}

static double crn_RAD2FT(double len, double rad, double wid)
{
   double retval;

   retval = len * (fabs(rad) + (0.5*wid));
   return retval;
}

static double crn_FT2RAD(double len, double rad, double wid)
{
   double retval;

   retval = len / (rad + (0.5*wid));
   return retval;
}

static double d_brk(double v)
{
   double retval;

   if (v < 250.0) {
      retval = 864.2473 - (0.217221*v) - (0.013173*v*v);
   }
   else
      retval = 0.0;
   return retval;
}

static double brkdist(double vi, double vf)
{
   if (vf >= vi)
      return 0.0;

   return ( (1.0 + ((vi - vf)*0.003)) * (d_brk(vf) - d_brk(vi)) );
}
//-----END spd_func.cpp-----//


Driver * getSBv1r4Instance()
{
   static int num = 1;
   
   return new SBv1r4(num++);
}

#undef SB_TESTING
