/*
 * SmoothB7.cpp
 * Dennis Lew (dennis_L76@yahoo.ca)
 *
 * ROBOT     : SmoothB2
 * DATA FILE : None
 *
 * Uses the OOP structure, call 'getSBv2r3Instance()' to get a
 * new instance.  All instances of SmoothB2 in a race will share
 * the same path data.
 *    
 * SmoothB will try and create its own path around the track its
 * on.  This routine will create a good path for all RARS tracks.
 * The data file can still be used to provide better paths if
 * necessary.
 *
 * If you want to use any part of this code in your own robot, or
 * are curious about anything I've done here, just let me know.
 *
 * Tracks this robot has been tested on the following tracks (71):
 *   aalborg.trk    adelaide.trk   albert.trk     albrtprk.trk
 *   anew.trk       austria.trk    barcelon.trk   brands.trk
 *   brazil.trk     buenos.trk     clkwis.trk     cstlcomb.trk
 *   doningtn.trk   dreiw.trk      elev2.trk      estoril.trk
 *   figure8.trk    fiorano.trk    fourmile.trk   hock.trk
 *   hungary.trk    imola.trk      indy500.trk    indygp.trk
 *   jerez.trk      kevin.trk      kimla01.trk    kimla02.trk
 *   kimla03.trk    kimla04.trk    kimla05.trk    longrand.trk
 *   loudon.trk     magnycrs.trk   michigan.trk   midohio.trk
 *   mlwaukee.trk   monaco.trk     montreal.trk   monza-55.trk
 *   monza-76.trk   mosport.trk    nazareth.trk   nurnburg.trk
 *   oval2.trk      phoenix.trk    pocono.trk     ra.trk
 *   ramdu.trk      rars.trk       s1.trk         s2.trk
 *   sepang(1).trk  silver97.trk   silverst.trk   spa.trk
 *   speed2.trk     stef2.trk      suzuka.trk     t_eight.trx
 *   t_spa.trx      t_zero.trx     timin.trk      tremblnt.trk
 *   trouble.trk    v01.trk        v02.trk        v03.trk
 *   watglen.trk    wierd.trk      zandvort.trk
 *
 * <CHANGES>
 * v1.4: -fixed NULL pointer bug when trying to load data
 *       -improved collision detection
 *       -robot will not actively try to pass other cars (UNDER DEVELOPMENT)
 *       -robot should be able to reliably finish races now
 *
 * v1.5: -bugs introduced by earlier improvements have been fixed
 *       -robot runs better on tracks with no data
 *
 * v1.6: -complete rewrite of the anti-collision code (MUCH BETTER)
 *
 * v2.0: -complete rewrite of the path creation routine (MUCH BETTER, BUT NEEDS MORE DEVELOPMENT)
 *       -new data file format (SMALLER THAN OLD VERSION)
 *       -improved control routine
 *       -improved anti-collision code
 *       -functional passing code (FINALLY!)
 *
 * v2.1: -changes for RARS v0.9
 *
 * v2.2: -defaults to using a constant string for its name
 *       -better control parameters for surfaces 0 and 2
 *       -small adjustments to anti-collision code
 *
 * v2.3: -rewrite of the path creation routine (WORKS ON ALL TRACKS!)
 *       -this routine takes longer than the last one:
 *           Avg Time (P200MMX): v2.2 =  0.3s
 *                               v2.3 =  1.3s
 *           Max Time (P200MMX): v2.2 = 17.7s
 *           (kimla04.trk)       v2.3 = 20.5s
 *           longrand.trk Time : v2.2 = 1min (cannot complete lap)
 *                               v2.3 = 9min
 *       -in most cases, this routine will create a faster path
 *       -small adjustments to the control routine, for better stability
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "os.h"
#include "car.h"
#include "track.h"

// Set the following macro to the number of SmoothB
// robots that will be used.
#define SB2_CARS 2
// Set the following macro to 1 if the default name
// will be used, or 0 if not
#define SB2_USE_DEFAULT_NAME 1


// CONSTANTS
static const char   *SB2_AUTHOR           = "Dennis Lew (dennis_L76@yahoo.ca)";
static const char   *SB2_DESCRIPTION      = "v2.3: Nov 29, 2002";
static const char   *SB2_NAME[1+SB2_CARS] = {"SmoothB2",    // the default name
                                             "SBv2r3-1",    // names for the other robots that may be used
                                             "SBv2r3-2"};
static const int     SB2_NOSE             = oLIGHTGRAY;
static const int     SB2_TAIL             = oGREEN;
static const char   *SB2_BITMAP2D         = "car_gray_green";
static const char   *SB2_MODEL3D          = NULL;

static const char   *DATA_FILE     = "SBv3.dat";
static const double  MY_PI         =    3.14159265358979323846;
static const double  MY_PI21       =    2.0 * MY_PI;
static const double  MY_PI12       =    0.5 * MY_PI;
static const double  MY_PI14       =    0.25 * MY_PI;
static const double  SCALE_CORNSPD =    0.98;
static const double  SB2_MAX_SPD    =  400.0;
static const double  SB2_FC         =    0.0003;  // fuel / ft
static const double  SB2_MARGIN     =   (0.5 * CARLEN) + 1.0;
static const double  SB2_MIN_DIST   =    3.0 * CARLEN;
static const double  SB2_CLOSE_RATE =   -5.0;
static const double  SB2_IN_WID     =    2.0;
static const double  SB2_OUT_WID    =    5.5;
static const double  SB2_PATH_ERR   =    0.5;
static const double  SB2_MIN_LEN    =   10.0;
static const int     SB2_NUM_RNDS   =  300;
static const int     SB2_SEARCH_LEN =    6;

// GLOBALS
static double        MAX_ACCEL;


// STATIC FUNCTIONS
//-----START spd_func.h-----//
#define SB2_car_mass(fuel) \
   (M + (fuel/g))
#define SB2_drag(damage) \
   (DRAG_CON * (1.0 + ((double)(damage << 1) / (double)MAX_DAMAGE)))
#define SB2_drag_accel(fuel, damage) \
   (SB2_drag(damage) / SB2_car_mass(fuel))


static double SB2_crn_spd(double radius);
static double SB2_brkdist(double vi, double vf, double drag_accel);
static double SB2_iu(double cur_u, int surfaceType);
static double SB2_fuel_needed(Stage s, int lapsleft, double fpl,
                             double cur_fuel);
//-----END spd_func.h-----//


// CLASSES
//-----START Primitive.h-----//
class SB2Primitive
{
public:
   virtual double dist2Point(const SB2Primitive &chk_pt) const                = 0;
   virtual int    intersect(const SB2Primitive &l) const                      = 0;
   virtual int    intersect(SB2Primitive &ret_pt, const SB2Primitive &l) const = 0;
   
protected:
   // CONSTRUCTOR
   SB2Primitive() {}
};
//-----END Primitive.h-----//

//-----START Vector.h-----//
class SB2Line;

class SB2Vector : public SB2Primitive
{
public:
   // MEMBERS
   double x, y;

   // CONSTRUCTORS
   SB2Vector();
   SB2Vector(double vx, double vy);
   SB2Vector(double angle);
   SB2Vector(const SB2Vector &v);
   
   // METHODS
   void   add(const SB2Vector &v);
   void   sub(const SB2Vector &v);
   double mag() const;
   double mag2() const;
   double dist2Point(const SB2Vector &chk_pt) const;
   double dist2Point2(const SB2Vector &chk_pt) const;
   double ang() const;
   double ang(const SB2Vector &v) const;
   void   scale(double factor);
   void   normalize();
   double dot(const SB2Vector &v) const;
   void   rotate(double angle);
   int    intersect(const SB2Line &l) const;
   int    intersect(SB2Vector &ret_pt, const SB2Line &l) const;

   double dist2Point(const SB2Primitive &chk_pt) const
      {return dist2Point((const SB2Vector&)chk_pt);}
   int    intersect(const SB2Primitive &l) const
      {return intersect((const SB2Line&)l);}
   int    intersect(SB2Primitive &ret_pt, const SB2Primitive &l) const
      {return intersect((SB2Vector&)ret_pt, (const SB2Line&)l);}
};
//-----END Vector.h-----//

//-----START Line.h-----//
class SB2Line : public SB2Primitive 
{
public:
   // MEMBERS
   SB2Vector pt;
   SB2Vector dir;

   // CONSTRUCTORS
   SB2Line();
   SB2Line(double px, double py, double dx, double dy);
   SB2Line(double px, double py, double d_ang);
   SB2Line(const SB2Line &l);
   SB2Line(const SB2Vector &Spt, const SB2Vector &Fpt);

   // METHODS
   int    intersect(const SB2Line &l) const;
   int    intersect(SB2Vector &ret_pt, const SB2Line &l) const;
   void   getPoint(SB2Vector &ret_pt, double dist) const;
   void   movePoint(double dist);
   int    movePoint(const SB2Line &l);
   double dist2Point(const SB2Vector &chk_pt) const;
   void   makeLine(const SB2Vector &Spt, const SB2Vector &Fpt);

   double dist2Point(const SB2Primitive &chk_pt) const
      {return dist2Point((const SB2Vector&)chk_pt);}
   int    intersect(const SB2Primitive &l) const
      {return intersect((const SB2Line&)l);}
   int    intersect(SB2Primitive &ret_pt, const SB2Primitive &l) const
      {return intersect((SB2Vector&)ret_pt, (const SB2Line&)l);}
};
//-----END Line.h-----//

//-----START Circle.h-----//
class SB2Circle : public SB2Primitive
{
public:
   static const int TAN_1, TAN_2, TAN_3, TAN_4;

   // MEMBERS
   SB2Vector c;
   double   R;

   // CONSTRUCTORS
   SB2Circle();
   SB2Circle(double rad);
   SB2Circle(double cx, double cy, double rad);
   SB2Circle(const SB2Circle &cc);

   // METHODS
   int    intersect(const SB2Line &l) const;
   int    intersect(SB2Vector &ret_pt, const SB2Line &l) const;
   int    intersect(SB2Vector &ret_pt1, SB2Vector &ret_pt2, 
                    const SB2Line &l) const;
   int    intersect(SB2Vector &ret_pt1, SB2Vector &ret_pt2,
                    const SB2Circle &cc) const;
   double dist2Point(const SB2Vector &chk_pt) const;
   void   getPoint(SB2Vector &ret_pt, double angle) const;
   double ang(SB2Vector &pt) const;
   int    tangent(const SB2Circle &c2, const int lineID) const;
   int    tangent(SB2Line &ret_l, const SB2Circle &c2, const int lineID) const;

   double dist2Point(const SB2Primitive &chk_pt) const
      {return dist2Point((const SB2Vector&)chk_pt);}
   int    intersect(const SB2Primitive &l) const
      {return intersect((const SB2Line&)l);}
   int    intersect(SB2Primitive &ret_pt, const SB2Primitive &l) const
      {return intersect((SB2Vector&)ret_pt, (const SB2Line&)l);}
};
//-----END Circle.h-----//

//-----START TrkSeg.h-----//
class SB2TrkSeg
{
public:
   // MEMBERS
   double       Sdst, Fdst;  // Start and End locations (measured on centre line of trk)
   double       len, rad;
   SB2Primitive *left_side;
   
   // METHODS
   virtual void   get_norm(SB2Line &ret_n, double trk_dist) const = 0;
   virtual double get_trk_dist(const SB2Vector &chk_pt) const     = 0;
   virtual int    path_ok(const SB2Line &path, double S, double F,
                          double min_w, double max_w) const      = 0;
   virtual int    path_ok(const SB2Circle &path, double S, double F,
                          double min_w, double max_w) const      = 0;

   int    same_dir(const SB2TrkSeg &seg) const;
   double to_left(const SB2Vector &chk_pt) const;
   
protected:
   // CONSTRUCTORS
   SB2TrkSeg(int ID);
};
//-----END TrkSeg.h-----//

//-----START TrkSeg_Straight.h-----//
class SB2TrkSeg_Straight : public SB2TrkSeg
{
public:
   // MEMBERS
   SB2Line   ls;
   SB2Vector norm;
   
   // CONSTRUCTORS
   SB2TrkSeg_Straight(int ID, segment *seg, double trk_dist);
   
   // METHODS
   void   get_norm(SB2Line &ret_n, double trk_dist) const;
   double get_trk_dist(const SB2Vector &chk_pt) const;
   int    path_ok(const SB2Line &path, double S, double F,
                  double min_w, double max_w) const;
   int    path_ok(const SB2Circle &path, double S, double F,
                  double min_w, double max_w) const;
};
//-----END TrkSeg_Straight.h-----//

//-----START TrkSeg_Turn.h-----//
class SB2TrkSeg_Turn : public SB2TrkSeg
{
public:
   // MEMBERS
   SB2Circle ls;
   double   beg_ang, end_ang;
   int      left;
   
   // CONSTRUCTORS
   SB2TrkSeg_Turn(int ID, segment *seg, double wid, double trk_dist);
   
   // METHODS
   void   get_norm(SB2Line &ret_n, double trk_dist) const;
   double get_trk_dist(const SB2Vector &chk_pt) const;
   int    path_ok(const SB2Line &path, double S, double F,
                  double min_w, double max_w) const;
   int    path_ok(const SB2Circle &path, double S, double F,
                  double min_w, double max_w) const;
   int    like_straight() const;
};
//-----END TrkSeg_Turn.h-----//

//-----START PathComponent.h-----//
class SB2PathComponent 
{
public:
   static const int LTURN, RTURN, STRAIGHT;

   // MEMBERS
   int           typeID;      // type of component
   double        len;
   double        Sdst, Fdst;  // Start and End locations (measured on centre line of trk)
   int           num_segs;    // number of track segments this component covers
   SB2TrkSeg    **seg;         // track segments (data is taken from left wall only)
   SB2Primitive  *comp;        // the actual path component

   SB2PathComponent **pass;    // alternate path route for passing
   double            pre_len, post_len, len_trk;
   
   // DESTRUCTORS
   ~SB2PathComponent();

   // METHODS
   virtual double get_spd(double trk_dist, double *max=0) const = 0;
   virtual void   get_tan(SB2Vector &ret_pt, SB2Line &ret_norm) const = 0;
   virtual double get_path_dist(const SB2Vector &pt) const = 0;
   virtual double get_dist2path(const SB2Vector &pt) const = 0;
   virtual void   getPoint2(SB2Vector &ret_pt, SB2Vector &ret_tan, 
                            double path_dist) const = 0;
   virtual double brk_dist(double dist, double v) const = 0;

   void   copy(const SB2PathComponent &pc);
   int    getPoint(SB2Vector &ret_pt, SB2Line &ret_norm, 
                   double trk_dist, int ret_tan=0) const;
   double to_side(SB2Vector &ret_pt, SB2Line &ret_norm, 
                  double trk_dist, int ret_tan=0) const;
   void   insert_TrkSeg(SB2TrkSeg *ts, int idx);
   void   insert_TrkSeg(const SB2PathComponent &pc, int add_to_tail);
   void   remove_TrkSeg(int idx);
   void   trim();
   int    find_trk_seg(const SB2Vector &pt, double trk_wid,
                       int start=0, int dir=1) const;
   int    find_trk_seg(double trk_dist, int start=0, int dir=1) const;
   int    is_after(int idx, double dst, const SB2PathComponent &prev,
                   int prev_idx, double prev_dst) const;
   int    is_after(double dst, const SB2PathComponent &prev,
                   double prev_dst) const;
   double dst_between(double dst, const SB2PathComponent &prev,
                      double prev_dst) const;
   int    valid(double trk_wid) const;
   
protected:
   // MEMBERS
   int seg_length;

   // CONSTRUCTORS
   SB2PathComponent();

   virtual int  path_ok(int seg_idx, double S, double F, 
                        double min_w, double max_w) const = 0;
   virtual void move_to_outside(SB2Line &ret_norm) const = 0;

private:
   // METHODS
   double dst_between(int idx, double dst, int prev_idx, 
                      double prev_dst) const;
};
//-----END PathComponent.h-----//

//-----START Straight.h-----//
class SB2Straight : public SB2PathComponent 
{
public:
   // CONSTRUCTORS
   SB2Straight();
   
   // METHODS
   double get_spd(double trk_dist, double *max=0) const;
   void   get_tan(SB2Vector &ret_pt, SB2Line &ret_norm) const;
   double get_path_dist(const SB2Vector &pt) const;
   double get_dist2path(const SB2Vector &pt) const;
   void   getPoint2(SB2Vector &ret_pt, SB2Vector &ret_tan, 
                    double path_dist) const;
   double brk_dist(double dist, double v) const;

   void   copy(const SB2Straight &sc);
   void   adjust(const SB2Vector &Spt, const SB2Vector &Fpt);
   int    intersect(SB2Vector &ret_pt, const SB2Straight &s) const;

protected:
   int  path_ok(int seg_idx, double S, double F, 
                double min_w, double max_w) const;
   void move_to_outside(SB2Line &ret_norm) const {};

private:
   // MEMBERS
   SB2Line l;
};
//-----END Straight.h-----//

//-----START Turn.h-----//
class SB2Turn : public SB2PathComponent 
{
public:
   static const int DATALEN,      // Lines per Turn in data file
                    BAD_ENTRY,    // Entry to turn needs adjustment
                    BAD_EXIT,     // Exit from turn needs adjustment
                    BAD_TURN,     // Turn is not completely on course
                    IGNORE_TURN;  // Turn is not needed in path
   
   // CONSTRUCTORS
   SB2Turn();
   SB2Turn(SB2TrkSeg **ts, int idx, int trk_segs,
          double min_w, double max_w, double trk_wid);
   
   // METHODS
   double get_spd(double trk_dist, double *max=0) const;
   void   get_tan(SB2Vector &ret_pt, SB2Line &ret_norm) const;
   double get_path_dist(const SB2Vector &pt) const;
   double get_dist2path(const SB2Vector &pt) const;
   void   getPoint2(SB2Vector &ret_pt, SB2Vector &ret_tan, 
                    double path_dist) const;
   double brk_dist(double dist, double v) const;

   void   init(double min_w, double max_w, double trk_wid);
   void   copy(const SB2Turn &tc, int copy_segs);
   void   copy(const SB2Turn &tc, int copy_segs, int trim_tail,
               int extra_segs);
   void   adjust(SB2Turn &pt, SB2Turn &nt, int adj_val,
                 double min_w, double max_w, double trk_wid);
   int    makeStraight(SB2Straight &ret, SB2Turn &pt, double trk_wid);
   void   set_spd_len(int data_loaded);
   int    makePass(SB2Straight *p_s, SB2Turn *p_t, SB2Turn &nt,
                   SB2Turn &last, double min_w, double max_w, 
                   double trk_wid);
   void   save(FILE *out, SB2TrkSeg **ts, double rotate) const;
   int    load(FILE *in, SB2TrkSeg **ts, int trk_segs, double rotate);

protected:
   int  path_ok(int seg_idx, double S, double F, 
                double min_w, double max_w) const;
   void move_to_outside(SB2Line &ret_norm) const;

private:
   // MEMBERS
   SB2Circle       curv;
   double         apex_dist;  // Apex position
   SB2TrkSeg_Turn *ts_turn;
   double         plen, nlen, Sang;
   double         min_spd, max_spd;
   
   // METHODS
   int    get_ts_turn_idx() const;
   double get_apex(double dst, double min_w, double max_w) const;
   double get_apex() const;
   void   moveApex(double amount, double min_w, double max_w, 
                   double trk_wid);
   double adjRadius(double amount, double min_w, double max_w);
   double moveTurn(double amount, double min_w, double max_w);
   void   setLimits(double trk_wid);
   void   set(double apex, double min_w, double max_w, double trk_wid);
   double get_weight(const SB2Turn &ot) const;
   double get_err(SB2Turn &pt, int get_pt_err, double trk_wid);
   void   fix_bad_turn(double min_w, double max_w, double trk_wid);
};
//-----END Turn.h-----//

//-----START TrkData.h-----//
class SB2TrkData
{
public:
   static const int NORMAL, OFF, START;

   // MEMBERS
   double pit_entry, pit_spd;

   // CONSTRUCTORS
   SB2TrkData();
   
   // DESTUCTOR
   ~SB2TrkData();
   
   // METHODS
   int    can_pass(double trk_dist, int pc) const;
   double get_trk_dist(int seg_ID, double to_end) const;
   double get_trk_dist(const SB2Vector &pt, double to_lft, 
                       int start_seg_ID=0) const;
   void   get_trk_pos_vel(SB2Line &ret_pv, double to_lft, 
                          double v, double vn, 
                          double trk_dist, int seg_ID) const;
   int    get_PathComponent(double trk_dist, int pass_flag=0,
                            int start=0) const;
   double get_tgt_spd(double v, double drag_accel, 
                      double path_dist, double trk_dist,
                      int pass_flag=0, int pc=-1) const;
   int    get_nex_pos(SB2Line &cV, SB2Vector &nV, int type,
                      double trk_dist, int pass_flag=0, int pc=-1) const;
   void   locate_car(double *dist2path, double *path_dist, 
                     const SB2Vector &pt, int start_pc, int end_pc,
                     double trk_dist=0.0, int pass_flag=0) const;
   void   set_path(Track *trk);
   
private:
   // MEMBERS
   char             *name;
   SB2PathComponent **path;
   SB2TrkSeg        **seg;
   int               num_segs, num_comps;
   double            trk_len, trk_wid;

   // PRIVATE METHODS
   int  get_trk_data(Track *trk);
   int  init_Turns(SB2Turn ***p_crn, int *p_num_turns);
   void adjust_Turns(SB2Turn **crn, SB2Straight **str, 
                     int num_turns, int rnds, int level, int *adj_val);
   void insert_Straights(SB2Turn **crn, SB2Straight **str, 
                         int num_turns, int data_loaded);
   void makePassingPath();
   void saveData() const;
   int  loadData(SB2Turn ***p_crn, int *p_num_turns);

   int  get_PassComponent(double trk_dist, int pc) const;
};
//-----END TrkData.h-----//

//-----START SB2Driver.h-----//
#ifdef SB2_DLL
class SBv2r3
#else
class SBv2r3 : public Driver
#endif
{
public:
   // CONSTRUCTORS
   SBv2r3(int i=-1);
   
   // DESTRUCTOR
   ~SBv2r3();

   // METHODS
   con_vec drive(situation &s);
   
private:
   // PRIVATE MEMBERS
   int      init;
   con_vec  cp;
   int      lastpit, pc, pass_flag, car2pass;
   double   fuelstart;
   SB2Line   cV;
   SB2Vector nV;

   // PRIVATE METHODS
   void   get_RARS_ctrl_params(situation &s);
   void   update_con_vec(int damp);
   void   locate_car(double *dist2path, double *path_dist,
                     const SB2Vector &pt, double to_lft, int seg_ID);
   double anti_collision(situation &s, double path_dist, double tgt_v,
                         double drag_accel);
   int    pass_clear(situation &s, double trk_dist);
};
//-----END SB2Driver.h-----//


//-----START Vector.cpp-----//
// CONSTRUCTORS
SB2Vector::SB2Vector() 
{
   x = y = 0.0;
}

SB2Vector::SB2Vector(double vx, double vy) 
{
   x = vx;
   y = vy;
}

SB2Vector::SB2Vector(double angle)
{
   x = cos(angle);
   y = sin(angle);
}

SB2Vector::SB2Vector(const SB2Vector &v)
{
   x = v.x;
   y = v.y;
}

// METHODS
void SB2Vector::add(const SB2Vector &v) 
{
   x += v.x;
   y += v.y;
}
   
void SB2Vector::sub(const SB2Vector &v) 
{
   x -= v.x;
   y -= v.y;
}

double SB2Vector::mag() const
{
   return sqrt((x*x) + (y*y));
}

double SB2Vector::mag2() const
{
   return ((x*x) + (y*y));
}

double SB2Vector::dist2Point(const SB2Vector &chk_pt) const
{
   double xd = x - chk_pt.x;
   double yd = y - chk_pt.y;

   return sqrt((xd*xd) + (yd*yd));
}

double SB2Vector::dist2Point2(const SB2Vector &chk_pt) const
{
   double xd = x - chk_pt.x;
   double yd = y - chk_pt.y;

   return ((xd*xd) + (yd*yd));
}

double SB2Vector::ang() const
{
   double retval;

   if (x != 0.0)
      retval = atan(y / x);
   else {
      if (y >= 0.0)
         retval =  MY_PI12;
      else
         retval = -MY_PI12;
   }

   if (x < 0.0)                       // Quad 2, Quad 3
      retval += MY_PI;
   else if ( (x >= 0.0)&&(y < 0.0) )  // Quad 4
      retval += MY_PI21;

   return retval;
}

double SB2Vector::ang(const SB2Vector &v) const
{
   double retval = v.ang() - ang();

   if (retval > MY_PI)
      retval -= MY_PI21;
   else if (retval < -MY_PI)
      retval += MY_PI21;
   return retval;
}
   
void SB2Vector::scale(double factor) 
{
   x *= factor;
   y *= factor;
}
   
void SB2Vector::normalize()
{
   double m;

   m = mag();
   if (m != 0.0)
      scale(1.0 / m);
}

double SB2Vector::dot(const SB2Vector &v) const
{
   return (x*v.x) + (y*v.y);
}

void SB2Vector::rotate(double angle)
{
   double m, a;

   m = mag();
   a = ang();
   x = m * cos(a + angle);
   y = m * sin(a + angle);
}

int SB2Vector::intersect(const SB2Line &l) const
{
   double   da;
   SB2Vector tv(*this);

   tv.sub(l.pt);
   da = fabs(tv.ang(l.dir));
   if (da > MY_PI12)
      da = MY_PI - da;
   return (da <= 0.017453292);
}

int SB2Vector::intersect(SB2Vector &ret_pt, const SB2Line &l) const
{
   int retval = intersect(l);
   
   if (retval) {
      ret_pt.x = x;
      ret_pt.y = y;
   }
   return retval;
}
//-----END Vector.cpp-----//

//-----START Line.cpp-----//
// CONSTRUCTORS
SB2Line::SB2Line()
   : pt(), dir()
{
}

SB2Line::SB2Line(double px, double py, double dx, double dy)
   : pt(px, py), dir(dx, dy)
{
   dir.normalize();
}

SB2Line::SB2Line(double px, double py, double d_ang)
   : pt(px, py), dir(d_ang)
{
}

SB2Line::SB2Line(const SB2Line &l)
   : pt(l.pt), dir(l.dir)
{
}

SB2Line::SB2Line(const SB2Vector &Spt, const SB2Vector &Fpt)
   : pt(Spt), dir(Fpt)
{
   dir.sub(Spt);
}

// METHODS
int SB2Line::intersect(const SB2Line &l) const
{
   double D = (l.dir.x*dir.y) - (l.dir.y*dir.x);
   return (D != 0.0);
}

int SB2Line::intersect(SB2Vector &ret_pt, const SB2Line &l) const
{
   double N, D;

   N = (dir.x*(l.pt.y - pt.y)) - (dir.y*(l.pt.x - pt.x));
   D = (l.dir.x*dir.y) - (l.dir.y*dir.x);

   if (D == 0.0)
      return 0;

   l.getPoint(ret_pt, (N/D));
   return 1;
}

void SB2Line::getPoint(SB2Vector &ret_pt, double dist) const
{
   ret_pt.x = pt.x + (dist*dir.x);
   ret_pt.y = pt.y + (dist*dir.y);
}

void SB2Line::movePoint(double dist)
{
   pt.x += dist * dir.x;
   pt.y += dist * dir.y;
}

int SB2Line::movePoint(const SB2Line &l)
{
   double N, D;

   N = (dir.x*(l.pt.y - pt.y)) - (dir.y*(l.pt.x - pt.x));
   D = (l.dir.x*dir.y) - (l.dir.y*dir.x);

   if (D == 0.0)
      return 0;

   l.getPoint(pt, (N/D));
   return 1;
}

double SB2Line::dist2Point(const SB2Vector &chk_pt) const
{
   double   N, D;
   SB2Vector norm(dir.ang() + MY_PI12);

   N = (dir.x*(chk_pt.y - pt.y)) - (dir.y*(chk_pt.x - pt.x));
   D = (norm.x*dir.y) - (norm.y*dir.x);

   // +ve return value means chk_pt is to right of this line
   // -ve return value means chk_pt is to left of this line
   return (N/D);
}

void SB2Line::makeLine(const SB2Vector &Spt, const SB2Vector &Fpt)
{
   pt.x  = Spt.x; pt.y  = Spt.y;
   dir.x = Fpt.x; dir.y = Fpt.y;
   dir.sub(Spt);
}
//-----END Line.cpp-----//

//-----START Circle.cpp-----//
const int SB2Circle::TAN_1 = 1,
          SB2Circle::TAN_2 = 2,
          SB2Circle::TAN_3 = 3,
          SB2Circle::TAN_4 = 4;


// CONSTRUCTORS
SB2Circle::SB2Circle()
   : c()
{
   R = 0.0;
}

SB2Circle::SB2Circle(double rad)
   : c()
{
   R = rad;
}

SB2Circle::SB2Circle(double cx, double cy, double rad)
   : c(cx, cy)
{
   R = rad;
}

SB2Circle::SB2Circle(const SB2Circle &cc)
   : c(cc.c)
{
   R = cc.R;
}

// METHODS
int SB2Circle::intersect(const SB2Line &l) const
{
   double   E, F, G, det;
   SB2Vector dif(l.pt);
   
   dif.sub(c);

   E = l.dir.mag2();
   F = 2.0 * dif.dot(l.dir);
   G = dif.mag2() - (R * R);

   det = (F*F) - (4.0*E*G);
   return (  (det >= 0.0)&&( (E != 0.0)||(F != 0.0) )  );
}

int SB2Circle::intersect(SB2Vector &ret_pt, const SB2Line &l) const
{
   double   E, F, G, det;
   double   d1, d2;
   SB2Vector dif(l.pt);
   
   dif.sub(c);

   E = l.dir.mag2();
   F = 2.0 * dif.dot(l.dir);
   G = dif.mag2() - (R * R);

   det = (F*F) - (4.0*E*G);
   if (  (det < 0.0)||( (E == 0.0)&&(F == 0.0) )  )
      return 0;

   if (E == 0.0)  // Only 1 solution
      l.getPoint(ret_pt, (-G / F));
   else {
      // 2 solutions, choose 1
      d1 = (-F + sqrt(det)) / (2.0*E);
      d2 = (-F - sqrt(det)) / (2.0*E);

      // find closest point
      if (fabs(d1) < fabs(d2))
         l.getPoint(ret_pt, d1);
      else
         l.getPoint(ret_pt, d2);
   }
   return 1;
}

int SB2Circle::intersect(SB2Vector &ret_pt1, SB2Vector &ret_pt2, 
                        const SB2Line &l) const
{
   double   E, F, G, det;
   double   d1, d2;
   SB2Vector dif(l.pt);

   dif.sub(c);

   E = l.dir.mag2();
   F = 2.0 * dif.dot(l.dir);
   G = dif.mag2() - (R * R);

   det = (F*F) - (4.0*E*G);
   if (  (det < 0.0)||( (E == 0.0)&&(F == 0.0) )  )
      return 0;

   if (E == 0.0) {
      // Only 1 solution
      l.getPoint(ret_pt1, (-G / F));
      ret_pt2.x = ret_pt1.x;
      ret_pt2.y = ret_pt1.y;
   }
   else {
      // 2 solutions
      d1 = (-F + sqrt(det)) / (2.0*E);
      d2 = (-F - sqrt(det)) / (2.0*E);

      // find closest point
      if (fabs(d1) < fabs(d2)) {
         l.getPoint(ret_pt1, d1);
         l.getPoint(ret_pt2, d2);
      }
      else {
         l.getPoint(ret_pt1, d2);
         l.getPoint(ret_pt2, d1);
      }
   }
   return 1;
}

int SB2Circle::intersect(SB2Vector &ret_pt1, SB2Vector &ret_pt2,
                        const SB2Circle &cc) const
{
   double N, D, dst;
   SB2Vector dc(cc.c);

   dc.sub(c);
   dst = dc.mag();
   if (R >= cc.R) {
      if ( ((dst+cc.R) < R)||((dst-cc.R) > R) )
         return 0;
   }
   else {
      if ( ((dst+R) < cc.R)||((dst-R) > cc.R) )
         return 0;
   }

   N = (dst*dst) + (R*R) - (cc.R*cc.R);
   D = 2.0 * dst * R;
   if (D == 0.0)
      return 0;
   N = acos(N/D);
   D = dc.ang();

   getPoint(ret_pt1, (D+N));
   getPoint(ret_pt2, (D-N));
   return 1;
}

double SB2Circle::dist2Point(const SB2Vector &chk_pt) const
{
   // +ve return value means chk_pt is outside this circle
   // -ve return value means chk_pt is inside this circle
   return (c.dist2Point(chk_pt) - R);
}

void SB2Circle::getPoint(SB2Vector &ret_pt, double angle) const
{
   ret_pt.x = c.x + (R * cos(angle));
   ret_pt.y = c.y + (R * sin(angle));
}

double SB2Circle::ang(SB2Vector &pt) const
{
   double retval;

   pt.sub(c);
   retval = pt.ang();
   pt.add(c);
   return retval;
}

int SB2Circle::tangent(const SB2Circle &c2, const int lineID) const
{
   double D = c.dist2Point(c2.c);

   if ( (lineID == SB2Circle::TAN_1)||(lineID == SB2Circle::TAN_2) ) {
      if ( (R < c2.R)&&((D+R) < c2.R) )
         return 0;
      else if ( (R > c2.R)&&((D+c2.R) < R) )
         return 0;
      D *= R + c2.R;
      if (D == 0.0)
         return 0;
      return 1;
   }
   else if ( (lineID == SB2Circle::TAN_3)||(lineID == SB2Circle::TAN_4) ) {
      if ( (D < (R+c2.R))||(D == 0.0) )
         return 0;
      return 1;
   }
   return 0;
}

int SB2Circle::tangent(SB2Line &ret_l, const SB2Circle &c2, 
                      const int lineID) const
{
   double N, D, ang;

   ret_l.pt.x = c.x; ret_l.pt.y = c.y;
   ret_l.pt.sub(c2.c);
   D = ret_l.pt.mag();

   if ( (lineID == SB2Circle::TAN_1)||(lineID == SB2Circle::TAN_2) ) {
      if ( (R < c2.R)&&((D+R) < c2.R) )
         return 0;
      else if ( (R > c2.R)&&((D+c2.R) < R) )
         return 0;
      D *= R + c2.R;
      if (D == 0.0)
         return 0;
      N  = (c2.R*c2.R) - (R*R);
      if (lineID == SB2Circle::TAN_1)
         ang = ret_l.pt.ang() - acos(N/D);
      else
         ang = ret_l.pt.ang() + acos(N/D);
      getPoint(ret_l.pt, ang);
      c2.getPoint(ret_l.dir, ang);
      ret_l.dir.sub(ret_l.pt);
      return 1;
   }
   else if ( (lineID == SB2Circle::TAN_3)||(lineID == SB2Circle::TAN_4) ) {
      if ( (D < (R+c2.R))||(D == 0.0) )
         return 0;
      else if (R != c2.R) {
         D  *= R - c2.R;
         N   = (c2.R*c2.R) - (R*R);
         ang = ret_l.pt.ang();
      }
      else {
         N   = 2.0 * R;
         ang = ret_l.pt.ang() - MY_PI;
      }

      if (lineID == SB2Circle::TAN_3)
         ang += acos(N/D);
      else
         ang -= acos(N/D);
      getPoint(ret_l.pt, ang);
      c2.getPoint(ret_l.dir, (ang+MY_PI));
      ret_l.dir.sub(ret_l.pt);
      return 1;
   }
   return 0;
}
//-----END Circle.cpp-----//

//-----START TrkSeg.cpp-----//
// CONSTRUCTORS
SB2TrkSeg::SB2TrkSeg(int ID)
{
   Sdst  = Fdst = 0.0;
   len   = rad  = 0.0;
   left_side = 0;
}

// METHODS
int SB2TrkSeg::same_dir(const SB2TrkSeg &seg) const
{
   if (rad == 0.0)
      return (seg.rad == 0.0);
   else if (rad > 0.0)
      return (seg.rad > 0.0);
   return (seg.rad < 0.0);
}

double SB2TrkSeg::to_left(const SB2Vector &chk_pt) const
{
   if (rad >= 0.0)
      return left_side->dist2Point(chk_pt);
   return -left_side->dist2Point(chk_pt);
}
//-----END TrkSeg.cpp-----//

//-----START TrkSeg_Straight.cpp-----//
// CONSTRUCTORS
SB2TrkSeg_Straight::SB2TrkSeg_Straight(int ID, segment *seg, double trk_dist)
   : SB2TrkSeg(ID),
     ls(seg->beg_x, seg->beg_y, seg->beg_ang),
     norm(seg->beg_ang - MY_PI12)
{
   Sdst = trk_dist;
   Fdst = trk_dist + seg->length;
   len  = seg->length;
   rad  = seg->radius;

   left_side = &ls;
}

// METHODS
void SB2TrkSeg_Straight::get_norm(SB2Line &ret_n, double trk_dist) const
{
   if ( (trk_dist < (Sdst-2.5))||(trk_dist > (Fdst+2.5)) ) {
      ret_n.pt.scale(0.0);
      ret_n.dir.scale(0.0);
   }
   else {
      ls.getPoint(ret_n.pt, (trk_dist - Sdst));
      ret_n.dir.x = norm.x;
      ret_n.dir.y = norm.y;
   }
}

double SB2TrkSeg_Straight::get_trk_dist(const SB2Vector &chk_pt) const
{
   double N, D;

   N = (norm.x*(chk_pt.y - ls.pt.y)) - (norm.y*(chk_pt.x - ls.pt.x));
   D = (ls.dir.y*norm.x) - (ls.dir.x*norm.y);

   return ((N/D) + Sdst);
}

int SB2TrkSeg_Straight::path_ok(const SB2Line &path, double S, double F,
                               double min_w, double max_w) const
{
   double   trk_dist;
   SB2Vector pt;
   SB2Line   edge(ls);

   // check finish edge
   edge.movePoint(F - Sdst);
   edge.dir.x = norm.x;
   edge.dir.y = norm.y;
   if (edge.intersect(pt, path)) {
      trk_dist = pt.dist2Point(edge.pt);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
   }
   else
      return 0;

   // check start edge
   edge.pt.x = ls.pt.x + ((S-Sdst) * ls.dir.x);
   edge.pt.y = ls.pt.y + ((S-Sdst) * ls.dir.y);
   if (edge.intersect(pt, path)) {
      trk_dist = pt.dist2Point(edge.pt);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
   }
   else
      return 0;

   // check left edge
   edge.pt.x  = ls.pt.x + (min_w * norm.x);
   edge.pt.y  = ls.pt.y + (min_w * norm.y);
   edge.dir.x = ls.dir.x;
   edge.dir.y = ls.dir.y;
   if (edge.intersect(pt, path)) {
      trk_dist = get_trk_dist(pt);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check right edge
   edge.pt.x += (max_w-min_w) * norm.x;
   edge.pt.y += (max_w-min_w) * norm.y;
   if (edge.intersect(pt, path)) {
      trk_dist = get_trk_dist(pt);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }
   return 1;
}

int SB2TrkSeg_Straight::path_ok(const SB2Circle &path, double S, double F,
                               double min_w, double max_w) const
{
   double   trk_dist;
   SB2Vector pt1;
   SB2Vector pt2;
   SB2Line   edge(ls);

   // check left edge
   edge.pt.x += min_w * norm.x;
   edge.pt.y += min_w * norm.y;
   if (path.intersect(pt1, pt2, edge)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check right edge
   edge.pt.x += (max_w-min_w) * norm.x;
   edge.pt.y += (max_w-min_w) * norm.y;
   if (path.intersect(pt1, pt2, edge)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }
   return 1;
}
//-----END TrkSeg_Straight.cpp-----//

//-----START TrkSeg_Turn.cpp-----//
// CONSTRUCTORS
SB2TrkSeg_Turn::SB2TrkSeg_Turn(int ID, segment *seg, double wid,
                             double trk_dist)
   : SB2TrkSeg(ID),
     ls(seg->cen_x, seg->cen_y, fabs(seg->radius))
{
   beg_ang = seg->beg_ang;
   end_ang = seg->end_ang;
   if (seg->radius > 0.0) {    // left turn
      len      = seg->length * (( 0.5*wid) + seg->radius);
      rad      = seg->radius;
      beg_ang -= MY_PI12;
      end_ang -= MY_PI12;
      left     = 1;
   }
   else {                     // right turn
      len      = seg->length * ((-0.5*wid) - seg->radius);
      rad      = seg->radius + wid;
      beg_ang += MY_PI12;
      end_ang += MY_PI12;
      left     = 0;
   }

   if (beg_ang < 0.0)
      beg_ang += MY_PI21;
   else if (beg_ang > MY_PI21)
      beg_ang -= MY_PI21;
   if (end_ang < 0.0)
      end_ang += MY_PI21;
   else if (end_ang > MY_PI21)
      end_ang -= MY_PI21;

   if ( (rad < 0.0)&&(end_ang > beg_ang) )
      beg_ang += MY_PI21;
   else if ( (rad > 0.0)&&(end_ang < beg_ang) )
      end_ang += MY_PI21;

   Sdst = trk_dist;
   Fdst = trk_dist + len;
   len  = seg->length;

   left_side = &ls;
}

// METHODS
void SB2TrkSeg_Turn::get_norm(SB2Line &ret_n, double trk_dist) const
{
   double ang;

   if ( (trk_dist < (Sdst-2.5))||(trk_dist > (Fdst+2.5)) ) {
      ret_n.pt.scale(0.0);
      ret_n.dir.scale(0.0);
      return;
   }

   ang  = (trk_dist - Sdst) / (Fdst - Sdst);
   ang *= end_ang - beg_ang;
   ls.getPoint(ret_n.pt, (ang + beg_ang));
   if (rad > 0.0) {
      ret_n.dir.x = ret_n.pt.x - ls.c.x;
      ret_n.dir.y = ret_n.pt.y - ls.c.y;
   }
   else {
      ret_n.dir.x = ls.c.x - ret_n.pt.x;
      ret_n.dir.y = ls.c.y - ret_n.pt.y;
   }
   ret_n.dir.normalize();
}

double SB2TrkSeg_Turn::get_trk_dist(const SB2Vector &chk_pt) const
{
   double   dist;
   SB2Vector loc(chk_pt);

   loc.sub(ls.c);
   dist  = loc.ang() - beg_ang;
   if (dist > (MY_PI21-0.0000001))
      dist -= MY_PI21-0.0000001;
   else if (dist < (-MY_PI21+0.0000001))
      dist += MY_PI21-0.0000001;
   if ( (beg_ang < end_ang)&&(dist < 0.0) )
      dist += MY_PI21;
   else if ( (beg_ang > end_ang)&&(dist > 0.0) )
      dist -= MY_PI21;
   dist *= (Fdst - Sdst) / (end_ang - beg_ang);
   return (dist + Sdst);
}

int SB2TrkSeg_Turn::path_ok(const SB2Line &path, double S, double F,
                           double min_w, double max_w) const
{
   double   trk_dist, ang;
   SB2Vector pt1, pt2;
   SB2Circle edge(ls);

   edge.R = fabs(rad);

   // check start and finish edges
   ang  = (S - Sdst) / (Fdst - Sdst);
   ang *= end_ang - beg_ang;
   edge.getPoint(pt1, (ang + beg_ang));
   ang  = (F - Sdst) / (Fdst - Sdst);
   ang *= end_ang - beg_ang;
   edge.getPoint(pt2, (ang + beg_ang));
   if (rad > 0.0) {  // left turn
      trk_dist = -path.dist2Point(pt1);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
      trk_dist = -path.dist2Point(pt2);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
   }
   else {
      trk_dist = path.dist2Point(pt1);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
      trk_dist = path.dist2Point(pt2);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
   }

   // check inner edge
   edge.R += min_w;
   if (edge.intersect(pt1, pt2, path)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check outer edge
   edge.R += max_w - min_w;
   if (edge.intersect(pt1, pt2, path)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      return 1;
   }
   return 0;
}

int SB2TrkSeg_Turn::path_ok(const SB2Circle &path, double S, double F,
                           double min_w, double max_w) const
{
   double   trk_dist;
   SB2Vector pt1, pt2;
   SB2Circle edge(ls);

   edge.R = fabs(rad);

   // check inner edge
   edge.R += min_w;
   if (edge.intersect(pt1, pt2, path)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check outer edge
   edge.R += max_w - min_w;
   if (edge.intersect(pt1, pt2, path)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }
   return 1;
}

int SB2TrkSeg_Turn::like_straight() const
{
   return ( ((fabs(rad) > 1000.0)&&(len < 0.55))||
            ((fabs(rad) > 1100.0)&&(len < 0.91))||
            ((fabs(rad) > 1400.0)&&(len < 1.05)) );
}
//-----END TrkSeg_Turn.cpp-----//

//-----START PathComponent.cpp-----//
const int SB2PathComponent::LTURN    =  1;
const int SB2PathComponent::RTURN    = -1;
const int SB2PathComponent::STRAIGHT =  0;


// CONSTRUCTORS
SB2PathComponent::SB2PathComponent()
{
   Sdst       = Fdst = len = 0.0;
   num_segs   = 0;
   seg        = (SB2TrkSeg **)malloc(sizeof(SB2TrkSeg *));
   seg[0]     = NULL;
   seg_length = 1;
   comp       = NULL;
   pass       = NULL;
   typeID     = 0;

   pre_len = post_len = len_trk = 0.0;
}

// DESTRUCTORS
SB2PathComponent::~SB2PathComponent()
{
   if (seg != NULL)
      free(seg);

   if (pass != NULL) {
      delete pass[0];
      delete pass[1];
      delete [] pass;
   }
}

// METHODS
void SB2PathComponent::copy(const SB2PathComponent &pc)
{
   typeID   = pc.typeID;
   len      = pc.len;
   Sdst     = pc.Sdst;
   Fdst     = pc.Fdst;
   num_segs = pc.num_segs;
   if ( (seg == NULL)||(num_segs > seg_length) ) {
      seg_length = num_segs + 2;
      seg        = (SB2TrkSeg **)realloc(seg, (seg_length*sizeof(SB2TrkSeg *)));
   }
   if (num_segs > 0)
      memcpy(seg, pc.seg, (num_segs*sizeof(SB2TrkSeg *)));
}

int SB2PathComponent::getPoint(SB2Vector &ret_pt, SB2Line &ret_norm,
                              double trk_dist, int ret_tan) const
{
   int i;

   for (i = 0; i < num_segs; i++) {
      if ( (trk_dist >= seg[i]->Sdst)&&
           (trk_dist <= seg[i]->Fdst) )
         break;
   }
   if (i >= num_segs) {
      for (i = 0; i < num_segs; i++) {
         if ( (trk_dist >= seg[i]->Sdst)&&
              (trk_dist < (seg[i]->Fdst+2.5)) )
            break;
      }
      if (i >= num_segs)
         return 0;
   }

   seg[i]->get_norm(ret_norm, trk_dist);
   move_to_outside(ret_norm);
   i = comp->intersect(ret_pt, ret_norm);
   if ( (i)&&(ret_tan) )
      get_tan(ret_pt, ret_norm);
   return i;
}

double SB2PathComponent::to_side(SB2Vector &ret_pt, SB2Line &ret_norm,
                                double trk_dist, int ret_tan) const
{
   int    i;
   double retval;

   for (i = 0; i < num_segs; i++) {
   if ( (trk_dist >= seg[i]->Sdst)&&
        (trk_dist <= seg[i]->Fdst) )
      break;
   }
   if (i >= num_segs)
      return 0.0;

   seg[i]->get_norm(ret_norm, trk_dist);
   move_to_outside(ret_norm);
   if (!comp->intersect(ret_pt, ret_norm))
      return 0.0;

   if (seg[i]->rad >= 0.0)
      retval =  seg[i]->left_side->dist2Point(ret_pt);
   else
      retval = -seg[i]->left_side->dist2Point(ret_pt);

   if (ret_tan)
      get_tan(ret_pt, ret_norm);
   return ((retval == 0.0) ? 0.001 : retval);
}

void SB2PathComponent::insert_TrkSeg(SB2TrkSeg *ts, int idx)
{
   int i;

   // check for duplicates
   for (i = 0; i < num_segs; i++) {
      if (seg[i] == ts)
         return;
   }

   if (idx < 0)
      idx = 0;
   else if (idx > num_segs)
      idx = num_segs;

   num_segs++;
   if (num_segs > seg_length) {
      seg_length = num_segs + 2;
      seg        = (SB2TrkSeg **)realloc(seg, (seg_length*sizeof(SB2TrkSeg *)));
   }
   for (i = num_segs-1; i > idx; i--) {
      seg[i] = seg[i-1];
   }
   seg[idx] = ts;
}

void SB2PathComponent::insert_TrkSeg(const SB2PathComponent &pc, 
                                    int add_to_tail)
{
   int s;

   if (add_to_tail) {
      for (s = 0; s < pc.num_segs; s++) {
         insert_TrkSeg(pc.seg[s], num_segs);
      }
   }
   else {
      for (s = pc.num_segs-1; s >= 0; s--) {
         insert_TrkSeg(pc.seg[s], 0);
      }
   }
}

void SB2PathComponent::remove_TrkSeg(int idx)
{
   int i;

   if ( (idx < 0)||(idx >= num_segs) )
      return;

   num_segs--;
   for (i = idx; i < num_segs; i++) {
      seg[i] = seg[i+1];
   }
   seg[i] = NULL;
}

void SB2PathComponent::trim()
{
   int s, f;

   if (num_segs > 0) {
      s = find_trk_seg(Sdst, 0, 1);
      f = find_trk_seg(Fdst, (num_segs-1), -1);
      if ( (s >= 0)&&(f >= 0)&&(f < s) ) {
         int t;
            
         t = s;
         s = f;
         f = t;
      }
      if (f >= 0)
         f = num_segs - 1 - f;
      
      // remove unused TrkSeg from component
      for (; s > 0; s--) {
         remove_TrkSeg(0);
      }
      for (; f > 0; f--) {
         remove_TrkSeg(num_segs-1);
      }
   }

   if (seg_length > num_segs) {
      // free unused memory
      seg_length = num_segs;
      seg        = (SB2TrkSeg **)realloc(seg, (seg_length*sizeof(SB2TrkSeg *)));
   }
}

int SB2PathComponent::find_trk_seg(const SB2Vector &pt, double trk_wid,
                                  int start, int dir) const
{
   int    i;
   double d, to_lft;

   if (dir < 0) {
      for (i = start; i >= 0; i--) {
         d      = seg[i]->get_trk_dist(pt);
         to_lft = seg[i]->to_left(pt);
         if ( (d >= seg[i]->Sdst)&&(d <= seg[i]->Fdst)&&
              (to_lft >= 0.0)&&(to_lft <= trk_wid) )
            return i;
      }
   }
   else {
      for (i = start; i < num_segs; i++) {
         d      = seg[i]->get_trk_dist(pt);
         to_lft = seg[i]->to_left(pt);
         if ( (d >= seg[i]->Sdst)&&(d <= seg[i]->Fdst)&&
              (to_lft >= 0.0)&&(to_lft <= trk_wid) )
            return i;
      }
   }
   return -1;
}

int SB2PathComponent::find_trk_seg(double trk_dist, int start, int dir) const
{
   int i;

   if (dir < 0) {
      for (i = start; i >= 0; i--) {
         if ( (trk_dist >= seg[i]->Sdst)&&(trk_dist <= seg[i]->Fdst) )
            return i;
      }
   }
   else {
      for (i = start; i < num_segs; i++) {
         if ( (trk_dist >= seg[i]->Sdst)&&(trk_dist <= seg[i]->Fdst) )
            return i;
      }
   }
   return -1;
}

int SB2PathComponent::is_after(int idx, double dst,
                              const SB2PathComponent &prev,
                              int prev_idx, double prev_dst) const
{
   int i;

   // check if both points are in same seg
   if (seg[idx] == prev.seg[prev_idx])
      return (prev_dst <= dst);
   // check if a seg in this component holds the previous point
   for (i = 0; i < num_segs; i++) {
      if (seg[i] == prev.seg[prev_idx])
         return (i < idx);
   }
   // check if a seg in the prev component holds the after point
   for (i = 0; i < prev.num_segs; i++) {
      if (prev.seg[i] == seg[idx])
         return (i > prev_idx);
   }
   /* There are no overlapping segs in the 2 components, therefore
      we must try to return the correct decision here. */
   if (seg[0]->Sdst == prev.seg[prev.num_segs-1]->Fdst)
      return true;   // this component is placed right after prev
   else if (prev.seg[0]->Sdst == seg[num_segs-1]->Fdst)
      return false;  // prev component is placed right after this
   return (prev_dst <= dst);
}

int SB2PathComponent::is_after(double dst, const SB2PathComponent &prev,
                              double prev_dst) const
{
   int idx, prev_idx;

   idx = find_trk_seg(dst, 0, 1);
   if (idx < 0)
      return 0;
   prev_idx = prev.find_trk_seg(prev_dst, 0, 1);
   if (prev_idx < 0)
      return 0;
   return is_after(idx, dst, prev, prev_idx, prev_dst);
}

double SB2PathComponent::dst_between(double dst, const SB2PathComponent &prev,
                                    double prev_dst) const
{
   int idx, prev_idx, i;

   idx = find_trk_seg(dst, 0, 1);
   if (idx < 0)
      return 0.0;
   prev_idx = prev.find_trk_seg(prev_dst, 0, 1);
   if (prev_idx < 0)
      return 0.0;

   // check if both points are in same seg
   if (seg[idx] == prev.seg[prev_idx]) {
      dst -= prev_dst;
      return ((dst == 0.0) ? 0.0001 : dst);
   }
   // check if a seg in this component holds the previous point
   for (i = 0; i < num_segs; i++) {
      if (seg[i] == prev.seg[prev_idx]) {
         return dst_between(idx, dst, i, prev_dst);
      }
   }
   // check if a seg in the prev component holds the after point
   for (i = 0; i < prev.num_segs; i++) {
      if (prev.seg[i] == seg[idx]) {
         return -prev.dst_between(prev_idx, prev_dst, i, dst);
      }
   }
   /* There are no overlapping segs in the 2 components, therefore
      we must try to return the correct value here. */
   if (seg[0]->Sdst == prev.seg[prev.num_segs-1]->Fdst) {
      // this component is placed right after prev
      dst  = dst_between(idx, dst, 0, seg[0]->Sdst);
      i    = prev.num_segs - 1;
      dst += prev.dst_between(i, prev.seg[i]->Fdst, prev_idx, prev_dst);
   }
   else if (prev.seg[0]->Sdst == seg[num_segs-1]->Fdst) {
      // prev component is placed right after this
      i    = num_segs - 1;
      dst  = -dst_between(i, seg[i]->Fdst, idx, dst);
      dst -= prev.dst_between(prev_idx, prev_dst, 0, prev.seg[0]->Sdst);
   }
   return ((dst == 0.0) ? 0.0001 : dst);
}

int SB2PathComponent::valid(double trk_wid) const
{
   int    i;
   double beg, end, s, f,
          in_w  = SB2_IN_WID - SB2_PATH_ERR,
          out_w = trk_wid - SB2_OUT_WID + SB2_PATH_ERR,
          max_w = trk_wid - in_w;
      
   if (is_after(Sdst, *this, Fdst)) {
      beg = Fdst;
      end = Sdst;
   }
   else {
      beg = Sdst;
      end = Fdst;
   }

   s = beg;
   f = end + 1.0;
   for (i = 0; (i < num_segs)&&(f != end); i++) {
      // start location
      if (s == beg) {
         if ( (beg < seg[i]->Sdst)||(beg > seg[i]->Fdst) )
            continue;
         beg += 1.0;  // found first segment, stop looking
      }
      else
         s = seg[i]->Sdst;
      // finish location
      if ( (end >= seg[i]->Sdst)&&(end <= seg[i]->Fdst) )
         f = end;     // found last segment
      else
         f = seg[i]->Fdst;

      // make sure this path is valid between start and finish locations
      if ( (seg[i]->rad == 0.0)||(typeID == STRAIGHT) ) {
         if (!path_ok(i, s, f, in_w, max_w))
            return 0;
      }
      else {
         if (!path_ok(i, s, f, in_w, out_w))
            return 0;
      }
   }
   return 1;
}

double SB2PathComponent::dst_between(int idx, double dst, int prev_idx, 
                                    double prev_dst) const
{
   if (prev_idx < idx) {
      prev_dst = seg[prev_idx++]->Fdst - prev_dst;
      for (; prev_idx < idx; prev_idx++) {
         prev_dst += seg[prev_idx]->Fdst - seg[prev_idx]->Sdst;
      }
      dst -= seg[prev_idx]->Sdst;
   }
   else {
      prev_dst = seg[prev_idx--]->Sdst - prev_dst;
      for (; prev_idx > idx; prev_idx--) {
         prev_dst -= seg[prev_idx]->Fdst - seg[prev_idx]->Sdst;
      }
      dst -= seg[prev_idx]->Fdst;
   }
   dst += prev_dst;
   return ((dst == 0.0) ? 0.0001 : dst);
}
//-----END PathComponent.cpp-----//

//-----START Straight.cpp-----//
// CONSTRUCTORS
SB2Straight::SB2Straight()
   : SB2PathComponent(), l()
{
   typeID = SB2PathComponent::STRAIGHT;
   comp   = &l;
}

// METHODS
double SB2Straight::get_spd(double trk_dist, double *max) const
{
   if (max)
      *max = SB2_MAX_SPD;
   return SB2_MAX_SPD;
}

void SB2Straight::get_tan(SB2Vector &ret_pt, SB2Line &ret_norm) const
{
   ret_norm.pt.x  = ret_norm.dir.x; ret_norm.pt.y  = ret_norm.dir.y;
   ret_norm.dir.x = l.dir.x;        ret_norm.dir.y = l.dir.y;
   ret_norm.pt.rotate(MY_PI12);
}

double SB2Straight::get_path_dist(const SB2Vector &pt) const
{
   SB2Vector v(pt);

   v.sub(l.pt);
   return (fabs(v.dot(l.dir)) / len);
}

double SB2Straight::get_dist2path(const SB2Vector &pt) const
{
   return l.dist2Point(pt);
}

void SB2Straight::getPoint2(SB2Vector &ret_pt, SB2Vector &ret_tan, 
                           double path_dist) const
{
   l.getPoint(ret_pt, (path_dist/len));
   ret_tan.x = l.dir.x; ret_tan.y = l.dir.y;
}

double SB2Straight::brk_dist(double dist, double v) const
{
   return dist;
}

void SB2Straight::copy(const SB2Straight &sc)
{
   SB2PathComponent::copy(sc);

   l.pt.x  = sc.l.pt.x;
   l.pt.y  = sc.l.pt.y;
   l.dir.x = sc.l.dir.x;
   l.dir.y = sc.l.dir.y;
}

void SB2Straight::adjust(const SB2Vector &Spt, const SB2Vector &Fpt)
{
   l.pt.x  = Spt.x; l.pt.y  = Spt.y;
   l.dir.x = Fpt.x; l.dir.y = Fpt.y;
   l.dir.sub(Spt);

   len = l.dir.mag();
}

int SB2Straight::intersect(SB2Vector &ret_pt, const SB2Straight &s) const
{
   return l.intersect(ret_pt, s.l);
}

int SB2Straight::path_ok(int seg_idx, double S, double F, 
                        double min_w, double max_w) const
{
   if (len < 1.0)
      return 1;
   return seg[seg_idx]->path_ok(l, S, F, min_w, max_w);
}
//-----END Straight.cpp-----//

//-----START Turn.cpp-----//
#define APEX_BEG  0.0
#define APEX_MID  0.5
#define APEX_END  1.0
#define ADJ_ANG   (2.0*(MY_PI/180.0))
#define ADJ_RAD   0.02

const int SB2Turn::DATALEN     = 2,  // Lines per Turn in data file
          SB2Turn::BAD_ENTRY   = 1,  // Entry to turn needs adjustment
          SB2Turn::BAD_EXIT    = 2,  // Exit from turn needs adjustment
          SB2Turn::BAD_TURN    = 4,  // Turn is not completely on course
          SB2Turn::IGNORE_TURN = 8;  // Turn is not needed in path


// CONSTRUCTORS
SB2Turn::SB2Turn()
   : SB2PathComponent(), curv()
{
   typeID  = SB2PathComponent::LTURN;
   comp    = &curv;
   ts_turn = NULL;
   plen    = 0.0;
   nlen    = 0.0;
   Sang    = 0.0;
   min_spd = 1.0;
   max_spd = 1.01;
}

SB2Turn::SB2Turn(SB2TrkSeg **ts, int idx, int trk_segs,
               double min_w, double max_w, double trk_wid)
   : SB2PathComponent(), curv()
{
   int            i, j;
   SB2TrkSeg_Turn *pt, *nt;

   typeID  = SB2PathComponent::LTURN;
   comp    = &curv;
   ts_turn = (SB2TrkSeg_Turn *)ts[idx];
   plen    = 0.0;
   nlen    = 0.0;
   Sang    = 0.0;
   min_spd = 1.0;
   max_spd = 1.01;

   insert_TrkSeg(ts[idx], 0);
   for (i = 1, j = 0; i < trk_segs; i++) {
      j = (idx + trk_segs - i) % trk_segs;
      if (ts[j]->rad != 0.0)
         break;
      insert_TrkSeg(ts[j], 0);
      plen += ts[j]->len;
   }
   pt = (SB2TrkSeg_Turn *)ts[j];
   for (i = 1, j = 0; i < trk_segs; i++) {
      j = (idx + i) % trk_segs;
      if (ts[j]->rad != 0.0)
         break;
      insert_TrkSeg(ts[j], num_segs);
      nlen += ts[j]->len;
   }
   nt = (SB2TrkSeg_Turn *)ts[j];

   insert_TrkSeg(pt, 0);
   insert_TrkSeg(nt, num_segs);

   init(min_w, max_w, trk_wid);
}

// METHODS
double SB2Turn::get_spd(double trk_dist, double *max) const
{
   double pct;

   if (max) {
      pct = dst_between(trk_dist, *this, Sdst);
      if (pct > 0.0)
         pct /= len_trk;

      if (pct <= 0.0)
         *max = min_spd;
      else
         *max = (pct*max_spd) + ((1.0-pct)*min_spd);
   }
   return min_spd;
}

void SB2Turn::get_tan(SB2Vector &ret_pt, SB2Line &ret_norm) const
{
   ret_norm.pt.x  = ret_norm.dir.x; ret_norm.pt.y  = ret_norm.dir.y;
   ret_norm.dir.x = ret_pt.x;       ret_norm.dir.y = ret_pt.y;
   ret_norm.dir.sub(curv.c);
   if (fabs(ret_norm.pt.ang(ret_norm.dir)) > MY_PI12)
      ret_norm.dir.rotate(MY_PI + MY_PI12);
   else
      ret_norm.dir.rotate(MY_PI12);
   ret_norm.pt.rotate(MY_PI12);
}

double SB2Turn::get_path_dist(const SB2Vector &pt) const
{
   double   a;
   SB2Vector v(pt);

   v.sub(curv.c);
   a = v.ang();

   if ( (typeID == SB2PathComponent::LTURN)&&(a < Sang) )
      a += MY_PI21;
   else if ( (typeID == SB2PathComponent::RTURN)&&(a > Sang) )
      a -= MY_PI21;
   a = fabs(a - Sang);
   if (a > (MY_PI21 - 0.05))
      a -= MY_PI21;
   return (curv.R * a);
}

double SB2Turn::get_dist2path(const SB2Vector &pt) const
{
   if (typeID == SB2PathComponent::LTURN)
      return curv.dist2Point(pt);
   return -curv.dist2Point(pt);
}

void SB2Turn::getPoint2(SB2Vector &ret_pt, SB2Vector &ret_tan, 
                       double path_dist) const
{
   if (typeID == SB2PathComponent::LTURN) {
      curv.getPoint(ret_pt, (Sang + (path_dist/curv.R)));
      ret_tan.x = ret_pt.x - curv.c.x; 
      ret_tan.y = ret_pt.y - curv.c.y;
      ret_tan.rotate(MY_PI12);
   }
   else {
      curv.getPoint(ret_pt, (Sang - (path_dist/curv.R)));
      ret_tan.x = ret_pt.x - curv.c.x; 
      ret_tan.y = ret_pt.y - curv.c.y;
      ret_tan.rotate(-MY_PI12);
   }
}

double SB2Turn::brk_dist(double dist, double v) const
{
   v *= v;
   v /= curv.R * MAX_ACCEL;
   v *= v;
   v  = 1.0 - v;
   if (v < 0.5625)
      dist *= 0.9 * 0.75;
   else
      dist *= 0.9 * sqrt(v);
   return dist;
}

void SB2Turn::init(double min_w, double max_w, double trk_wid)
{
   int            i;
   double         apex, min_len;
   SB2TrkSeg_Turn *pt = NULL,
                 *nt = NULL;

   i = get_ts_turn_idx() - 1;
   while ( (i >= 0)&&(seg[i]->rad == 0.0) ) {
      i--;
   }
   if (i >= 0)
      pt = (SB2TrkSeg_Turn *)seg[i];
      
   i = get_ts_turn_idx() + 1;
   while ( (i < num_segs)&&(seg[i]->rad == 0.0) ) {
      i++;
   }
   if (i < num_segs) {
      nt = (SB2TrkSeg_Turn *)seg[i];
      if (pt == NULL)
         pt = nt;
   }
   else if (pt != NULL)
      nt = pt;

   min_len = 3.5 * trk_wid;
   apex    = APEX_MID;
   if ( (plen < min_len)&&(!pt->like_straight())&&
        (nlen < min_len)&&(!nt->like_straight()) ) {
      // no straights between the prev turn, this turn, and next turn
      if ( (ts_turn->same_dir(*pt))&&(ts_turn->same_dir(*nt)) ) {
         // In the middle of 3 or more corners in same direction
         if (ts_turn->like_straight())  // this turn can be considered straight
            min_w = max_w - 5.0;
         else if ( (plen == 0.0)&&(nlen == 0.0) )
            max_w = min_w + 1.0;
         else {
            if (plen == 0.0)
               apex = -1.0;
            else if (nlen == 0.0)
               apex =  2.0;
            if ( (plen < nlen)||(nlen == 0.0) )
               max_w = (0.5 + 0.5*(plen/min_len))*(max_w - min_w);
            else
               max_w = (0.5 + 0.5*(nlen/min_len))*(max_w - min_w);
            max_w += min_w;
         }
      }
      else if (ts_turn->same_dir(*pt)) {
         // 2 turns in same dir, followed by 1 in opposite dir
         if (ts_turn->like_straight())  // this turn can be considered straight
            apex = APEX_END;
         else if (plen == 0.0)
            apex = -1.0;
         else
            apex = get_apex(plen, min_w, max_w);
      }
      else if (ts_turn->same_dir(*nt)) {
         // 2 turns in same dir, preceeded by 1 in opposite dir
         if (ts_turn->like_straight())  // this turn can be considered straight
            apex = APEX_BEG;
         else if (nlen == 0.0)
            apex = 2.0;
         else
            apex = 1.0 - get_apex(nlen, min_w, max_w);
      }
      else {
         // prev and next turns in same dir, this turn in opposite dir
         if (ts_turn->like_straight())  // this turn can be considered straight
            max_w = min_w + 5.0;
      }
   }
   else if (ts_turn->like_straight()) {
      // this turn can be considered straight
      if (pt->same_dir(*nt)) {
         if (ts_turn->same_dir(*pt))
            min_w = max_w - 5.0;
         else
            max_w = min_w + 5.0;
      }
      else {
         if (ts_turn->same_dir(*pt))
            apex = APEX_END;
         else
            apex = APEX_BEG;
      }
   }
   else if ( (plen < min_len)&&(!pt->like_straight()) ) {
      // no straight between the prev turn, and this turn
      // there is a straight between this turn, and the next
      if (ts_turn->same_dir(*pt)) { // this turn in same dir as prev
         if (plen == 0.0)
            apex = -1.0;
         else
            apex = get_apex(plen, min_w, max_w);
      }
   }
   else if ( (nlen < min_len)&&(!nt->like_straight()) ) {
      // there is a straight between the prev turn, and this turn
      // no straight between this turn, and the next
      if (ts_turn->same_dir(*nt)) { // this turn in same dir as next
         if (nlen == 0.0)
            apex = 2.0;
         else
            apex = 1.0 - get_apex(nlen, min_w, max_w);
      }
   }
   set(apex, min_w, max_w, trk_wid);
}

void SB2Turn::copy(const SB2Turn &tc, int copy_segs)
{
   if (copy_segs)
      SB2PathComponent::copy(tc);

   curv.c.x  = tc.curv.c.x;
   curv.c.y  = tc.curv.c.y;
   curv.R    = tc.curv.R;
   ts_turn   = tc.ts_turn;
   apex_dist = tc.apex_dist;
   plen      = tc.plen;
   nlen      = tc.nlen;
}

void SB2Turn::copy(const SB2Turn &tc, int copy_segs, int trim_tail,
                  int extra_segs)
{
   int end;

   copy(tc, copy_segs);
   if (copy_segs) {
      end = get_ts_turn_idx();
      if (trim_tail) {
         end += extra_segs;
         if (end >= num_segs)
            end = num_segs - 1;
         if (is_after(Fdst, *this, seg[end]->Fdst))
            Fdst = seg[end]->Fdst;
         num_segs = end + 1;
      }
      else {
         end -= extra_segs;
         if (end < 0)
            end = 0;
         if (is_after(seg[end]->Sdst, *this, Sdst))
            Sdst = seg[end]->Sdst;
         while (end > 0) {
            remove_TrkSeg(0);
            end--;
         }
      }
   }
   else {
      if (trim_tail)
         Sdst = tc.Sdst;
      else
         Fdst = tc.Fdst;
   }
}

void SB2Turn::adjust(SB2Turn &pt, SB2Turn &nt, int adj_val,
                    double min_w, double max_w, double trk_wid)
{
   double p_dst, n_dst, weight;

   if (adj_val == 0)
      return;
   else if ((adj_val&BAD_TURN) != 0) {
      fix_bad_turn(min_w, max_w, trk_wid);
      return;
   }

   // find the error at the entry and exit of this turn
   p_dst = get_err(pt, 0, trk_wid);
   n_dst = nt.get_err(*this, 1, trk_wid);

   if (ts_turn->like_straight()) {
      // adjust these turns here
      if ( (p_dst < 0.0)&&(n_dst < 0.0) )
         moveTurn( 2.0, min_w, max_w);  // move turn closer to inside wall
      else if ( (p_dst > 2.0)&&(n_dst > 2.0) )
         moveTurn(-2.0, min_w, max_w);  // move turn closer to outside wall
      else if (p_dst < 0.0)
         moveApex(-ADJ_ANG, min_w, max_w, trk_wid);  // move apex closer to entry
      else if (n_dst < 0.0)
         moveApex( ADJ_ANG, min_w, max_w, trk_wid);  // move apex closer to exit
      return;
   }
   
   // determine how much adjustment this turn will allow
   if (adj_val == BAD_ENTRY)
      weight = get_weight(pt);
   else if (adj_val == BAD_EXIT)
      weight = get_weight(nt);
   else
      weight = 0.5 * (get_weight(pt) + get_weight(nt));

   if (adj_val == BAD_ENTRY) {
      if (p_dst < 0.0) {
         // move entry of turn closer to inside wall
         if ( (ts_turn->Sdst < apex_dist)&&(apex_dist <= ts_turn->Fdst) ) {
            // move apex to entry
            moveApex((-ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
         else if (apex_dist > ts_turn->Fdst) {
            // reduce turn radius
            weight *= -ADJ_RAD * curv.R;
            min_w  -= 0.1;
            moveTurn(weight, min_w, max_w);
            adjRadius(weight, min_w, max_w);
         }
      }
      else if (p_dst > 2.0) {
         // move entry of turn closer to outside wall
         if ( (ts_turn->Sdst <= apex_dist)&&(apex_dist < ts_turn->Fdst) ) {
            // move apex to exit
            moveApex(( ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
      }
   }
   else if (adj_val == BAD_EXIT) {
      if (n_dst < 0.0) {
         // move exit of turn closer to inside wall
         if ( (ts_turn->Sdst <= apex_dist)&&(apex_dist < ts_turn->Fdst) ) {
            // move apex to exit
            moveApex(( ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
         else if (ts_turn->Sdst > apex_dist) {
            // reduce turn radius
            weight *= -ADJ_RAD * curv.R;
            min_w  -= 0.1;
            moveTurn(weight, min_w, max_w);
            adjRadius(weight, min_w, max_w);
         }
      }
      else if (n_dst > 2.0) {
         // move exit of turn closer to outside wall
         if ( (ts_turn->Sdst < apex_dist)&&(apex_dist <= ts_turn->Fdst) ) {
            // move apex to entry
            moveApex((-ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
      }
   }
   else if (p_dst < 0.0) {
      // move entry of turn closer to inside wall
      if (n_dst > 2.0) {
         // move exit of turn closer to outside wall
         if (ts_turn->Sdst < apex_dist) {
            // move apex to entry
            moveApex((-ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
      }
      else {
         // reduce turn radius
         weight *= -ADJ_RAD * curv.R;
         moveTurn(weight, min_w, max_w);
         adjRadius(weight, min_w, max_w);
      }
   }
   else if ( (p_dst > 2.0)&&(n_dst < 0.0)&&
             (apex_dist < ts_turn->Fdst) ) {
      // move entry of turn closer to outside wall
      // move exit of turn closer to inside wall
      // move apex to exit
      moveApex(( ADJ_ANG*weight), min_w, max_w, trk_wid);
   }
}

int SB2Turn::makeStraight(SB2Straight &ret, SB2Turn &pt, double trk_wid)
{
   int    i, j, k, k1, ts1, ts2;
   double d1, d2;
   SB2Line s;

   ts1 = get_ts_turn_idx();
   ts2 = pt.get_ts_turn_idx();

   // first check if straight is needed
   s.dir.x = pt.curv.c.x - curv.c.x;
   s.dir.y = pt.curv.c.y - curv.c.y;
   d1      = s.dir.mag();
      
   if (ts_turn->same_dir(*(pt.ts_turn))) {
      j   = SB2Circle::TAN_1;
      d1 -= fabs(curv.R - pt.curv.R);
   }
   else {
      j   = SB2Circle::TAN_3;
      d1 -= curv.R + pt.curv.R;
   }

   if (fabs(d1) < 2.5) {
      curv.getPoint(s.pt, s.dir.ang());
      k = find_trk_seg(s.pt, trk_wid, ts1, -1);
      if (k >= 0) {
         // do not need a straight here
         ret.adjust(s.pt, s.pt);
         ret.num_segs = 0;
         ret.insert_TrkSeg(seg[k], 0);

         // start and finish locations
         d2 = seg[k]->get_trk_dist(s.pt);
         ret.Sdst = pt.Fdst = d2;
         ret.Fdst = Sdst    = d2;
         return 1;
      }
   }

   for (i = 0; i < 2; i++) {
      j += i;
      if (!curv.tangent(s, pt.curv, j))
         continue;
      s.dir.add(s.pt);

      k = find_trk_seg(s.pt, trk_wid, ts1, -1);
      if (k < 0) {
         k = find_trk_seg(s.pt, trk_wid, ts1+1, 1);
         if (k < 0)
            continue;
      }
      k1 = k;
      d1 = seg[k]->get_trk_dist(s.pt);
      if ( (ts_turn->len > (0.7*MY_PI))&&
           (!is_after(ts1, ts_turn->Fdst, *this, k, d1)) )
         continue;

      k = pt.find_trk_seg(s.dir, trk_wid, ts2, 1);
      if (k < 0) {
         k = pt.find_trk_seg(s.dir, trk_wid, ts2-1, -1);
         if (k < 0)
            continue;
      }
      d2 = pt.seg[k]->get_trk_dist(s.dir);
      if ( (pt.ts_turn->len > (0.7*MY_PI))&&
           (!pt.is_after(k, d2, pt, ts2, pt.ts_turn->Sdst)) )
         continue;

      if (is_after(k1, d1, pt, k, d2)) {
         ts1 = k1;
         ts2 = k;
         break;
      }
   }

   if (i < 2) {
      ret.adjust(s.dir, s.pt);

      // insert TrkSeg into Straight component
      ret.num_segs = 0;
      for (k = pt.num_segs-1; k >= ts2; k--) {
         ret.insert_TrkSeg(pt.seg[k], 0);
      }
      for (k = 0; k <= ts1; k++) {
         ret.insert_TrkSeg(seg[k], ret.num_segs);
      }

      // start and finish locations
      ret.Sdst = pt.Fdst = d2;
      ret.Fdst = Sdst    = d1;
      return 1;
   }
   return 0;
}

void SB2Turn::set_spd_len(int data_loaded)
{
   SB2Vector p;
   SB2Line   n1, n2;

   // set the length of this component
   getPoint(p, n1, Sdst, 1);
   Sang = curv.ang(p);
   getPoint(p, n2, (Sdst+2.0), 1);
   if (n1.dir.ang(n2.dir) < 0.0)
      typeID = SB2PathComponent::RTURN;
   else
      typeID = SB2PathComponent::LTURN;
   getPoint(p, n2, Fdst, 1);
   len = n1.dir.ang(n2.dir);
   if ( (typeID == SB2PathComponent::LTURN)&&(len < 0.0) )
      len += MY_PI21;
   else if ( (typeID == SB2PathComponent::RTURN)&&(len > 0.0) )
      len -= MY_PI21;
   p.x = fabs(len);
   len = fabs(len) * curv.R;

   // set the speed for this turn
#ifndef SB2_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 0:
      min_spd *= 0.945;
      max_spd *= 0.945;
      if (min_spd > 1.0)
         min_spd = 1.0;
      if (max_spd > 1.05)
         max_spd = 1.05;
      break;
   case 2:
      min_spd *= 0.99;
      max_spd *= 0.985;
   };

#ifndef SB2_DLL
   if ( (data_loaded)||(args.m_iSurface == 0) ) {
#else
   if ( (data_loaded)||(args->m_iSurface == 0) ) {
#endif
      min_spd *= SCALE_CORNSPD * SB2_crn_spd(curv.R);
      max_spd *= SCALE_CORNSPD * SB2_crn_spd(curv.R);
   }
   else {
      if (p.x < 0.175) {
         min_spd = max_spd = SB2_MAX_SPD;
      }
      else {
         min_spd *= SCALE_CORNSPD * SB2_crn_spd(curv.R);
         max_spd *= SCALE_CORNSPD * SB2_crn_spd(curv.R);
         if ( (p.x < 0.37)&&(len < 150.0) ) {
            min_spd *= ((p.x-0.37)*(1.6-1.1)/(0.175-0.37)) + 1.1;
            max_spd *= ((p.x-0.37)*(1.6-1.1)/(0.175-0.37)) + 1.1;
         }
      }
   }
   if (min_spd > SB2_MAX_SPD)
      min_spd = SB2_MAX_SPD;
   if (max_spd > SB2_MAX_SPD)
      max_spd = SB2_MAX_SPD;
}

int SB2Turn::makePass(SB2Straight *p_s, SB2Turn *p_t, SB2Turn &nt,
                     SB2Turn &last, double min_w, double max_w, 
                     double trk_wid)
{
   int      retval;
   double   dist;
   SB2Vector p;
   SB2Line   n;

   if (nt.ts_turn->like_straight())
      return 0;  // turn too fast for passing

   dist  = nt.curv.R - nt.curv.c.dist2Point(nt.ts_turn->ls.c);
   dist -= fabs(nt.ts_turn->rad);
   if (dist > max_w)
      return 0;  // apex point too far from inner edge

   // attempt to create a passing path
   p_t->copy(nt, 1);
   p_t->set(nt.get_apex(), min_w, max_w, trk_wid);
   dist      = Fdst;
   retval    = p_t->makeStraight(*p_s, *this, trk_wid);
   Fdst      = dist;
   p_t->Fdst = nt.apex_dist;

   p_t->min_spd = 1.0;
   p_t->max_spd = 1.0;
   p_t->set_spd_len(1);
#ifndef SB2_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 0:
      if (p_t->min_spd <= (SB2_MAX_SPD-2.0)) {
         p_t->min_spd += 1.0;
         p_t->max_spd += 2.0;
      }
      break;
   case 1:
      if (p_t->min_spd <= (SB2_MAX_SPD-3.0)) {
         p_t->min_spd += 1.5;
         p_t->max_spd += 3.0;
      }
      break;
   case 2:
      if (p_t->min_spd <= (SB2_MAX_SPD-5.0)) {
         p_t->min_spd += 3.0;
         p_t->max_spd += 5.0;
      }
   };

   // check if path is valid
   retval = ( (retval)&&(p_s->len > 20.0)&&(p_t->len > 20.0)&&
              (p_s->valid(trk_wid))&&
              (is_after(p_s->Sdst, *this, Sdst))&&
              (p_t->is_after(p_t->Fdst, *p_t, p_t->Sdst)) );

   if (retval) {
      // make sure passing path isn't too close to racing path
      dist = nt.to_side(p, n, nt.Sdst);
      if (p_s->is_after(p_s->Fdst, nt, nt.Sdst))
         dist -= p_s->to_side(p, n, nt.Sdst);
      else
         dist -= p_t->to_side(p, n, nt.Sdst);
      if (nt.ts_turn->rad < 0.0)
         dist = -dist;
      retval = (dist >= CARLEN);
   }

   if ( (retval)&&(last.pass != NULL) ) {
      // make sure there is no overlap
      if (is_after(last.pass[1]->Fdst, *this, p_s->Sdst)) {
         // there is an overlap, so decide which one to keep
         if ((nt.ts_turn->len*nt.get_apex()) > (ts_turn->len*get_apex())) {
            // drop the last one
            delete last.pass[0];
            delete last.pass[1];
            delete [] last.pass;
            last.pass = NULL;
         }
         else
            retval = 0;  // drop this one
      }
   }

   pre_len = nt.post_len = 0.0;
   if (retval) {
      // store path
      if (pass != NULL) {
         // remove these
         delete pass[0];
         delete pass[1];
      }
      else
         pass = new SB2PathComponent*[2];
      p_s->trim();
      p_t->trim();
      pass[0] = p_s;
      pass[1] = p_t;

      p_s->len_trk = p_s->dst_between(p_s->Fdst, *p_s, p_s->Sdst);
      p_t->len_trk = p_t->dst_between(p_t->Fdst, *p_t, p_t->Sdst);

      getPoint(p, n, p_s->Sdst);
      pre_len = get_path_dist(p);
      
      nt.getPoint(p, n, p_t->Fdst);
      nt.post_len = nt.len - nt.get_path_dist(p);
   }
   else if (pass != NULL) {
      // clean up components
      p_t->num_segs = 0;
      p_s->num_segs = 0;

      // remove these
      delete pass[0];
      delete pass[1];
      delete [] pass;
      pass = NULL;
   }
   return retval;
}

void SB2Turn::save(FILE *out, SB2TrkSeg **ts, double rotate) const
{
   int      i;
   SB2Vector cen(curv.c);

   cen.sub(ts_turn->ls.c);
   cen.rotate(rotate);

   // SAVE LINE 1
   i = 0;
   while (ts[i] != ts_turn) {
      i++;
   }
   fprintf(out, "%d ", i);
   i = 0;
   while (ts[i] != seg[0]) {
      i++;
   }
   fprintf(out, "%d %d ", i, num_segs);

   fprintf(out, "%10.3lf %10.3lf %10.3lf\n", Sdst, apex_dist, Fdst);

   // SAVE LINE 2
   fprintf(out, "%11.3lf %11.3lf %10.3lf ", cen.x, cen.y, curv.R);

   cen.x = min_spd / (SCALE_CORNSPD * SB2_crn_spd(curv.R));
   cen.y = max_spd / (SCALE_CORNSPD * SB2_crn_spd(curv.R));
   fprintf(out, "%5.2lf %5.2lf\n", cen.x, cen.y);
}

int SB2Turn::load(FILE *in, SB2TrkSeg **ts, int trk_segs, double rotate)
{
   const char *ws = " \t\n\r";
   int  i, j, k;
   char buf[128];

   // LOAD LINE 1
   if (!fgets(buf, 128, in))
      return 0;
   k       = atoi(strtok(buf, ws));
   if ( (k < 0)||(k >= trk_segs) )
      return 0;
   ts_turn = (SB2TrkSeg_Turn *)ts[k];
   i       = atoi(strtok(NULL, ws));
   j       = atoi(strtok(NULL, ws)) - 1;
   if ( (i < 0)||(i >= trk_segs)||(j < 0)||(j >= trk_segs) )
      return 0;
   for (; j >= 0; j--) {
      insert_TrkSeg(ts[(i+j)%trk_segs], 0);
   }

   plen = nlen = 0.0;
   for (i = 1, j = 0; i < trk_segs; i++) {
      j = (k + trk_segs - i) % trk_segs;
      if (ts[j]->rad != 0.0)
         break;
      plen += ts[j]->len;
   }
   for (i = 1, j = 0; i < trk_segs; i++) {
      j = (k + i) % trk_segs;
      if (ts[j]->rad != 0.0)
         break;
      nlen += ts[j]->len;
   }

   Sdst      = atof(strtok(NULL, ws));
   apex_dist = atof(strtok(NULL, ws));
   Fdst      = atof(strtok(NULL, ws));

   // LOAD LINE 2
   if (!fgets(buf, 128, in))
      return 0;
   curv.c.x = atof(strtok(buf, ws));
   curv.c.y = atof(strtok(NULL, ws));
   curv.R   = atof(strtok(NULL, ws));
   curv.c.rotate(rotate);
   curv.c.add(ts_turn->ls.c);

   min_spd = atof(strtok(NULL, ws));
   max_spd = atof(strtok(NULL, ws));
   return 1;
}

// PRIVATE METHODS
int SB2Turn::get_ts_turn_idx() const
{
   int i;

   for (i = 0; i < num_segs; i++) {
      if (seg[i] == ts_turn)
         return i;
   }
   return 0;
}

double SB2Turn::get_apex(double dst, double min_w, double max_w) const
{
   double A, B, C, dA,
          dB = 0.0,
          dC = cos(ts_turn->len),
          b  = dst / tan(0.5*ts_turn->len),
          d  = dst / sin(0.5*ts_turn->len);
   int    i  = 0;

   d *= d;
   B  = 0.0;
   do {
      i++;
      B += 0.5 * dB;
      A  = sqrt((B*B) + (2.0*B*b) + d);
      C  = dC * B;
      dA = (B + b) / A;
      dB = (b + C + max_w - A - min_w) / (dA - dC);
   } while ( (i < 50)&&(fabs(dB) > 1.0) );

   if (fabs(dB) > 1.0)
      return 0.5;

   B = atan(dst / (b+B));
   if (B < 0.0)
      B += MY_PI;

   B /= ts_turn->len;
   if (B < 0.0)
      B = 0.0;
   else if ( (0.5 < B)&&(B <= 1.0) )
      B = 1.0 - B;
   return B;
}

double SB2Turn::get_apex() const
{
   double apex;

   apex = ts_turn->Fdst - ts_turn->Sdst;
   if (apex != 0.0)
      apex = (apex_dist-ts_turn->Sdst) / apex;
   return apex;
}

void SB2Turn::moveApex(double amount, double min_w, double max_w,
                      double trk_wid)
{
   double apex  = apex_dist - ts_turn->Sdst,
          old_R = curv.R;

   apex = get_apex() + (amount / ts_turn->len);
   if (apex > APEX_END)
      apex = APEX_END;
   else if (apex < APEX_BEG)
      apex = APEX_BEG;

   set(apex, min_w, max_w, trk_wid);
   if (old_R < curv.R) {
      apex   = old_R - curv.R;
      curv.R = old_R;
      moveTurn(apex, min_w, max_w);
   }
}

double SB2Turn::adjRadius(double amount, double min_w, double max_w)
{
   double Ad = curv.c.dist2Point(ts_turn->ls.c);

   if ((curv.R-Ad+amount) < (fabs(ts_turn->rad)+min_w))
      amount = fabs(ts_turn->rad) + min_w - curv.R + Ad;
   else if ((curv.R-Ad+amount) > (fabs(ts_turn->rad)+max_w))
      amount = fabs(ts_turn->rad) + max_w - curv.R + Ad;
   curv.R += amount;
   return amount;
}

double SB2Turn::moveTurn(double amount, double min_w, double max_w)
{
	double Ad;
	SB2Line l;

	l.pt.add(curv.c);
	l.dir.add(curv.c);
	l.dir.sub(ts_turn->ls.c);
	Ad = l.dir.mag();
	if (Ad == 0.0)
		return 0.0;
	else if ((curv.R-Ad-amount) < (fabs(ts_turn->rad)+min_w))
		amount = curv.R - Ad - fabs(ts_turn->rad) - min_w;
	else if ((curv.R-Ad-amount) > (fabs(ts_turn->rad)+max_w))
		amount = curv.R - Ad - fabs(ts_turn->rad) - max_w;
	if ((Ad+amount) < 0.5)
		amount = 0.5 - Ad;
	l.dir.normalize();
	l.getPoint(curv.c, amount);
	return amount;
}

void SB2Turn::setLimits(double trk_wid)
{
	int            i;
	double         dst;
	SB2Line         a;
	SB2Circle      *cir;
	SB2TrkSeg_Turn *trn;
		
	a.dir.x = ts_turn->ls.c.x - curv.c.x;
	a.dir.y = ts_turn->ls.c.y - curv.c.y;
	dst     = a.dir.mag();
	Sdst    = dst * sin(get_apex()*ts_turn->len);
	Fdst    = dst * sin((1.0 - get_apex())*ts_turn->len);

	if (Sdst <= plen) {
		if (Sdst <= 0.0)
			Sdst = ts_turn->Sdst;
		else {
			dst = Sdst;
			i   = get_ts_turn_idx() - 1;
			for (; (i >= 0)&&(dst > 0.0); i--) {
				dst -= seg[i]->Fdst - seg[i]->Sdst;
			}
			Sdst = seg[i+1]->Sdst - dst;
		}
	}
	else {
		if ( (ts_turn->like_straight())||(get_apex() < 0.01) )
			Sdst = ts_turn->Sdst;
		else {
			dst = Sdst;
			i   = get_ts_turn_idx() - 1;
			for (; (i >= 0)&&(dst > 0.0); i--) {
				if (seg[i]->rad == 0.0) {
					seg[i]->get_norm(a, seg[i]->Sdst);
					cir = &curv;
				}
				else {
					trn     = (SB2TrkSeg_Turn *)seg[i];
					a.dir.x = trn->ls.c.x - curv.c.x;
					a.dir.y = trn->ls.c.y - curv.c.y;
					cir     = &(trn->ls);
				}
			
				cir->getPoint(a.pt, a.dir.ang());
				Sdst = seg[i]->get_trk_dist(a.pt);
				if ( (Sdst > (seg[i]->Sdst-2.5))&&
				     (Sdst < (seg[i]->Fdst+2.5)) )
					break;
				cir->getPoint(a.pt, (a.dir.ang()+MY_PI));
				Sdst = seg[i]->get_trk_dist(a.pt);
				if ( (Sdst > (seg[i]->Sdst-2.5))&&
				     (Sdst < (seg[i]->Fdst+2.5)) )
					break;
            if (!seg[i]->path_ok(curv, seg[i]->Sdst, seg[i]->Fdst,
                                 0.0, trk_wid)) {
               Sdst = seg[i]->Fdst;
               break;
            }
				dst -= seg[i]->Fdst - seg[i]->Sdst;
			}
			if ( (dst <= 0.0)||(i == -1) )
				Sdst = seg[i+1]->Fdst;
         else if (!seg[i]->path_ok(curv, Sdst, seg[i]->Fdst,
                                   0.0, trk_wid)) {
            do {
               Sdst *= 0.5;
               Sdst += 0.5 * seg[i]->Fdst;
               if ((seg[i]->Fdst-Sdst) < 5.0)
                  break;
            } while (!seg[i]->path_ok(curv, Sdst, seg[i]->Fdst,
                                      0.0, trk_wid));
         }
		}
	}

	if (Fdst <= nlen) {
		if (Fdst <= 0.0)
			Fdst = ts_turn->Fdst;
		else {
			dst = Fdst;
			i   = get_ts_turn_idx() + 1;
			for (; (i < num_segs)&&(dst > 0.0); i++) {
				dst -= seg[i]->Fdst - seg[i]->Sdst;
			}
			Fdst = seg[i-1]->Fdst + dst;
		}
	}
	else {
		if ( (ts_turn->like_straight())||(get_apex() > 0.99) )
			Fdst = ts_turn->Fdst;
		else {
			dst = Fdst;
			i   = get_ts_turn_idx() + 1;
			for (; (i < num_segs)&&(dst > 0.0); i++) {
				if (seg[i]->rad == 0.0) {
					seg[i]->get_norm(a, seg[i]->Fdst);
					cir = &curv;
				}
				else {
					trn     = (SB2TrkSeg_Turn *)seg[i];
					a.dir.x = trn->ls.c.x - curv.c.x;
					a.dir.y = trn->ls.c.y - curv.c.y;
					cir     = &(trn->ls);
				}

				cir->getPoint(a.pt, (a.dir.ang()+MY_PI));
				Fdst = seg[i]->get_trk_dist(a.pt);
				if ( (Fdst > (seg[i]->Sdst-2.5))&&
				     (Fdst < (seg[i]->Fdst+2.5)) )
					break;
				cir->getPoint(a.pt, a.dir.ang());
				Fdst = seg[i]->get_trk_dist(a.pt);
				if ( (Fdst > (seg[i]->Sdst-2.5))&&
				     (Fdst < (seg[i]->Fdst+2.5)) )
					break;
            if (!seg[i]->path_ok(curv, seg[i]->Sdst, seg[i]->Fdst,
                                 0.0, trk_wid)) {
               Fdst = seg[i]->Sdst;
               break;
            }
				dst -= seg[i]->Fdst - seg[i]->Sdst;
			}
			if ( (dst <= 0.0)||(i == num_segs) )
				Fdst = seg[i-1]->Sdst;
         else if (!seg[i]->path_ok(curv, seg[i]->Sdst, Fdst,
                                   0.0, trk_wid)) {
            do {
               Fdst *= 0.5;
               Fdst += 0.5 * seg[i]->Sdst;
               if ((Fdst-seg[i]->Sdst) < 5.0)
                  break;
            } while (!seg[i]->path_ok(curv, seg[i]->Sdst, Fdst,
                                      0.0, trk_wid));
         }
		}
	}
}

void SB2Turn::set(double apex, double min_w, double max_w, double trk_wid)
{
	double Ad;
	SB2Line a;

	if (apex == APEX_BEG)
		apex_dist = ts_turn->Sdst;
	else if (apex == APEX_END)
		apex_dist = ts_turn->Fdst;
	else if ( (APEX_BEG < apex)&&(apex < APEX_END) )
		apex_dist = (apex*ts_turn->Fdst) + ((1.0-apex)*ts_turn->Sdst);
	else {
		if (apex < APEX_BEG)
			apex_dist = ts_turn->Sdst - 2.0;
		else
			apex_dist = ts_turn->Fdst + 2.0;
		apex   = get_apex();
		min_w -= 0.1;
	}
	ts_turn->get_norm(a, apex_dist);
	if (ts_turn->rad > 0.0)
		a.dir.rotate(MY_PI);

   if (apex < APEX_MID) {
		apex *= ts_turn->len;
		Ad    = (max_w - min_w) / (1.0 - cos(ts_turn->len - apex));
	}
	else {
		apex *= ts_turn->len;
		Ad    = (max_w - min_w) / (1.0 - cos(apex));
	}

	a.pt.x = ts_turn->ls.c.x;
	a.pt.y = ts_turn->ls.c.y;
	a.getPoint(curv.c, Ad);
	curv.R = Ad + fabs(ts_turn->rad) + min_w;
		
	setLimits(trk_wid);
}

double SB2Turn::get_weight(const SB2Turn &ot) const
{
   double weight;

   weight = curv.R / (curv.R + ot.curv.R);
   if ( (0.15 < weight)&&(weight < 0.85) ) {
      double t = fabs(ts_turn->rad),
             o = fabs(ot.ts_turn->rad);
                
      weight  = 0.7 * (t / (t + o));
      t       = curv.R    / t;
      o       = ot.curv.R / o;
      weight += 0.3 * (o / (t + o));
   }
   return weight;
}

double SB2Turn::get_err(SB2Turn &pt, int get_pt_err, double trk_wid)
{
   double err;
      
   if (!ts_turn->same_dir(*(pt.ts_turn))) {
      err  = curv.c.dist2Point(pt.curv.c);
      err -= curv.R + pt.curv.R;
   }
   else {
      double org_Sdst    = Sdst,
             org_Fdst    = Fdst,
             org_pt_Sdst = pt.Sdst,
             org_pt_Fdst = pt.Fdst;

      pt.setLimits(trk_wid);
      setLimits(trk_wid);
      if (pt.is_after(pt.Fdst, *this, Sdst)) {
         err  = pt.dst_between(pt.Fdst, *this, Sdst);
         err /= pt.curv.R + curv.R;
         if (get_pt_err)
            err = -pt.curv.R;
         else
            err = -curv.R;
      }
      else {
         SB2Vector p;
         SB2Line   n;
            
         if (ts_turn->rad < 0.0)
            err = to_side(p, n, Sdst) - pt.to_side(p, n, pt.Fdst);
         else
            err = pt.to_side(p, n, pt.Fdst) - to_side(p, n, Sdst);
         if (get_pt_err)
            err = -err;
      }
      Sdst    = org_Sdst;
      Fdst    = org_Fdst;
      pt.Sdst = org_pt_Sdst;
      pt.Fdst = org_pt_Fdst;
   }
   return err;
}

void SB2Turn::fix_bad_turn(double min_w, double max_w, double trk_wid)
{
   int      i, j, end;
   double   w, start, seg_len,
            amount = 0.0;
   SB2Vector p;
   SB2Line   n;
      
   // check apex
   p.add(ts_turn->ls.c);
   p.sub(curv.c);
   w = curv.R - p.mag() - fabs(ts_turn->rad);
   if (w < (min_w-SB2_PATH_ERR))
      moveTurn((w-min_w), min_w, max_w);
   else if (w > (max_w+SB2_PATH_ERR))
      moveTurn((w-max_w), min_w, max_w);

   // check entry
   w = to_side(p, n, Sdst);
   if (ts_turn->rad < 0.0)
      w = (ts_turn->ls.R + ts_turn->rad) - w;
   if (w < (min_w+SB2_PATH_ERR)) {
      // increase radius
      if (w < (min_w-1.0))
         amount = min_w - w;
      else 
         amount = 1.0;
   }
   else {
      // decrease radius
      if (w > (max_w+2.0))
         amount = max_w - w;
      else
         amount = -2.0;
   }

   // check the rest of the turn
   i   = find_trk_seg(Sdst, 0, 1);
   end = find_trk_seg(Fdst, (num_segs-1), -1) + 1;
   for (; i < end; i++) {
      if ( (seg[i]->Sdst <= Sdst)&&(Sdst <= seg[i]->Fdst) )
         start = Sdst;
      else
         start = seg[i]->Sdst;
      if ( (seg[i]->Sdst <= Fdst)&&(Fdst <= seg[i]->Fdst) )
         seg_len = Fdst - start;
      else
         seg_len = seg[i]->Fdst - start;
      for (j = 1; j <= 5; j++) {
         w = to_side(p, n, (start + (0.2*(double)j*seg_len)));
         if (ts_turn->rad < 0.0)
            w = (ts_turn->ls.R + ts_turn->rad) - w;
         if (w < (min_w+SB2_PATH_ERR)) {
            // increase radius
            if (w < (min_w-1.0))
               w = min_w - w;
            else 
               w = 1.0;
         }
         else {
            // decrease radius
            if (w > (max_w+2.0))
               w = max_w - w;
            else
               w = -2.0;
         }

         // update the amount of adjustment needed
         if ( (amount > 0.0)&&(w > 0.0) ) {
            if (w > amount)
               amount = w;  // use larger value
         }
         else if (w < amount)
            amount = w;  // use smaller value
      }
   }

   // fix the turn
   curv.R += amount;
   moveTurn(0.0, min_w, max_w);
   setLimits(trk_wid);
}

int SB2Turn::path_ok(int seg_idx, double S, double F, 
                    double min_w, double max_w) const
{
   return seg[seg_idx]->path_ok(curv, S, F, min_w, max_w);
}

void SB2Turn::move_to_outside(SB2Line &ret_norm) const
{
   if (ts_turn->rad > 0.0) {
      // move point from inside wall to outside wall
      ret_norm.movePoint(1000.0);
   }
}

#undef APEX_BEG
#undef APEX_MID
#undef APEX_END
#undef ADJ_ANG
#undef ADJ_RAD
//-----END Turn.cpp-----//

//-----START TrkData.cpp-----//
const int SB2TrkData::NORMAL = 0,
          SB2TrkData::OFF    = 1,
          SB2TrkData::START  = 2;


// CONSTRUCTORS
SB2TrkData::SB2TrkData()
{
   name = NULL;
   path = NULL;
   seg  = NULL;

   num_segs = num_comps = 0;
   trk_len  = trk_wid   = 0.0;
   pit_spd  = pit_entry = 0.0;
}

// DESTRUCTOR
SB2TrkData::~SB2TrkData()
{
   int i;

   if (path) {
      for (i = 0; i < num_comps; i++) {
         delete path[i];
      }
      delete [] path;
      path = NULL;
   }
   
   if (seg) {
      for (i = 0; i < num_segs; i++) {
         delete seg[i];
      }
      delete [] seg;
      seg = NULL;
   }

   if (name) {
      free(name);
      name = NULL;
   }
}

// METHODS
int SB2TrkData::can_pass(double trk_dist, int pc) const 
{
   return ( (path[pc]->pass != NULL)&&(get_PassComponent(trk_dist, pc) < 1) );
}

double SB2TrkData::get_trk_dist(int seg_ID, double to_end) const
{
   double d;

   to_end /= seg[seg_ID]->len;
   
   d = ((to_end*seg[seg_ID]->Sdst) + ((1.0-to_end)*seg[seg_ID]->Fdst));
   if (d > trk_len)
      d -= trk_len;
   else if (d < 0.0)
      d += trk_len;
   return d;
}

double SB2TrkData::get_trk_dist(const SB2Vector &pt, double to_lft,
                               int start_seg_ID) const
{
   int i, j;

   for (j = 0; j < num_segs; j++) {
      i = (start_seg_ID + j) % num_segs;

      if (fabs(seg[i]->to_left(pt) - to_lft) < 5.0)
         return seg[i]->get_trk_dist(pt);
   }
   return -trk_len;
}

void SB2TrkData::get_trk_pos_vel(SB2Line &ret_pv, double to_lft, 
                                double v, double vn, 
                                double trk_dist, int seg_ID) const
{
   double ang;

   seg[seg_ID]->get_norm(ret_pv, trk_dist);

   // current position
   ret_pv.movePoint(to_lft);

   // current velocity
   ang          = ret_pv.dir.ang();
   ret_pv.dir.x = -vn;
   ret_pv.dir.y = sqrt((v*v) - (vn*vn));
   ret_pv.dir.rotate(ang);
}

int SB2TrkData::get_PathComponent(double trk_dist, int pass_flag,
                                 int start) const
{
   int              i, j, k;
   double           Fdst;
   SB2PathComponent *pc;

   if (trk_dist > trk_len)
      trk_dist -= trk_len;
   else if (trk_dist < 0.0)
      trk_dist += trk_len;

   if (start < 0)
      start += num_comps;

   for (j = 0; j < num_comps; j++) {
      i  = (start + j) % num_comps;
      pc = path[i];

      if ( (pass_flag)&&(pc->pass) )
         Fdst = pc->pass[0]->Sdst;
      else
         Fdst = pc->Fdst;
      if ( (pc->is_after(trk_dist, *pc, pc->Sdst))&&
           (pc->is_after(Fdst, *pc, trk_dist)) )
         return i;

      if ( (pass_flag)&&(pc->pass) ) {
         for (k = 0; k < 2; k++) {
            pc = path[i]->pass[k];
            if ( (pc->is_after(trk_dist, *pc, pc->Sdst))&&
                 (pc->is_after(pc->Fdst, *pc, trk_dist)) )
               return i;
         }
      }
   }
   return -1;
}

double SB2TrkData::get_tgt_spd(double v, double drag_accel, 
                              double path_dist, double trk_dist,
                              int pass_flag, int pc) const
{
   int    i, j;
   double nex_spd, tst_spd, brk, brk_dist, brk_end;
   double dist2comp, end_comp, max_dist, min_dist;

   if (pc < 0) {
      pc = get_PathComponent(trk_dist, pass_flag);
      if (pc < 0)
         return 0.0;
   }

   i         = pc;
   max_dist  = SB2_brkdist(v, 0.0, drag_accel);
   min_dist  = v * delta_time;
   dist2comp = brk_dist = 0.0;
   end_comp  = -path_dist;
   if ( (pass_flag)&&(path[i]->pass) )
      j = get_PassComponent(trk_dist, i);
   else
      j = -1;

   if (j >= 0) {  
      // on the passing path
      path[i]->pass[j]->get_spd(trk_dist, &nex_spd);
      end_comp += path[i]->pre_len + path[i]->pass[0]->len;
      if (j == 1) {
         end_comp += path[i]->pass[j]->len;
         dist2comp = -2.0 * v * delta_time;
         brk_dist  = dist2comp;
      }
      brk_end = end_comp;
   }
   else {
      // on the racing path
      path[i]->get_spd(trk_dist, &nex_spd);
      end_comp += path[i]->len;
      brk_end   = path[i]->brk_dist(end_comp, v);
   }
   j++;

   do {
      dist2comp += end_comp;
      brk_dist  += brk_end;
      if ( (pass_flag)&&(path[i]->pass)&&(j >= 0)&&(j <= 2) ) {
         if (j < 2) {
            // check speed on passing path
            if (!j)
               dist2comp += path[i]->pre_len - path[i]->len;
            end_comp = path[i]->pass[j]->len;
            brk_end  = path[i]->pass[j]->brk_dist(end_comp, v);
            tst_spd  = path[i]->pass[j]->get_spd(path[i]->pass[j]->Sdst);
            j++;
         }
         else {
            // check speed on component after passing path
            j = i;
            do {
               i = (i + 1) % num_comps;
            } while (path[i]->typeID == SB2PathComponent::STRAIGHT);
            end_comp = path[i]->post_len;
            brk_end  = path[i]->brk_dist(end_comp, v);
            tst_spd  = path[i]->get_spd(path[j]->pass[1]->Fdst);
            j        = 0;
         }
      }
      else {
         // check speed on next component
         i        = (i + 1) % num_comps;
         end_comp = path[i]->len;
         brk_end  = path[i]->brk_dist(end_comp, v);
         tst_spd  = path[i]->get_spd(path[i]->Sdst);
      }

      brk = SB2_brkdist(v, tst_spd, drag_accel);
      if (brk_dist > min_dist)
         brk *= dist2comp / brk_dist;
      if ( (brk >= dist2comp)&&(nex_spd > tst_spd) )
         nex_spd = tst_spd;
      else if ( (dist2comp <= min_dist)&&(tst_spd > nex_spd) )
         nex_spd = tst_spd;
   } while (dist2comp < max_dist);
   return nex_spd;
}

int SB2TrkData::get_nex_pos(SB2Line &cV, SB2Vector &nV, int type,
                           double trk_dist, int pass_flag, int pc) const
{
   int      i, retval;
   double   v, step, MAX_dA, to_side;
   SB2Vector t1, t2;

   if (pc < 0) {
      pc = get_PathComponent(trk_dist, pass_flag);
      if (pc < 0)
         return 0;
   }

   to_side = nV.y;
   v       = nV.x;
   t2.x    = cV.dir.mag();
   if (t2.x < 73.0)
      step = 20.0 / (t2.x * delta_time);
   else
      step =  5.0;

   if ( (pass_flag)&&(path[pc]->pass) ) {
      i = get_PassComponent(trk_dist, pc);
      if (i < 0) {
         // check if car has entered the passing path
         t1.x   = path[pc]->Fdst;
         t1.y   = path[pc]->pass[0]->Sdst;
         retval = ( (path[pc]->is_after(trk_dist, *(path[pc]), t1.x))&&
                    (path[pc]->is_after(t1.y, *(path[pc]), trk_dist)) );
      }
      else if (i == 1) {
         // check if car has left the passing path
         t1.x  = path[pc]->pass[1]->dst_between(path[pc]->pass[1]->Fdst,
                                                *(path[pc]->pass[1]),
                                                trk_dist);
         retval = (t1.x > (v*delta_time));
      }
      else
         retval = 1;
   }
   else {
      i      = -1;
      retval = 0;
   }

   // get current point, and tangent vector
   if ( (i >= 0)&&(i < 2) ) {
      trk_dist = path[pc]->pass[i]->get_path_dist(cV.pt);
      MAX_dA   = path[pc]->pass[i]->comp->dist2Point(cV.pt);
      path[pc]->pass[i]->getPoint2(nV, t1, trk_dist);
      if (path[pc]->pass[i]->typeID == SB2PathComponent::RTURN)
         to_side = trk_wid - to_side;
      else if (path[pc]->pass[i]->typeID == SB2PathComponent::STRAIGHT)
         to_side = 0.2 * trk_wid;
   }
   else {
      trk_dist = path[pc]->get_path_dist(cV.pt);
      MAX_dA   = path[pc]->comp->dist2Point(cV.pt);
      path[pc]->getPoint2(nV, t1, trk_dist);
      if (path[pc]->typeID == SB2PathComponent::RTURN)
         to_side = trk_wid - to_side;
      else if (path[pc]->typeID == SB2PathComponent::STRAIGHT)
         to_side = 0.2 * trk_wid;
   }

   // find new point
   trk_dist += step * t2.x * delta_time;
   do {
      if ( (pass_flag)&&(path[pc]->pass) ) {
         if (i < 0) {
            if (trk_dist <= path[pc]->pre_len)
               path[pc]->getPoint2(nV, t2, trk_dist);
            trk_dist -= path[pc]->pre_len;
            i         = 0;
         }
         else if (i == 0) {
            if (trk_dist <= path[pc]->pass[0]->len)
               path[pc]->pass[0]->getPoint2(nV, t2, trk_dist);
            trk_dist -= path[pc]->pass[0]->len;
            i         = 1;
         }
         else {
            if (trk_dist <= path[pc]->pass[1]->len)
               path[pc]->pass[1]->getPoint2(nV, t2, trk_dist);
            trk_dist -= path[pc]->pass[1]->len;
            if (trk_dist > 0.0) {
               // next component
               do {
                  pc = (pc + 1) % num_comps;
               } while (path[pc]->typeID == SB2PathComponent::STRAIGHT);
               trk_dist += path[pc]->len - path[pc]->post_len;
               i         = -1;
            }
         }
      }
      else {
         if (trk_dist <= path[pc]->len)
            path[pc]->getPoint2(nV, t2, trk_dist);
         trk_dist -= path[pc]->len;
         pc = (pc + 1) % num_comps;
      }
   } while (trk_dist > 0.0);

   // new velocity vector
   nV.sub(cV.pt);
   nV.normalize();
   nV.scale(v);
   
   // limit the change in direction
   if (type != SB2TrkData::OFF) {
      MAX_dA = fabs(MAX_dA) / trk_wid;
      if (MAX_dA < 0.1)
         MAX_dA = 0.1;
      else if (MAX_dA > 0.4)
         MAX_dA = 0.4;
      if (type == SB2TrkData::START)
         MAX_dA *= 0.6;
   }
   else
      MAX_dA = 0.4;

   v = cV.dir.mag();
#ifndef SB2_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 0:
      if (v > 142.0)
         v  = 1.0;
      else
         v *= 0.007;
      break;
   case 1:
      if (v > 166.0)
         v  = 1.0;
      else
         v *= 0.006;
      break;
   case 2:
      v *= 0.0025;
   };

   if (to_side <= 0.0)
      to_side = 0.0;
   else if (to_side >= trk_wid)
      to_side = v;
   else
      to_side = v * (to_side / trk_wid);
   t1.rotate(to_side * t1.ang(t2));
   v    = t1.ang(nV);
   step = v / step;
   if (step > MAX_dA)
      nV.rotate(MAX_dA - v);
   else if (step < -MAX_dA)
      nV.rotate(-MAX_dA - v);
   else if (type != SB2TrkData::OFF) {
      MAX_dA = 1.0 - MAX_dA;
      nV.rotate((step - v)*MAX_dA);
   }
   return retval;
}

void SB2TrkData::locate_car(double *dist2path, double *path_dist, 
                           const SB2Vector &pt, int start_pc, int end_pc,
                           double trk_dist, int pass_flag) const
{
   int i, j;

   if ( (pass_flag)&&(path[end_pc]->pass) )
      j = get_PassComponent(trk_dist, end_pc);
   else
      j = -1;

   if (path_dist) {
      *path_dist = 0.0;
      for (i = start_pc; i != end_pc; i = (i+1)%num_comps) {
         *path_dist += path[i]->len;
      }
   }
   else
      i = end_pc;

   if ( (j >= 0)&&(j < 2) ) {
      if (path_dist) {
         *path_dist += path[i]->pre_len;
         if (j)
            *path_dist += path[i]->pass[0]->len;
         *path_dist += path[i]->pass[j]->get_path_dist(pt);
      }
      if (dist2path)
         *dist2path  = path[i]->pass[j]->get_dist2path(pt);
   }
   else {
      if (path_dist)
         *path_dist += path[i]->get_path_dist(pt);
      if (dist2path)
         *dist2path  = path[i]->get_dist2path(pt);
   }
}

void SB2TrkData::set_path(Track *trk)
{
   int          i, num_turns, data_loaded;
   SB2Turn     **crn = NULL;
   SB2Straight **str = NULL;

   // remove old data
   this->~SB2TrkData();

   // get RARS track data
   num_turns = get_trk_data(trk);

   // Make some corners
   data_loaded = init_Turns(&crn, &num_turns);

   // Make some straights
   str = new SB2Straight*[num_turns];
   for (i = 0; i < num_turns; i++) {
      if (crn[i]->Sdst >= trk_len)
         crn[i]->Sdst -= trk_len;
      if (crn[i]->Sdst < 0.0)
         crn[i]->Sdst += trk_len;
      if (crn[i]->Fdst >= trk_len)
         crn[i]->Fdst -= trk_len;
      if (crn[i]->Fdst < 0.0)
         crn[i]->Fdst += trk_len;
      str[i] = new SB2Straight();  // initialize the straights
   }

   // Try and make everything fit
   adjust_Turns(crn, str, num_turns, 
                ((data_loaded)?(SB2_NUM_RNDS-1):0), 0, NULL);

   // Combine turns and straights into the path
   insert_Straights(crn, str, num_turns, data_loaded);

   // free memory
   delete [] crn;
   delete [] str;

   // Create the passing path
   makePassingPath();

//   saveData();
}


// PRIVATE METHODS
int SB2TrkData::get_trk_data(Track *trk)
{
   int i, num_turns;

   // store track name
   i       = strlen(trk->m_sFileName);
   name    = (char *)malloc(i + 1);
   name[i] = 0;
   for (; i >= 0; i--) {
      name[i] = (char)tolower((int)trk->m_sFileName[i]);
   }

   num_turns = 0;
   trk_len   = 0.0;
   trk_wid   = trk->width;
   num_segs  = trk->NSEG;
   pit_spd   = trk->m_fPitLaneSpeed + 50.0;
   pit_entry = trk->m_fPitEntry;
   seg       = new SB2TrkSeg*[num_segs];
   for (i = 0; i < num_segs; i++) {
      if (trk->lftwall[i].radius == 0.0)   // straight
         seg[i] = new SB2TrkSeg_Straight(i, (trk->lftwall + i), trk_len);
      else {                               // turn
         seg[i] = new SB2TrkSeg_Turn(i, (trk->lftwall + i), trk_wid, trk_len);
         num_turns++;
      }
      trk_len += seg[i]->Fdst - seg[i]->Sdst;
   }
   return num_turns;
}

int SB2TrkData::init_Turns(SB2Turn ***p_crn, int *p_num_turns)
{
   int      i, j;
   double   min_w, max_w;
   SB2Turn **crn;

   // check if there is data that can be loaded
   if (loadData(p_crn, p_num_turns))
      return 1;

   min_w = SB2_IN_WID;
   max_w = trk_wid - SB2_OUT_WID;
   crn   = *p_crn = new SB2Turn*[*p_num_turns];
   for (i = 0, j = 0; i < num_segs; i++) {
      if (seg[i]->rad == 0.0)   // straight
         continue;
      if (j == *p_num_turns)
         break;

      crn[j++] = new SB2Turn(seg, i, num_segs, min_w, max_w, trk_wid);
   }
   return 0;
}

void SB2TrkData::adjust_Turns(SB2Turn **crn, SB2Straight **str, 
                             int num_turns, int rnds, int level, 
                             int *adj_val)
{
   int        i, j, k, del, chk, stop;
   SB2Turn     ct, ot;
   SB2Straight ps;
   double     min_w = SB2_IN_WID, 
              max_w = trk_wid - SB2_OUT_WID;

   adj_val = new int[num_turns];
   for (i = 0; i < num_turns; i++) {
      adj_val[i] = SB2Turn::BAD_ENTRY|SB2Turn::BAD_EXIT;
   }

   stop = ((num_turns-1) > SB2_SEARCH_LEN) ? SB2_SEARCH_LEN : (num_turns-1);
   for (; rnds < SB2_NUM_RNDS; rnds++) {
      // Find problems in current path
      chk = 0;
      for (j = 0; j < num_turns; j++) {
         if ( (crn[j] == NULL)||(adj_val[j]&SB2Turn::IGNORE_TURN) )
            continue;

         if (!crn[j]->valid(trk_wid))
            adj_val[j] |= SB2Turn::BAD_TURN;
         else {
            // Check entry to turn
            ct.copy(*(crn[j]), 1, 1, (num_segs-crn[j]->num_segs));
            adj_val[j] |= SB2Turn::BAD_ENTRY;
            for (k = 1, del = 0; (k < stop)||(k == 1); k++) {
               if (k > 1) {
                  // insert TrkSegs from the last turn into this one
                  while (crn[i] == NULL) {
                     i = (i + 1) % num_turns;
                  }
                  if (i != j)
                     ct.insert_TrkSeg(*(crn[i]), 0);
               }
               i = (j + num_turns - k) % num_turns;
               if ( (crn[i] == NULL)||(adj_val[i]&SB2Turn::IGNORE_TURN) )
                  continue;
               ot.copy(*(crn[i]), 1, 0, (num_segs-crn[i]->num_segs));
               if (k > 1)
                  ot.insert_TrkSeg(ct, 1);

               if ( (ct.makeStraight(ps, ot, trk_wid))&&
                    (crn[j]->is_after(crn[j]->Fdst, ct, ct.Sdst))&&
                    (ct.valid(trk_wid))&&(ot.valid(trk_wid))&&
                    (ps.valid(trk_wid)) ) {
                  adj_val[j] &= ~SB2Turn::BAD_ENTRY;
                  str[j]->copy(ps);
                  if ( (k == 1)||
                       (!(adj_val[i]&(SB2Turn::BAD_TURN|SB2Turn::BAD_ENTRY))) ) {
                     // this turn is good
                     del = k - 1;
                     crn[i]->copy(ot, 0, 0, 0);
                     crn[j]->copy(ct, 0, 1, 0);
                     if (k > 1) {
                        crn[i]->insert_TrkSeg(ot, 1);
                        crn[j]->insert_TrkSeg(ct, 0);
                     }
                  }
               }
               if (adj_val[j]&(SB2Turn::BAD_TURN|SB2Turn::BAD_EXIT))
                  break;  // do not try and remove any turns
            }
            // Remove unecessary turns
            for (; del > 0; del--) {
               // remove these
               i = (j + num_turns - del) % num_turns;
               if (crn[i] != NULL) {
                  delete crn[i];
                  crn[i] = NULL;
               }
               if (str[i] != NULL) {
                  delete str[i];
                  str[i] = NULL;
               }
            }

            // Check exit from turn
            ct.copy(*(crn[j]), 1, 0, (num_segs-crn[j]->num_segs));
            adj_val[j] |= SB2Turn::BAD_EXIT;
            for (i = 1, del = 0; (i < stop)||(i == 1); i++) {
               if (i > 1) {
                  // insert TrkSegs from the last turn into this one
                  while (crn[k] == NULL) {
                     k = (k + num_turns - 1) % num_turns;
                  }
                  if (k != j)
                     ct.insert_TrkSeg(*(crn[k]), 1);
               }
               k = (j + i) % num_turns;
               if ( (crn[k] == NULL)||(adj_val[k]&SB2Turn::IGNORE_TURN) )
                  continue;
               ot.copy(*(crn[k]), 1, 1, (num_segs-crn[k]->num_segs));
               if (i > 1)
                  ot.insert_TrkSeg(ct, 0);

               if ( (ot.makeStraight(ps, ct, trk_wid))&&
                    (ct.is_after(ct.Fdst, *(crn[j]), crn[j]->Sdst))&&
                    (ct.valid(trk_wid))&&(ot.valid(trk_wid))&&
                    (ps.valid(trk_wid)) ) {
                  adj_val[j] &= ~SB2Turn::BAD_EXIT;
                  str[k]->copy(ps);
                  if ( (i == 1)||
                     (!(adj_val[k]&(SB2Turn::BAD_TURN|SB2Turn::BAD_EXIT))) ) {
                     // this turn is good
                     del = i - 1;
                     crn[j]->copy(ct, 0, 0, 0);
                     crn[k]->copy(ot, 0, 1, 0);
                     if (i > 1) {
                        crn[j]->insert_TrkSeg(ct, 1);
                        crn[k]->insert_TrkSeg(ot, 0);
                     }
                  }
               }
               if (adj_val[j]&(SB2Turn::BAD_TURN|SB2Turn::BAD_ENTRY))
                  break;  // do not try and remove any turns
            }
            // Remove unecessary turns
            for (; del > 0; del--) {
               // remove this
               k = (j + del) % num_turns;
               if (crn[k] != NULL) {
                  delete crn[k];
                  crn[k] = NULL;
               }
               if (str[k] != NULL) {
                  delete str[k];
                  str[k] = NULL;
               }
            }
            
            adj_val[j] &= ~SB2Turn::BAD_TURN;  // remove this flag
         }     
         chk += adj_val[j];
      }
      if ( (chk == 0)||((rnds+1) == SB2_NUM_RNDS) )
         break;

      // Fix problems in the path
      chk = 0;
      for (j = 0; j < num_turns; j++) {
         if ( (crn[j] == NULL)||(adj_val[j] == 0)||
              (adj_val[j]&SB2Turn::IGNORE_TURN) )
            continue;

         i = j;  // prev turn index
         do {
            i = (i + num_turns - 1) % num_turns;
         } while ( (crn[i] == NULL)||(adj_val[i]&SB2Turn::IGNORE_TURN) );
         k = j;  // next turn index
         do {
            k = (k + 1) % num_turns;
         } while ( (crn[k] == NULL)||(adj_val[k]&SB2Turn::IGNORE_TURN) );

         if ( (!(adj_val[i]&SB2Turn::BAD_TURN))&&
              (!(adj_val[k]&SB2Turn::BAD_TURN)) ) {
            ct.copy(*(crn[i]), 1);
            ct.insert_TrkSeg(*(crn[j]), 1);
            ot.copy(*(crn[k]), 1);
            ot.insert_TrkSeg(*(crn[j]), 0);
            if (ct.is_after(ct.Fdst, ot, ot.Sdst)) {
               adj_val[j] |= SB2Turn::IGNORE_TURN;
               continue;  // this turn is no longer needed
            }
         }
         if ( (crn[(j+num_turns-1) % num_turns] == NULL)||
              (crn[(j+1) % num_turns] == NULL) )
            continue;  // stop adjusting this turn

         del = adj_val[j];  // save value
         if ( (adj_val[j]&SB2Turn::BAD_ENTRY)&&
              (adj_val[i]&SB2Turn::BAD_TURN) )
            adj_val[j] &= ~SB2Turn::BAD_ENTRY;
         if ( (adj_val[j]&SB2Turn::BAD_EXIT)&&
              (adj_val[k]&SB2Turn::BAD_TURN) )
            adj_val[j] &= ~SB2Turn::BAD_EXIT;

         // Adjust turn
         crn[j]->adjust(*(crn[i]), *(crn[k]), adj_val[j], 
                        min_w, max_w, trk_wid);
         adj_val[j] = del;  // restore value

         chk++;
      }
      if (chk == 0)
         break;
   }

   if ( (chk != 0)&&(level < 4)&&(rnds > 0) ) {
      level++;
      stop    = 1;
      ct.len  = (double)level / 9.0;
      ct.Sdst = (min_w * (1.0-ct.len)) + (max_w * ct.len);
      ct.Fdst = (max_w * (1.0-ct.len)) + (min_w * ct.len);

      for (j = 0; j < num_turns; j++) {
         if ( (crn[j] == NULL)||(adj_val[j] == 0)||
              (adj_val[j] == SB2Turn::BAD_ENTRY)||
              (adj_val[j] == SB2Turn::BAD_EXIT) )
            continue;

         // a problem in the path could not be fixed, try again
         stop       = 0;
         adj_val[j] = SB2Turn::BAD_ENTRY|SB2Turn::BAD_EXIT;
         crn[j]->init(ct.Sdst, ct.Fdst, trk_wid);
      }
   
      if (!stop)
         adjust_Turns(crn, str, num_turns, 0, level, adj_val);
      level--;
   }
   if (level == 0)
      delete [] adj_val;   // free memory
}

void SB2TrkData::insert_Straights(SB2Turn **crn, SB2Straight **str, 
                                 int num_turns, int data_loaded)
{
   int    i, j, k;
   double d, Sdst, Fdst;

   // Check for invalid straights
   for (j = 0; j < num_turns; j++) {
      if (str[j] == NULL)
         continue;

      if ( (str[j]->Sdst > 0.0)||(str[j]->Fdst > 0.0) ) {
         k = j;  // find the previous turn
         do {
            k = (k + num_turns - 1) % num_turns;
         } while (crn[k] == NULL);
         crn[j]->Sdst = str[j]->Fdst;
         crn[k]->Fdst = str[j]->Sdst;
      }

      if (str[j]->len < SB2_MIN_LEN) {
         delete str[j];
         str[j] = NULL;   // remove
      }
   }
      
   // Now put it all together
   num_comps = 0;
   for (j = 0; j < num_turns; j++) {
      if (crn[j] != NULL)
         num_comps++;
      if (str[j] != NULL)
         num_comps++;
   }
   path = new SB2PathComponent*[num_comps];
   for (i = 0, j = 0; i < num_comps; i++, j++) {
      // find a valid turn
      while (crn[j] == NULL) {
         j++;
      }
         
      if (str[j] != NULL) {
         Sdst = str[j]->Sdst;

         // put this straight component into the path
         path[i] = str[j];
         path[i]->trim();
         path[i]->len_trk = path[i]->dst_between(path[i]->Fdst, 
                                                 *(path[i]), 
                                                 path[i]->Sdst);
         i++;
      }
      else {
         k = j;  // find the previous turn
         do {
            k = (k + num_turns - 1) % num_turns;
         } while (crn[k] == NULL);

         // connect these 2 turns
         d  = crn[j]->dst_between(crn[j]->Sdst, *(crn[k]), crn[k]->Fdst);
         d *= 0.5;
         d += crn[k]->Fdst;
         if (d < 0.0)
            d += trk_len;
         else if (d > trk_len)
            d -= trk_len;
         crn[k]->Fdst = crn[j]->Sdst = Sdst = d;
      }
      crn[j]->set_spd_len(data_loaded);
      path[i] = crn[j];

      k = j;  // find the next turn
      do {
         k = (k + 1) % num_turns;
      } while (crn[k] == NULL);
      d                = Sdst;
      Sdst             = path[i]->Sdst;  // store real values
      Fdst             = path[i]->Fdst;
      path[i]->Sdst    = d;              // extend PathComponent
      path[i]->Fdst    = crn[k]->Sdst;
      path[i]->trim();
      path[i]->Sdst    = Sdst;           // restore real values
      path[i]->Fdst    = Fdst;
      path[i]->len_trk = path[i]->dst_between(path[i]->Fdst, 
                                              *(path[i]), 
                                              path[i]->Sdst);
   }
}

void SB2TrkData::makePassingPath()
{
   int         i, j, _can_pass;
   double      max_w;
   SB2Turn     *ct, *nt, *p_t, *last;
   SB2Straight *p_s;

   if (num_comps <= 0)
      return;

   // find last turn
   i = num_comps;
   do {
      i--;
   } while ( (path[i] == NULL)||
             (path[i]->typeID == SB2PathComponent::STRAIGHT) );
   last = (SB2Turn *)path[i];

   // create passing path
   max_w = 2.0 * SB2_OUT_WID;
   p_s   = NULL;
   p_t   = ct = nt = NULL;
   for (i = 0; i < num_comps; i++) {
      if ( (path[i] == NULL)||(path[i]->typeID == SB2PathComponent::STRAIGHT) )
         continue;

      // find next turn
      j = i;
      do {
         j = (j + 1) % num_comps;
      } while ( (path[j] == NULL)||
                (path[j]->typeID == SB2PathComponent::STRAIGHT) );
      ct = (SB2Turn *)path[i];
      nt = (SB2Turn *)path[j];

      if ( (p_s == NULL)&&(p_t == NULL) ) {
         p_s = new SB2Straight();
         p_t = new SB2Turn();
      }
      if ( (p_s == NULL)||(p_t == NULL) )
         break;

      _can_pass = ct->makePass(p_s, p_t, *nt, *last, 
                               SB2_IN_WID, max_w, trk_wid);
      last      = ct;
      if (_can_pass) {
         // get new components next time through loop
         p_s = NULL;
         p_t = NULL;
      }
   }

   // remove unused components
   if (p_s != NULL)
      delete p_s;
   if (p_t != NULL)
      delete p_t;
}

void SB2TrkData::saveData() const
{
   FILE   *out;
   int     i, nt;
   double  rotate;
   SB2Turn *st;
   SB2Line *start;

   for (i = 0, nt = 0; i < num_comps; i++) {
      if (path[i]->typeID != SB2PathComponent::STRAIGHT)
         nt++;
   }

   out = fopen("SB2path_out.txt", "w");
   if (!out)
      return;

   fprintf(out, "v3\n%s\n%d\n", name, nt);

   start  = (SB2Line *)seg[0]->left_side;
   rotate = -start->dir.ang();
   for (i = 0; i < num_comps; i++) {
      if (path[i]->typeID != SB2PathComponent::STRAIGHT) {
         st = (SB2Turn *)path[i];
         st->save(out, seg, rotate);
      }
   }
   fclose(out);
}

int SB2TrkData::loadData(SB2Turn ***p_crn, int *p_num_turns)
{
   const char *ws = " \t\n\r";
   int      i, len, found_trk;
   char     buf[128];
   double   rotate;
   FILE    *in;
   SB2Turn **crn;
   SB2Line  *start;

   in = fopen(DATA_FILE, "r");
   if (!in)
      goto SB2TD_LOAD_EXIT_0;

   len       = 0;
   found_trk = 0;

   if (!fgets(buf, 128, in))
      goto SB2TD_LOAD_EXIT_1;
   if (strncmp(buf, "v3", 2))
      goto SB2TD_LOAD_EXIT_1;

   while (!found_trk) {
      if (!fgets(buf, 128, in))
         goto SB2TD_LOAD_EXIT_1;
      found_trk = (!strncmp(buf, name, strlen(name)));

      if (!fgets(buf, 128, in))
         goto SB2TD_LOAD_EXIT_1;
      len = atoi(strtok(buf, ws));
      if (found_trk)
         break;
      len *= SB2Turn::DATALEN;
      for (i = 0; i < len; i++) {
         if (!fgets(buf, 128, in))
            goto SB2TD_LOAD_EXIT_1;
      }
   }

   start  = (SB2Line *)seg[0]->left_side;
   rotate = start->dir.ang();
   crn    = new SB2Turn*[len];
   for (i = 0; i < len; i++) {
      crn[i] = new SB2Turn();
      if (!crn[i]->load(in, seg, num_segs, rotate))
         goto SB2TD_LOAD_EXIT_2;
   }
   fclose(in);
   *p_crn       = crn;
   *p_num_turns = len;
   return 1;

SB2TD_LOAD_EXIT_2:
   do {
      delete crn[i--];
   } while (i >= 0);
   delete [] crn;
SB2TD_LOAD_EXIT_1:
   fclose(in);
SB2TD_LOAD_EXIT_0:
   return 0;
}

int SB2TrkData::get_PassComponent(double trk_dist, int pc) const
{
   int              i;
   SB2PathComponent *comp;

   if (trk_dist > trk_len)
      trk_dist -= trk_len;
   else if (trk_dist < 0.0)
      trk_dist += trk_len;

   for (i = 0; i < 2; i++) {
      comp = path[pc]->pass[i];
      if ( (comp->is_after(trk_dist, *comp, comp->Sdst))&&
           (comp->is_after(comp->Fdst, *comp, trk_dist)) )
         return i;
   }
   return -1;
}
//-----END TrkData.cpp-----//

//-----START SB2Driver.cpp-----//
// GLOBALS
static int       trk_num = -1;
static SB2TrkData trk_data;

// CONSTRUCTORS
SBv2r3::SBv2r3(int i)
   : cV(), nV()
{
#ifndef SB2_DLL
   if ( (0 <= i)&&(i < SB2_CARS) )
      m_sName = SB2_NAME[i+1];
   else
      m_sName = SB2_NAME[0];

   m_sAuthor       = SB2_AUTHOR;
   m_sDescription  = SB2_DESCRIPTION;
   m_iNoseColor    = SB2_NOSE;
   m_iTailColor    = SB2_TAIL;
   m_sBitmapName2D = SB2_BITMAP2D;
   m_sModel3D      = SB2_MODEL3D;
#endif
   init      = 2;
   lastpit   = 0;
   fuelstart = 0.0;
   pc        = 0;
   pass_flag = 0;
   car2pass  = MAX_CARS;
}

// DESTRUCTOR
SBv2r3::~SBv2r3()
{
   if (trk_num >= 0) {
      // free memory used by the track and path data
      trk_data.~SB2TrkData();
      trk_num = -1;
   }
}


// METHODS
con_vec SBv2r3::drive(situation &s)
{
#ifndef SB2_DLL
   if ( (args.m_iCurrentTrack != trk_num)&&(currentTrack) ) {
      trk_num = args.m_iCurrentTrack;
      switch (args.m_iSurface) {
#else
   if (!args)
      return cp;
   else if ( (args->m_iCurrentTrack != trk_num)&&(currentTrack) ) {
      trk_num = args->m_iCurrentTrack;
      switch (args->m_iSurface) {
#endif
      case 0:
         MAX_ACCEL = MYU_MAX0;
         break;
      case 1:
         MAX_ACCEL = MYU_MAX1;
         break;
      case 2:
         MAX_ACCEL = MYU_MAX2;
      }
      MAX_ACCEL *= g;
      trk_data.set_path(currentTrack);
   }
   else if (!currentTrack)
      return cp;

   if (s.starting) {
      init             = (init < 1) ? 1 : init;
      s.side_vision    = 1;
      cp.request_pit   = 0;
      cp.repair_amount = 0;
      cp.fuel_amount   = SB2_fuel_needed(s.stage, s.laps_to_go, 
                                        (SB2_FC * currentTrack->length),
                                        0.0);
      lastpit          = 0;
      fuelstart        = cp.fuel_amount;
      if ( (s.stage == RACING)&&(s.position > 3) ) {
         // use passing path for opening lap
         pc        = -2;
         pass_flag =  1;
      }
   }
   
   if (init == 2) {
      init--;
      cp.vc    = SB2_MAX_SPD;
      cp.alpha = 0.0;
   }
   else
      get_RARS_ctrl_params(s);
   return cp;
}


// PRIVATE METHODS
void SBv2r3::get_RARS_ctrl_params(situation &s)
{
   int    type;
   double trk_dist, tgt_v, acc;
   double path_dist, drag_accel;

   // Handle fuel / damage
   if (s.laps_done < 0)
      fuelstart = s.fuel;
   else if ( (s.lap_flag)&&(s.laps_to_go > 1)&&(s.laps_done > lastpit) ) {
      cp.fuel_amount  = fuelstart - s.fuel;
      cp.fuel_amount /= (double)(s.laps_done - lastpit);
      if ( (s.fuel < (cp.fuel_amount*2.25))||
           ((s.laps_to_go > 4)&&(s.damage > 20000)) ) {
         cp.fuel_amount = SB2_fuel_needed(s.stage, (s.laps_to_go + 1), 
                                         cp.fuel_amount, 
                                         (s.fuel - cp.fuel_amount));
         cp.request_pit = (cp.fuel_amount > 1.0)||(s.damage > 20000);
         if (cp.request_pit)
            lastpit = s.laps_done + 1;
      }
   }
   if (cp.request_pit) {
      if ( (s.out_pits)&&(s.laps_done == lastpit) ) {
         cp.repair_amount = 0;
         cp.request_pit   = 0;
         fuelstart        = s.fuel;
      }
      else {
         acc = (cp.fuel_amount+s.fuel) / MAX_FUEL;
         if (acc > 1.0)
            acc = 1.0;
         if (acc < 0.9) {
            acc = (1.0 - acc) * 20000.0;
            if ((double)s.damage > acc)
               cp.repair_amount = s.damage - (unsigned long)acc;
            else
               cp.repair_amount = 0; // don't repair
         }
         else
            cp.repair_amount = (s.damage > 20000) ? 20000 : s.damage;
      }
   }

   // get current track/path position and velocity
   trk_dist  = trk_data.get_trk_dist(s.seg_ID, s.to_end);
   pc        = trk_data.get_PathComponent(trk_dist, pass_flag, pc);
   trk_data.get_trk_pos_vel(cV, s.to_lft, s.v, s.vn, trk_dist, s.seg_ID);
   locate_car(NULL, &path_dist, cV.pt, trk_dist, -1);

   // target velocity
   drag_accel = SB2_drag_accel(s.fuel, s.damage);
   tgt_v      = trk_data.get_tgt_spd(s.v, drag_accel, path_dist, 
                                     trk_dist, pass_flag, pc);
   if ( (cp.request_pit)&&(s.v > trk_data.pit_spd)&&
        (s.distance < trk_data.pit_entry) ) {
      acc = SB2_brkdist(s.v, trk_data.pit_spd, drag_accel);
      if ( (acc > (trk_data.pit_entry - s.distance))&&
           (tgt_v > trk_data.pit_spd) )
         tgt_v = trk_data.pit_spd;
   }

   // avoid other cars
   car2pass = MAX_CARS;
   if (s.v > 40.0)
      tgt_v *= anti_collision(s, path_dist, tgt_v, drag_accel);
   if ( (!pass_flag)&&(car2pass < MAX_CARS) ) {
      // try to use the passing path
      type = pc;
      pc   = trk_data.get_PathComponent(trk_dist, 1, (pc-1));
      if ( (trk_data.can_pass(trk_dist, pc))&&(pass_clear(s, trk_dist)) ) {
         pass_flag = 1;
         locate_car(NULL, &path_dist, cV.pt, trk_dist, -1);

         // target velocity
         acc = trk_data.get_tgt_spd(s.v, drag_accel, path_dist, 
                                    trk_dist, pass_flag, pc);
         if (acc < (tgt_v-5.0)) {
            // stay on racing path
            pass_flag = 0;
            pc        = type;
         }
      }
      else
         pc = type;
   }

   // limit acceleration
   acc = (tgt_v - s.v) / delta_time;
   if (acc > (5.0*MAX_ACCEL))
      tgt_v = s.v + (5.0 * MAX_ACCEL * delta_time);
   else if (acc < (-5.0*MAX_ACCEL))
      tgt_v = s.v - (5.0 * MAX_ACCEL * delta_time);

   // find target position
   if (init == 1) {
      if ( (pc != 1)&&(s.seg_ID != 1) )
         type = SB2TrkData::START;
      else
         init--;
   }
   else if ( ((s.to_lft < 0.0)||(s.to_rgt < 0.0))&&
             (!s.out_pits) )
      type = SB2TrkData::OFF;
   else
      type = SB2TrkData::NORMAL;
   nV.x      = tgt_v;
   nV.y      = s.to_lft;
   pass_flag = trk_data.get_nex_pos(cV, nV, type, trk_dist, pass_flag, pc);

   // get new control parameters
   update_con_vec((s.to_lft >= SB2_IN_WID)&&(s.to_rgt >= SB2_IN_WID));
}

void SBv2r3::update_con_vec(int damp)
{
   SB2Vector dV, L, W;
   double   acc;

   nV.rotate(-cV.dir.ang());
   acc      = cV.dir.mag();
   cV.dir.x = acc;
   cV.dir.y = 0.0;

   dV.add(nV);
   dV.sub(cV.dir);

   L.add(dV);
   L.normalize();
   acc = (dV.mag() / delta_time) / g;
#ifndef SB2_DLL
   L.scale(-SB2_iu(acc, args.m_iSurface));
#else
   L.scale(-SB2_iu(acc, args->m_iSurface));
#endif

   W.add(L);
   W.sub(cV.dir);

   cp.vc = W.mag();
   if ( (!damp)||(pass_flag) )
      cp.alpha = W.ang() - MY_PI;
   else {
#ifndef SB2_DLL
      switch (args.m_iSurface) {
#else
      switch (args->m_iSurface) {
#endif
      case 0:
         cp.alpha *= 0.1;
         cp.alpha += 0.9*(W.ang() - MY_PI);
         break;
      case 1:
         cp.alpha *= 0.15;
         cp.alpha += 0.85*(W.ang() - MY_PI);
         break;
      case 2:
         cp.alpha *= 0.2;
         cp.alpha += 0.8*(W.ang() - MY_PI);
      };
   }
}

void SBv2r3::locate_car(double *dist2path, double *path_dist, 
                         const SB2Vector &pt, double to_lft, int seg_ID)
{
   int    car_pc;
   double trk_dist;

   if (seg_ID >= 0) {
      trk_dist = trk_data.get_trk_dist(pt, to_lft, seg_ID);
      car_pc   = trk_data.get_PathComponent(trk_dist, pass_flag, pc);
   }
   else {
      trk_dist = to_lft;
      car_pc   = pc;
   }
   trk_data.locate_car(dist2path, path_dist, pt, pc, car_pc,
                       trk_dist, pass_flag);
}

double SBv2r3::anti_collision(situation &s, double path_dist, 
                               double tgt_v, double drag_accel)
{
   double brk, min_v, v;
   double oc_off_L, oc_path_dist;
   double max_dist, sn, cs, x, y;
   int    j, close;
   SB2Line oc;

   oc.dir.x = cV.dir.x; oc.dir.y = cV.dir.y;
   oc.dir.normalize();

   max_dist = SB2_brkdist(s.v, 0.0, drag_accel);
   min_v    = tgt_v;
   sn       = sin(0.75*s.alpha);
   cs       = cos(0.75*s.alpha);
   for (j = 0; j < NEARBY_CARS; j++) {
      if (s.nearby[j].who >= MAX_CARS)
         continue;
      
      // rotate the relative axis
      x   = s.nearby[j].rel_x;
      y   = s.nearby[j].rel_y;
      brk = (s.nearby[j].rel_x*cs) - (s.nearby[j].rel_y*sn);
      v   = (s.nearby[j].rel_y*cs) + (s.nearby[j].rel_x*sn);
      s.nearby[j].rel_x = brk;
      s.nearby[j].rel_y = v;
      brk = (s.nearby[j].rel_xdot*cs) - (s.nearby[j].rel_ydot*sn);
      v   = (s.nearby[j].rel_ydot*cs) + (s.nearby[j].rel_xdot*sn);
      s.nearby[j].rel_xdot = brk;
      s.nearby[j].rel_ydot = v;

      // check if car is approaching
      brk  = s.nearby[j].rel_x * s.nearby[j].rel_xdot;
      v    = s.nearby[j].rel_xdot * s.nearby[j].rel_xdot;
      brk += s.nearby[j].rel_y * s.nearby[j].rel_ydot;
      v   += s.nearby[j].rel_ydot * s.nearby[j].rel_ydot;
      v    = sqrt(v);
      if ( (brk > 0.0)&&(v > 3.0) )
         continue;   // not approaching this car

      // locate this car
      if (s.nearby[j].rel_y > SB2_MIN_DIST) {
         oc.pt.x = cV.pt.x; oc.pt.y = cV.pt.y;
         oc.movePoint(y);
         oc.dir.rotate(-MY_PI12);
         oc.movePoint(x);
         oc.dir.rotate(MY_PI12);
         locate_car(&oc_off_L, &oc_path_dist, oc.pt, 
                    s.nearby[j].to_lft, s.seg_ID);
         oc_path_dist -= path_dist;
         
         if ( ((s.nearby[j].to_lft < 0.0)&&
               ((s.nearby[j].to_lft-oc_off_L) > CARLEN))||
              ((s.nearby[j].to_rgt < 0.0)&&
               ((s.nearby[j].to_rgt+oc_off_L) > CARLEN)) )
            continue;   // car is off track and not in the way
         close = (oc_path_dist < SB2_MIN_DIST);
      }
      else {
         if ( (s.nearby[j].rel_y < CARLEN)&&
              ((s.nearby[j].to_lft < 0.0)||(s.nearby[j].to_rgt < 0.0)) )
            continue;   // car is off track and not in the way
         close = 1;
      }

      if (!close) {
         // this car is not too close
         brk = -0.95 * s.v;
         if (brk < SB2_CLOSE_RATE) {
            v = (oc_path_dist-SB2_MIN_DIST) / max_dist;
            if (v <= 0.0)
               v = SB2_CLOSE_RATE;
            else if (v >= 1.0)
               v = brk;
            else {
               v *= brk - SB2_CLOSE_RATE;
               v += SB2_CLOSE_RATE;
            }
         }
         else 
            v = SB2_CLOSE_RATE;
         brk  = CARLEN + 2.0;
         brk += SB2_brkdist(s.v, (s.nearby[j].v-SB2_CLOSE_RATE), drag_accel);

         if (oc_path_dist > brk)
            continue;   // car is too far away to worry about
         else if (s.nearby[j].v < (s.v+v))
            v = s.nearby[j].v * 0.75;   // approaching this car too fast
         else if ( (s.nearby[j].to_lft < 0.0)||(s.nearby[j].to_rgt < 0.0) )
            v = s.nearby[j].v;         // car is off track, slow down
         else {
            v = s.nearby[j].v;
            if ( (v < (s.v+SB2_CLOSE_RATE))&&(fabs(oc_off_L) < SB2_MARGIN) )
               v -= SB2_CLOSE_RATE;  // car is too close to the path
            else
               continue;            // car is not in the way
         }
      }
      else {
         // this car is too close
         oc_path_dist = s.nearby[j].rel_y;
         oc_off_L     = s.nearby[j].rel_x;

         if (oc_path_dist >= CARLEN) {
            if ( ((s.nearby[j].to_lft < 0.0)&&
                  ((s.nearby[j].to_lft-oc_off_L) > CARWID))||
                 ((s.nearby[j].to_rgt < 0.0)&&
                  ((s.nearby[j].to_rgt+oc_off_L) > CARWID)) )
               continue;   // car is off track and not in the way
            else if ( (s.nearby[j].to_lft < 0.0)||
                      (s.nearby[j].to_rgt < 0.0) ) {
               v = s.nearby[j].v * 0.9;
               if (v < min_v)
                  min_v = v;
               continue;   // approaching this car too fast
            }
         }

         if ( (car2pass >= MAX_CARS)&&(fabs(oc_off_L) < CARLEN) )
            car2pass = j;  // found a car to pass

         if ( (oc_path_dist < (-0.65*CARLEN))||(brk >= 0.0)||
              ((brk < 0.0)&&(v < 2.5)) )
            continue;   // already passed this car

         brk = fabs(oc_off_L);
         if (brk > CARWID) {
            if (oc_path_dist == 0.0)
               brk = MY_PI12;
            else {
               brk = atan(brk/oc_path_dist);
               if (brk < 0.0)
                  brk += MY_PI;
            }
            if (brk > 0.48) {
               // car is not directly in front
               if (s.nearby[j].rel_ydot >= 0.0)
                  continue;   // ignore this car
               brk = atan(fabs(s.nearby[j].rel_xdot/s.nearby[j].rel_ydot));
               if (v < 25.0) {
                  v -=  2.5;
                  v *= -((80.0/90.0)*MY_PI12) / 22.5;
                  v +=  (83.0/90.0)*MY_PI12;
               }
               else
                  v  =  (3.0/90.0)*MY_PI12;
               if (brk < v)
                  continue;   // impact angle is small enough to ignore
            }
         }
         // find displacements in y dir
         v = (0.5*CARLEN) + SB2_MARGIN;
         if (oc_path_dist > v) {
            oc.pt.x =  v - oc_path_dist;
            oc.pt.y = -v - oc_path_dist;
         }
         else {
            oc.pt.x = 0.0;
            if (s.nearby[j].rel_ydot >= 0.0)
               oc.pt.y =  v - oc_path_dist;
            else
               oc.pt.y = -v - oc_path_dist;
         }
         
         // get times in y dir
         if ( ((oc.pt.y < 0.0)&&(s.nearby[j].rel_ydot < 0.0))||
              ((oc.pt.y > 0.0)&&(s.nearby[j].rel_ydot > 0.0)) ) {
            oc.pt.x /= s.nearby[j].rel_ydot;
            oc.pt.y /= s.nearby[j].rel_ydot;
            oc.pt.y *= 0.9;
            oc.pt.y += 0.1 * oc.pt.x;
         }
         else if ( (oc.pt.x == 0.0)&&(s.nearby[j].rel_ydot == 0.0) )
            oc.pt.y = 100000.0;
         else
            continue;   // not going to hit this car

         // find displacements in x dir
         v = (0.5*CARWID) + SB2_MARGIN;
         if (oc_off_L < -v) {
            brk = -v - oc_off_L;
            v   =  v - oc_off_L;
         }
         else if (oc_off_L > v) {
            brk =  v - oc_off_L;
            v   = -v - oc_off_L;
         }
         else {
            brk = 0.0;
            if (s.nearby[j].rel_xdot >= 0.0)
               v =  v - oc_off_L;
            else
               v = -v - oc_off_L;
         }

         // get times in x dir
         if ( ((v > 0.0)&&(s.nearby[j].rel_xdot > 0.0))||
              ((v < 0.0)&&(s.nearby[j].rel_xdot < 0.0)) ) {
            brk /= s.nearby[j].rel_xdot;
            v   /= s.nearby[j].rel_xdot;
         }
         else if ( (brk == 0.0)&&(s.nearby[j].rel_xdot == 0.0) )
            v = 100000.0;
         else
            continue;   // not going to hit this car

         // try to avoid crashes
         if ( (brk <= oc.pt.x)&&(oc.pt.x < v) )
            brk = oc.pt.x; // time to possible rear end collision
         else if ( (oc.pt.x > brk)||(brk >= oc.pt.y) )
            continue;      // no collision detected

         v = s.nearby[j].v;
         if (s.nearby[j].braking)
            v *= 0.96;
         if (brk > delta_time) {
            v  = (v - s.v) * (delta_time / brk);
            v += s.v;
         }
      }

      if (v < min_v)
         min_v = v;
   }
   return ((min_v < tgt_v) ? (min_v / tgt_v) : 1.0);
}

int SBv2r3::pass_clear(situation &s, double trk_dist)
{
   int    j;
   double min_to_lft, max_to_lft, max_dist;

   locate_car(&min_to_lft, NULL, cV.pt, trk_dist, -1);
   if (fabs(min_to_lft) > (2.5*CARLEN))
      return 0;
   else if (min_to_lft > CARWID) {
      // pass on the left side of track
      min_to_lft = 0.0;
      max_dist   = 2.0 * SB2_MIN_DIST;

      for (j = 0; j < NEARBY_CARS; j++) {
         if ( (s.nearby[j].who >= MAX_CARS)||
              (s.nearby[j].rel_y < (-0.5*CARLEN))||
              (s.nearby[j].rel_y > max_dist)||
              (s.nearby[j].to_lft < -CARLEN)||
              (s.nearby[j].to_rgt < 0.0) )
            continue;
         if ( ((s.nearby[j].rel_y < CARLEN)||(j == car2pass))&&
              (s.nearby[j].to_lft < (s.to_lft-CARWID)) )
            return 0;
         else if (j != car2pass) {
            max_to_lft  = 1.0 - (s.nearby[j].rel_y / max_dist);
            max_to_lft *= s.to_lft - min_to_lft - CARWID;
            max_to_lft += min_to_lft + CARWID;
            if ( (s.nearby[j].to_lft > min_to_lft)&&
                 (s.nearby[j].to_lft < max_to_lft) )
               return 0;
         }
      }
   }
   else if (min_to_lft < -CARWID) {
      // pass on the right side of track
      max_to_lft = s.to_lft + s.to_rgt;
      max_dist   = 2.0 * SB2_MIN_DIST;

      for (j = 0; j < NEARBY_CARS; j++) {
         if ( (s.nearby[j].who >= MAX_CARS)||
              (s.nearby[j].rel_y < (-0.5*CARLEN))||
              (s.nearby[j].rel_y > max_dist)||
              (s.nearby[j].to_lft < 0.0)||
              (s.nearby[j].to_rgt < -CARLEN) )
            continue;
         if ( ((s.nearby[j].rel_y < CARLEN)||(j == car2pass))&&
              (s.nearby[j].to_lft > (s.to_lft+CARWID)) )
            return 0;
         else if (j != car2pass) {
            min_to_lft  = 1.0 - (s.nearby[j].rel_y / max_dist);
            min_to_lft *= s.to_lft - max_to_lft + CARWID;
            min_to_lft += max_to_lft - CARWID;
            if ( (s.nearby[j].to_lft > min_to_lft)&&
                 (s.nearby[j].to_lft < max_to_lft) )
               return 0;
         }
      }
   }
   return 1;   // passing path is clear
}
//-----END SB2Driver.cpp-----//


// STATIC FUNCTIONS
//-----START spd_func.cpp-----//
double SB2_crn_spd(double radius)
{
   double retval;

   if (radius == 0.0)
      return SB2_MAX_SPD;

   retval = sqrt(radius * MAX_ACCEL);
   if (retval > SB2_MAX_SPD)
      retval = SB2_MAX_SPD;

   return retval;
}

double SB2_brkdist(double vi, double vf, double drag_accel)
{
   double scale;

   if (vf >= vi)
      return 0.0;

#ifndef SB2_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 0:
      scale = 1.0 + ((vi - vf)*0.0020);
      break;
   case 1:
      scale = 1.0 + ((vi - vf)*0.0018);
      break;
   case 2:
      scale = 1.0 + ((vi - vf)*0.0015);
   };

   vi *= vi * drag_accel;
   vi += MAX_ACCEL;
   vf *= vf * drag_accel;
   vf += MAX_ACCEL;
   return (scale * log(vf/vi) / (-2.0*drag_accel));
}

// inverse friction function
double SB2_iu(double cur_u, int surfaceType)
{
   double retval;

   switch (surfaceType) {
   case 2:
      if (cur_u >= MYU_MAX2)
         retval = 20.0;
      else
         retval = -SLIPPING * log(1.0 - (cur_u/MYU_MAX2));
      break;
   case 1:
      if (cur_u >= MYU_MAX1)
         retval = 20.0;
      else
         retval = -SLIPPING * log(1.0 - (cur_u/MYU_MAX1));
      break;
   default:
      if (cur_u >= MYU_MAX0)
         retval = 20.0;
      else
         retval = ((SLIPPING*cur_u)/MYU_MAX0) / (1.0 - (cur_u/MYU_MAX0));
   }
   return retval;
}

double SB2_fuel_needed(Stage s, int lapsleft, double fpl, double cur_fuel)
{
   double retval;

   if (fpl <= 0.0)
      return MAX_FUEL;
   else if (s == PRACTICE)
      retval = fpl * 10.0;
   else if (s == QUALIFYING)
      retval = fpl * 2.5;
   else
      retval = fpl * (double)lapsleft;

   retval -= cur_fuel;
   if (retval < 0.0)
      return 0.0;
   else if (retval > MAX_FUEL)
      return MAX_FUEL;
   return retval;
}
//-----END spd_func.cpp-----//


Driver * getSBv2r3Instance()
{
   static int i = -SB2_USE_DEFAULT_NAME;
   
   return new SBv2r3(i++);
}

#undef SB2_car_mass
#undef SB2_drag
#undef SB2_drag_accel
#undef SB2_CARS
#undef SB2_USE_DEFAULT_NAME
